open Arg
open Fscw_tree
open Wl
open Wl_require
open Wl_generate
open Ctt_abstree

let dprintf l f = Debug.dprintf ~category:2 l f
let get_debug_flags () = Debug.get_debug_flags ~category:2

let make_cpp_cmdline systemdir includedirs defines fname = 
  let sysincludedir =
    if systemdir = None || Option.get systemdir = "" then failwith "panic: systemdir not specified"
    else Option.get systemdir ^ "/include"
  in
  let definition_endian =
    if Fsc_config.is_littleendian
    then "-D__failsafeC_little_endian"
    else "-D__failsafeC_big_endian" in
  let args =
    ["cpp"; "-nostdinc"] @
    ["-idirafter"; sysincludedir] @
    (Util.map_flatten (fun d -> ["-I"; d]) includedirs) @
    ["-ansi"; "-undef"; "-D__failsafeC"; definition_endian] @ defines @
    [fname]
  in
  (String.concat " "
     (Util.map_flatten
	(fun s -> [Filename.quote s]) args))

let separate_fscw fn =
  Fscw_tree.reset fn;
  let chan = open_in fn in
  let lexbuf = Lexing.from_channel chan in
  let tokenizer = Fscw_lexer.make_tokenizer () in
  let t = Fscw_parser.parse_fscw tokenizer lexbuf in
  t

let read_chan chan =
  C_parserhelper.reset "";
  let lexbuf = Lexing.from_channel chan in
  C_parser.translation_unit C_lexer.token lexbuf

let output_string_and_pos outch (str, (file, line)) =
  Printf.fprintf outch "#line %d \"%s\"\n" line file;
  output_string outch str

let process ~infile ~outfile systemdir includedirs defines =
  let fscw = separate_fscw infile in
  (*Printf.printf "P:%s\nC:%s\nF:%s\nR:%s\nN:%s\n"
    fscw.r_prologue fscw.r_c_prologue fscw.r_fsc_prologue fscw.r_required_decls fscw.r_native_implementation;*)

  let outch = open_out outfile in
  begin try
  output_string_and_pos outch fscw.r_c_prologue;
  let (_, (_, line_1)), (_, (_, line_2)) = fscw.r_c_prologue, fscw.r_fsc_prologue in
  Printf.fprintf outch "#line %d \"%s\"\n"
                       (line_2 - line_1 + 2)
                       outfile;
  let tmpfn, tmpch = Filename.open_temp_file "fscw" ".c" in
  output_string_and_pos tmpch fscw.r_fsc_prologue;
  close_out tmpch;
  let cmdline = make_cpp_cmdline systemdir includedirs defines tmpfn in

  let chan = Unix.open_process_in cmdline in
  dprintf 1 "parsing...";
  let t =
    try read_chan chan with e -> close_in chan; raise e
  in
  let r = Unix.close_process_in chan in
  Sys.remove tmpfn;
  if r <> Unix.WEXITED 0 then
    (dprintf 1 "preprocessor failed."; exit 1);

  let t = C_abstree_remove_definitions.f t in
  let genv, t, gv, dep, orig = C_global_only_typing.parse_global_declarations t in
  dprintf 1 "cleaning...";
  let t = Expunge_localstatic.reduce_program ~genv t in
  let genv, t = Ctt_pad_structures.translate_program genv t in
  let genv, t = Ctt_reduce_mdarray.reduce_program ~genv t in
  let genv, t = Ctt_reduce_unions.translate_program ~genv t in
  let genv, t = Ctt_add_tentative_initializer.reduce_program ~genv t in
  (*let genv, t = Ctt_remove_unuseddecls.f ~genv t in*)
  dprintf 1 "analyzing assignments...";
  let genv, t = genv, Cttm_expunge_assignment.translate_program t in
  let genv, t = genv, Separate_side_effect.translate_program_to_il0 ~genv t in
  let genv, t = Il0_translate_bitfield.translate_program ~genv t in
  let genv, t = Il0_type_canonify.translate_program ~genv t in

  let genv2 = make_genv2 genv gv dep orig in

  let req_list = List.map
    (function
    | RequireTypedef name -> if name = "fscw_primitive_types" then WLR_primitive_type  else WLR_function (WLTabstract name)
    (*| RequireStruct name -> WLR_typedef (WLTstruct (wl_struct_of_struct_tag ~genv2 name)) *)
    | RequireStruct name -> WLR_function (WLTstruct (wl_struct_of_struct_name ~genv2 name))
    | RequireFunction name -> WLR_wrapper_decl name
    | AutoGenerate name -> WLR_autogen name
    | EmitTypeinfo name -> WLR_emit_typeinfo (WLTstruct (wl_struct_of_struct_name ~genv2 name))
    ) (fst fscw.r_required_decls) in

  let req_list = Wl_require.resolve_required ~genv2 req_list in

  Record_globalnames.clear ();
  List.iter (function 
             | WLR_typeinfo wty when wty <> WLTvoid ->
               let p = parse_wl_type ~genv2 wty in
               Record_globalnames.require_name (Record_globalnames.GNtypeinfo p.tp_c_t)
             | WLR_emit_typeinfo (WLTstruct s) ->
               Add_support_funcs.specially_emit_structs := s.wl_struct_id::!Add_support_funcs.specially_emit_structs
             | _ -> ()) req_list;
  let rr = Add_support_funcs.f_standalone ~genv [] in
  C_pp.pp_print_program (Format.formatter_of_out_channel outch) rr;

  List.iter (pp_wlr (Format.formatter_of_out_channel outch) ~genv2) req_list;

  output_string_and_pos outch fscw.r_native_implementation;
  close_out outch;
  with
    e -> close_out outch; (try Sys.remove outfile with _ -> ()); raise e
  end;
  exit 0

let parse_f str =
  if str = "stdlib-implementation" then
    Translate.compiler_mode := Translate.StdlibImplementation
  else if str = "multi-module" then
    Translate.compiler_mode := Translate.MultiModule
  else
    raise (Arg.Bad ("unknown option -f" ^ str))

let _ =
  let systemdir = ref None in
  let includedirs = ref [] in
  let infile = ref None in
  let outfile = ref None in
  let setfile var s = var := Some s in
  let defines = ref [] in
  if not !Sys.interactive then begin
    let argspec = Arg.align
	([
	 "-I", String (fun s -> includedirs := !includedirs @ [s]), " specify include directories";
	 "-D", String (fun s -> defines := !defines @ ["-D" ^ s]), " declare preproessor symbols";
	 "-U", String (fun s -> defines := !defines @ ["-U" ^ s]), " undeclare preproessor symbols";
         (*"--debug", String (Debug.parse_debug_flags), " set compiler-debugging options"; *)
         "-f", String (fun s -> Translate.compiler_mode := Translate.StdlibImplementation), " (not documented)";
         "---systemdir",  String (setfile systemdir), " specify system include directory";
         "---input-file",  String (setfile infile), " specify name for .fscw input file";
         "---output-file", String (setfile outfile), " specify name for .c output file";
	 ])
    in
    let usagemsg = "Usage: fscw [options]" in
    let usage s = Arg.usage argspec (s ^ usagemsg); exit 1 in
    Arg.parse argspec (fun s -> raise (Bad "bad usage")) usagemsg;
    if !infile = None  then usage "no input file specified: ";
    if !outfile = None then usage "no output file specified: ";
    let infile  = Option.get !infile in
    let outfile = Option.get !outfile in
    Install_fsc_extension.install ();
    process ~infile:infile ~outfile:outfile !systemdir !includedirs !defines
  end

