(* 
   Part of Fail-Safe C Compiler. Produced by Yutaka Oiwa.
   (c) 2001-2005 Yutaka Oiwa.
   (c) 2005-2006 AIST.

   This file is written by Yutaka Oiwa in 2005-2006. *)

open Linker_types
open Parse_typeinfo

include Debug.Install (struct let category = 200 end)

module StrDesc_Set = Set.Make(struct type t = struct_desc let compare = compare end)

type equi_struct = struct_desc * StrDesc_Set.t

type linker_environment = {
    hash : string;
    modules : (string * module_info) list;
    struct_decls : (struct_desc, string * string * struct_declaration_info) Hashtbl.t;
    provided_decls : (string, string * ltype) Hashtbl.t;
    mutable equi_struct_rel : equi_struct list;

    (* updated during unification *)
    rename_func_set : (string, ltype) Hashtbl.t;
    referred_provided_decls : (string, string * ltype) Hashtbl.t;
  }

type t = { 
    new_module_hash : string;
    rename_list : 
      (struct_desc, struct_desc) Hashtbl.t;
    renamed_typeinfo :
      (ltype, ltype) Hashtbl.t;
    required_typeinfo :
      (ltype, unit) Hashtbl.t;
    provided_renamed_identifiers : (string * ltype * ltype) list;
    requested_renamed_identifiers : (string * ltype * ltype) list;
    struct_table :
      (int * (string option * (string * string * struct_declaration_info))) list;
    provided_values : (string, string * ltype) Hashtbl.t;
    renamed_unknown_functions : (string * ltype) list;
  }

exception Unification_failure

let assert_unify b = if not b then raise Unification_failure

let strof_strdesc = function
    LTSnamed s -> s
  | LTShashed(h,s) -> h ^ "::" ^ string_of_int s

let add_eqstr ~env s1 s2 = 
  (* s1, s2 is principle *)
  (* in adding stage, s1 should be new principal *)
  let iter_found first_occurence another_equi =
    let rec iter = function
	[] ->
	  dprintf 6 "  we do not found %s as principal. make %s principal, let %s be alias."
	    (strof_strdesc another_equi) (strof_strdesc s1) (strof_strdesc s2);
	  [s1, StrDesc_Set.add s2 (snd first_occurence)]
      | h :: t ->
	  if fst h = another_equi then begin
	    dprintf 6 "  we also find %s as principal. keep %s as principal, let %s be alias." 
	      (strof_strdesc another_equi) (strof_strdesc s1) (strof_strdesc s2);
	    (s1, 
	     StrDesc_Set.union
	       (StrDesc_Set.add s2 (snd first_occurence))
	       (snd h))::t
	  end else
	    h :: iter t
    in
    iter
  in

  let rec iter_not_found = function
      [] ->
	dprintf 6 "add_eqstr::iter_not_found: not found. take %s as principal, %s as alias." 
	  (strof_strdesc s1) (strof_strdesc s2);
	[s1, StrDesc_Set.singleton s2]
    | h :: t ->
	if fst h = s1 then begin
	  dprintf 6 "  we already has %s as principal. search for %s." 
	    (strof_strdesc s1) (strof_strdesc s2);
	  iter_found h s2 t
	end else if fst h = s2 then begin
	  dprintf 6 "  we already has %s as principal. search for %s." 
	    (strof_strdesc s2) (strof_strdesc s1);
	  iter_found h s1 t
	end else begin
	  h :: iter_not_found t
	end
  in
  dprintf 6 "add_eqstr: required to add relation %s <= %s" 
    (strof_strdesc s1) (strof_strdesc s2);
  env.equi_struct_rel <- iter_not_found env.equi_struct_rel

let get_struct_impl ~env s = 
  Hashtbl.find env.struct_decls s

let find_principal ~env s =
  let rec iter = function
      [] -> s
    | (p, l)::t ->
	if p = s then p
	else if StrDesc_Set.mem s l then p else iter t
  in iter env.equi_struct_rel

let rec check_eqtype ~env ?(left_nullarray_ok = false) ~symbol t1 t2 = 
  try
    match t1, t2 with
      LTconcrete c1, LTconcrete c2 -> 
	assert_unify (c1 = c2)
    | LTarray(s1,et1), LTarray(s2,et2) ->
	assert_unify ((s1 = s2 || left_nullarray_ok && s1 = 0));
	check_eqtype ~env ~symbol et1 et2
    | LTpointer et1, LTpointer et2 ->
	check_eqtype ~env ~symbol et1 et2
    | LTstruct s1, LTstruct s2 -> begin
	if (s1 = s2) then () else
	let s1 = find_principal ~env s1 in
	let s2 = find_principal ~env s2 in
	if s1 = s2 then () else
	let (_, nam1, decl1) = get_struct_impl ~env s1 in
	let (_, nam2, decl2) = get_struct_impl ~env s2 in
	
	(* both are equally unnamed or named *)
	assert_unify (nam1 = nam2);
	
	match decl1, decl2 with
	  Sspecial, Sspecial ->
	    (* special structs should be name-equal *)
	    assert_unify false;
	| Sspecial, Sconcrete _ 
	| Sconcrete _, Sspecial ->
	    (* special structs should not be concrete *)
	    assert_unify false;
	| Sconcrete c1, Sconcrete c2 ->
	    add_eqstr ~env s1 s2; (* for recursive check *)
	    check_struct_equality ~env ~symbol c1 c2
	| (Sconcrete _ | Sspecial), Sabstract ->
	    add_eqstr ~env s1 s2
	| Sabstract, (Sconcrete _ | Sspecial) ->
	    add_eqstr ~env s2 s1
	| Sabstract, Sabstract ->
	    add_eqstr ~env s1 s2
    end
    | LTfunction(at1,vp1,rt1), LTfunction(at2,vp2,rt2) ->
	assert_unify (vp1 = vp2);
	assert_unify (List.length at1 = List.length at2);
	List.iter2 (check_eqtype ~env ~symbol) at1 at2;
	check_eqtype ~env ~symbol rt1 rt2
    | _ -> assert_unify false
  with
    Unification_failure ->
      failwith ("error while resolving " ^ symbol ^ ":\n  type " ^ strof_ltype t1
		^ " is not compatible with " ^ strof_ltype t2)

and check_struct_equality ~env ~symbol l1 l2 = 
  assert_unify (List.length l1 = List.length l2);
  List.iter2
    (fun (nam1, t1) (nam2, t2) ->
      assert_unify (nam1 = nam2);
      check_eqtype ~env ~symbol t1 t2) l1 l2

let gather_struct_decls mods = 
  let ht = Hashtbl.create 16 in
  List.iter
    (fun (modname, modinfo) ->
      List.iter
	(fun (desc, snam, dcl) ->
	  if Hashtbl.mem ht desc then
	    match desc, dcl, Hashtbl.find ht desc with
	      LTSnamed _, (Sspecial | Sconcrete _), (_, _, Sabstract) ->
		Hashtbl.replace ht desc (modname, snam, dcl);
	    | LTSnamed _, Sabstract, (_, _, (Sspecial | Sconcrete _ | Sabstract)) ->
		()
	    | LTSnamed _, Sconcrete t, (_, _, Sconcrete t') when t = t' -> ()
	    | _, _, (omod, _, _) ->
		failwith ("duplicate decl of struct " ^ snam ^ " " ^ strof_strdesc desc 
			 ^ " between " ^ modname ^ " / " ^ omod)
	  else
	    Hashtbl.add ht desc (modname, snam, dcl))
	modinfo.mi_structs) mods;
  ht

let gather_provided_values mods = 
  let ht = Hashtbl.create 16 in
  List.iter
    (fun (modname, modinfo) ->
      List.iter
	(fun (nam, typ) ->
	  if Hashtbl.mem ht nam then
	    failwith ("duplicate decl of value " ^ nam ^ " between "
		      ^ modname ^ " / " 
		      ^ fst (Hashtbl.find ht nam))
	  else
	    Hashtbl.add ht nam (modname, typ))
	modinfo.mi_provided_values) mods;
  ht

let unify_typeof_referred_values ~env modules = 
  List.iter
    (fun (modname, modinfo) ->
      dprintf_progress "[%s]" modname;
      List.iter
	(fun (nam, rtyp) ->
	  try
	    let (mname, dtyp) as p = Hashtbl.find env.provided_decls nam in
	    Hashtbl.replace env.referred_provided_decls nam p;
	    if rtyp = LTunknownfunc then begin
	      match dtyp with
		LTfunction _ -> Hashtbl.add env.rename_func_set nam dtyp
	      | _ -> failwith 
		    (Printf.sprintf "%s is used as unknown function in module \"%s\", but provided as value in module \"%s\""
		       nam modname mname)
	    end else
	      check_eqtype ~env ~left_nullarray_ok:true ~symbol:nam rtyp dtyp
	  with
	    Not_found -> failwith ("no decl of value " ^ nam ^ " provided"))
	modinfo.mi_required_values) modules;
  (* add main as referred value, if provided *)
  if not (Hashtbl.mem env.referred_provided_decls "main") &&
    Hashtbl.mem env.provided_decls "main" then
    Hashtbl.add env.referred_provided_decls "main"
      (Hashtbl.find env.provided_decls "main");
  ()

let check_equi_relation ~env = 
  (* equi_relation should not contain types from the same module twice. *)
  List.iter
    (fun (p, l) ->
      let t = Hashtbl.create 16 in
      let check = function
	  LTSnamed _ -> ()
	| LTShashed (s, _) ->
	    if Hashtbl.mem t s then
	      failwith ("two structs from the module " ^ s ^ " have been unified during type checking")
	    else 
	      Hashtbl.replace t s ()
      in
      check p;
      StrDesc_Set.iter check l)
    env.equi_struct_rel

let dump_equi_relation ~env = 
  prerr_endline "\nEqui-type list:\n";
  List.iter
    (fun (p, l) ->
      prerr_endline (strof_strdesc p);
      StrDesc_Set.iter
	(fun v -> prerr_endline ("  = " ^ strof_strdesc v)) l;
      prerr_newline ())
    env.equi_struct_rel

exception Exit_StrDesc of struct_desc

let make_rename_list env = 
  let ht = Hashtbl.create 16 in
  let find_representive p l = 
    let check = function
      | (LTSnamed s as r) -> raise (Exit_StrDesc r)
      | _ -> ()
    in
    try
      check p;
      StrDesc_Set.iter check l;
      None
    with 
      Exit_StrDesc r -> Some r
  in
  List.iter
    (fun (p, l) ->
      match find_representive p l with
	None ->
	  StrDesc_Set.iter (fun s -> Hashtbl.replace ht s p) l
      | Some r ->
	  StrDesc_Set.iter
	    (fun s -> if s <> r then Hashtbl.replace ht s r)
	    (StrDesc_Set.add p l))
    env.equi_struct_rel;
  ht

let dump_rename_list renlist = 
  prerr_endline "\nType rename list:\n";
  Hashtbl.iter
    (fun r t ->
      Printf.eprintf "%s ==> %s\n" (strof_strdesc r) (strof_strdesc t)) renlist;
  prerr_newline ()

let dump_typeinfo_list tlist = 
  prerr_endline "\nTypeInfo list:\n";
  let l = Hashtbl.fold (fun x () y -> x::y) tlist [] in
  let l = List.sort compare l in
  List.iter
    (fun r -> prerr_endline (strof_ltype r)) l;
  prerr_newline ()

let dump_decl_rename_list s renlist = 
  prerr_endline ("\n" ^ s ^ " declaration rename list:\n");
  List.iter
    (fun (id, ot, nt) ->
      Printf.eprintf "%s: %s ==> %s\n" id (strof_ltype ot) (strof_ltype nt)) renlist;
  prerr_newline ()

let rec rename_type ren
    = function
	(LTconcrete _ as c) -> c
      | LTpointer t -> LTpointer (rename_type ren t)
      | LTarray(s, t) -> LTarray(s, rename_type ren t)
      | LTfunction(at,vp,rt) -> LTfunction(List.map (rename_type ren) at, vp, rename_type ren rt)
      | LTstruct s -> LTstruct (rename_struct_type ren s)
      | LTunknownfunc -> LTunknownfunc
and rename_struct_type ren s = 
  try
    Hashtbl.find ren s
  with
    Not_found -> s

let generate_renamed_typeinfo_list ren mods = 
  let gtbl = Hashtbl.create 16 in
  let rtbl = Hashtbl.create 16 in
  List.iter
    (fun (modname, modinfo) ->
      List.iter
	(fun typ ->
	  let ntyp = rename_type ren typ in
	  Hashtbl.replace gtbl ntyp ();
	  if ntyp <> typ then
	    Hashtbl.replace rtbl typ ntyp)
	modinfo.mi_required_typeinfo) mods;
  rtbl, gtbl

let generate_provided_values_table ren vals = 
  let ht = Hashtbl.create 16 in
  Hashtbl.iter
    (fun nam (modname, typ) ->
      Hashtbl.replace ht nam (modname, rename_type ren typ))
    vals;
  ht

let dump_provided_values table = 
  prerr_endline ("\nList of (referred) provided values:\n");
  Hashtbl.iter
    (fun id (m, lt) ->
      Printf.eprintf "%s: %s (defined in %s)\n" id (strof_ltype lt) m) table;
  prerr_newline ()

let generate_provided_renamed_decls ren mods = 
  let g = Glist.empty () in
  List.iter
    (fun (modname, modinfo) ->
      List.iter
	(fun (nam, typ) ->
	  let new_type = rename_type ren typ in
	  if new_type <> typ then
	    Glist.put g (nam, typ, new_type))
	modinfo.mi_provided_values) mods;
  Glist.to_list g

let generate_requested_renamed_decls ren mods = 
  let g = Glist.empty () in
  List.iter
    (fun (modname, modinfo) ->
      List.iter
	(fun (nam, typ) ->
	  let new_type = rename_type ren typ in
	  if new_type <> typ then
	    Glist.put g (nam, typ, new_type))
	modinfo.mi_required_values) mods;
  Glist.to_list g

let generate_rename_func_set ren s =
  let g = Glist.empty () in
  Hashtbl.iter
    (fun id lt ->
      Glist.put g (id, rename_type ren lt))
    s;
  Glist.to_list g

let generate_renamed_struct_decl_list ~env ren strdecls = 
  let new_rename_tbl = Hashtbl.create 16 in
  let new_decls = Glist.empty () in
  let idctr = ref 0 in
  Hashtbl.iter
    (fun desc declinfo ->
      dprintf 5 "\niter : %s" (strof_strdesc desc);
      if Hashtbl.mem new_rename_tbl desc then begin
	match desc with
	  LTSnamed s -> ()
	| _ -> assert (rename_struct_type ren desc = desc);
      end else begin
	let canon_desc = rename_struct_type ren desc in
	dprintf 5 "canon : %s" (strof_strdesc canon_desc);
	let updated_desc = 
	  try 
	    Hashtbl.find new_rename_tbl canon_desc
	  with
	    Not_found -> begin
	      let newid = (incr idctr; !idctr) in
	      let name, updated_desc = 
		match canon_desc with
		  LTShashed _ -> None, LTShashed (env.hash, newid)
		| LTSnamed s -> Some s, canon_desc
	      in
	      Glist.put new_decls (newid, (name, Hashtbl.find env.struct_decls canon_desc));
	      Hashtbl.replace new_rename_tbl canon_desc updated_desc;
	      updated_desc
	    end
	in
	dprintf 5 "upd : %s" (strof_strdesc updated_desc);
	Hashtbl.replace new_rename_tbl desc updated_desc
      end)
    strdecls;
  new_rename_tbl,
  List.map
    (fun (id, (namopt, (modnam, orgnam, decl))) ->
      let decl =
	match decl with
	  Sconcrete l ->
	    Sconcrete
	      (List.map
		 (fun (n, t) -> n, rename_type new_rename_tbl t)
		 l)
	| (Sabstract | Sspecial) -> decl
      in
      (id, (namopt, (modnam, orgnam, decl))))
    (Glist.to_list new_decls)

let dump_renamed_struct_decl_list decls = 
  prerr_endline "\nStruct list after rename:\n";
  List.iter
    (fun (id, (nam, (modnam, origname, fldinfos))) ->
      prerr_endline
	(string_of_int id ^ 
	 (match nam with None -> ", "
	 | Some s -> " named \"" ^ s ^ "\", ")
	 ^ "defined in " ^ modnam ^ " as \"" ^ origname ^ "\"");
  match fldinfos with
	Sspecial -> prerr_endline "  Special struct.\n"
      | Sabstract -> prerr_endline "  Abstract (??).\n"
      | Sconcrete l ->
	  prerr_endline 
	    ("  { " ^
	     String.concat ", "
	       (List.map
		  (fun (fn, ft) ->
		    fn ^ ": " ^ strof_ltype ft)
		  l)
	       ^ "}.\n"))
    decls


let resolve module_list = 
  let debug = get_debug_flags () >= 4 in
  let start_time = Unix.gettimeofday () in
  dprintf_start "resolving symbols...";

  let struct_decls = gather_struct_decls module_list in
  let value_decls = gather_provided_values module_list in
  let hash = Printf.sprintf "%08x" (Hashtbl.hash_param max_int max_int (struct_decls, value_decls)) in
  if debug then Printf.eprintf "Module hash is %s\n\n" hash;
  let env = {
    hash = hash;
    modules = module_list;
    struct_decls = struct_decls;
    provided_decls = value_decls;
    equi_struct_rel = [];

    rename_func_set = Hashtbl.create 16;
    referred_provided_decls = Hashtbl.create 16  
  } in
  let rename_func_set = Hashtbl.create 16 in

  unify_typeof_referred_values ~env module_list;
  check_equi_relation ~env;
  if debug then dump_equi_relation ~env;

  dprintf_progress "(PRN)";
  let pre_rename_list = make_rename_list env in
  if debug then dump_rename_list pre_rename_list;

  dprintf_progress "(RSL)";
  let rename_list, new_struct_decls = 
    generate_renamed_struct_decl_list ~env pre_rename_list env.struct_decls in
  if debug then begin
    dump_renamed_struct_decl_list new_struct_decls;
    dump_rename_list rename_list;
  end;

  dprintf_progress "(RTI)";
  let renamed_typeinfo, new_required_typeinfos = 
    generate_renamed_typeinfo_list rename_list module_list in
  if debug then dump_typeinfo_list new_required_typeinfos;

  dprintf_progress "(RDC)";
  let provided_renamed_decls = 
    generate_provided_renamed_decls rename_list module_list in
  if debug then dump_decl_rename_list "provided" provided_renamed_decls;

  let required_renamed_decls = 
    generate_requested_renamed_decls rename_list module_list in
  if debug then dump_decl_rename_list "required" required_renamed_decls;

  dprintf_progress "(PVT)";
  let provided_values_table = 
    generate_provided_values_table rename_list env.referred_provided_decls in
  if debug then dump_provided_values provided_values_table;

  dprintf_progress "(RFT)";
  let rename_func_table =
    generate_rename_func_set rename_list env.rename_func_set in
  
  dprintf_end "done (%.3f sec)." (Unix.gettimeofday() -. start_time);
  { new_module_hash = hash;
    rename_list = rename_list;
    renamed_typeinfo = renamed_typeinfo;
    required_typeinfo = new_required_typeinfos;
    provided_renamed_identifiers = provided_renamed_decls;
    requested_renamed_identifiers = required_renamed_decls;
    struct_table = new_struct_decls;
    provided_values = provided_values_table;
    renamed_unknown_functions = rename_func_table;
  }

let f modules = 
  resolve modules
