(* 
   Part of Fail-Safe C Compiler. Produced by Yutaka Oiwa.
   (c) 2001-2005 Yutaka Oiwa.
   (c) 2005-2007 AIST.

   This file is written by Yutaka Oiwa in 2003-2004. *)

open Big_int
open Ctt_abstree
open C_typing

let dummy_loc = Locterm.dummy_location

let initializer_zero typ = 
  let d = (CTTexpConstant (CTTconstInteger zero_big_int)) in
  CTTinitExp(make_expr d typ dummy_loc)

let initializer_float_zero typ = 
  let d = (CTTexpConstant (CTTconstFloat 0.0)) in
  CTTinitExp(make_expr d typ dummy_loc)

let initializer_null typ = 
  let d = (CTTexpConstant CTTconstNull) in
  CTTinitExp(make_expr d typ dummy_loc)

let rec fill_initializer ~genv ~is_static typ = 
  match typ.ct_ty with
    Tvoid -> assert false
  | Tbuiltin (Tfloat | Tdouble | Tlongdouble) ->
      typ, initializer_float_zero typ
  | Tbuiltin _ ->
      typ, initializer_zero typ
  | Tpointer _ ->
      typ, initializer_null typ
  | Tarray(et, Some sz) ->
      let et, i = fill_initializer ~genv ~is_static et in
      let szi = int_of_big_int sz in
      let init =
	if !C_typing.fill_zero_initializer 
	then
	  CTTinitList(Util.list_repeat i szi)
	else
	  CTTinitList [i]
      in
      Ctt_abstree.make_c_type (Tarray(et, Some sz)), init
  | Tarray(et, None) ->
      if is_static then
	failwith "incomplete static declaration"
      else
	let et, init = fill_initializer ~genv ~is_static et in
	Ctt_abstree.make_c_type (Tarray(et, Some unit_big_int)),
	CTTinitList [init]
  | Tstruct(id) ->
      let sdesc = Ctt_abstree.get_struct_desc ~genv id in
      if sdesc.str_size = None then
	failwith "initializing incomplete struct"
      else
	typ,
	CTTinitList
	  (fill_struct_initializer ~genv ~is_static sdesc.str_fields)
  | Tabstract(_) -> assert false
  | Tfunction _ -> assert false (* see reduce_gdecl *)

and fill_struct_initializer ~genv ~is_static = function
  | [] -> []
  | (_, NormalField { sf_type = t }) :: tl ->
      snd (fill_initializer ~genv ~is_static t) :: fill_struct_initializer ~genv ~is_static tl
  | (_, BitField { s_bf_fields = bfs }) :: tl ->
      let rec iter = function
	  [] -> fill_struct_initializer ~genv ~is_static tl
	| (None, _, _, _)::l -> iter l
	| (Some _, t, _, _)::l ->
	    snd (fill_initializer ~genv ~is_static t) :: iter l
      in
      iter bfs

let reduce_gdecl ~genv = function
    CTTdeclFunction(_, typ, ident, _, _)
  | CTTdeclVariable(_, typ, ident, Some _)
  | CTTdeclVariable(Extern _, typ, ident, None) as decl ->
      decl
  | CTTdeclVariable((Global | ModuleStatic) as sclass, typ, ident, None) -> begin
      let is_static = sclass = ModuleStatic in
      match typ.ct_ty with
	Tfunction _ -> 
	  CTTdeclVariable
	    ((if is_static then ModuleStatic else Extern []),
	     typ, ident, None)
      | _ ->
	  let typ, init = fill_initializer ~genv ~is_static typ in
	  CTTdeclVariable(sclass, typ, ident, Some init)
  end

let rec reduce_program ~genv prog = 
  let prog = Util.list_map (reduce_gdecl ~genv) prog in
  genv, prog
