(* 
   Part of Fail-Safe C Compiler. Produced by Yutaka Oiwa.
   (c) 2001-2005 Yutaka Oiwa.
   (c) 2005-     AIST.

   This file is written by Yutaka Oiwa in 2006. *)

open Util
open Big_int_infix
open Locterm
open Ctt_abstree
open Cttm_abstree

(** helper functions. *)

external is_printable : char -> bool = "caml_is_printable"

let encode_c_string s = 
  (* borrowed from OCaml's string.ml *)
  let n = ref 0 in
  for i = 0 to String.length s - 1 do
    n := !n +
        (match String.unsafe_get s i with
          '"' | '\\' | '\n' | '\t' -> 2
        | c -> if is_printable c then 1 else 4)
  done;
  if !n = String.length s then s else begin
    let s' = String.create !n in
    n := 0;
    for i = 0 to String.length s - 1 do
      begin
        match String.unsafe_get s i with
          ('"' | '\\') as c ->
            String.unsafe_set s' !n '\\'; incr n; String.unsafe_set s' !n c
        | '\010' ->
            String.unsafe_set s' !n '\\'; incr n; String.unsafe_set s' !n 'n'
        | '\t' ->
            String.unsafe_set s' !n '\\'; incr n; String.unsafe_set s' !n 't'
        | c ->
            if is_printable c then
              String.unsafe_set s' !n c
            else begin
              let a = Char.code c in
              String.unsafe_set s' !n '\\';
              incr n;
              String.unsafe_set s' !n (Char.chr (48 + a / 64));
              incr n;
              String.unsafe_set s' !n (Char.chr (48 + (a / 8) mod 8));
              incr n;
              String.unsafe_set s' !n (Char.chr (48 + a mod 8))
            end
      end;
      incr n
    done;
    s'
  end

(** to_string functions for common data types *)

let strof_builtin_type = function
    Tchar -> "char"
  | Tschar -> "signed char"
  | Tuchar -> "unsigned char"
  | Tshort -> "short"
  | Tushort -> "unsigned short"
  | Tint -> "int"
  | Tuint -> "unsigned int"
  | Tlong -> "long"
  | Tulong -> "unsigned long"
  | Tlonglong -> "long long"
  | Tulonglong -> "unsigned long long"
  | Tfloat -> "float"
  | Tdouble -> "double"
  | Tlongdouble -> "long double"

let rec strof_c_type_qualifiers t = 
  match t.ct_volatile_p, t.ct_const_p with
    true, true -> "volatile const "
  | true, false -> "volatile "
  | false, true -> "const "
  | false, false -> ""

let rec strof_c_type ?(name="") ?(argnames=[]) t = 
  let rec iter t sp lv s tl = 
    match t.ct_ty with
    | Tstruct(id) ->
	strof_c_type_qualifiers t ^ "struct S" ^ string_of_int id ^ sp ^ s ^ tl
    | Tabstract str ->
	"<abstract>" ^ str ^ sp ^ s ^ tl
    | Tbuiltin(bt) ->
	strof_c_type_qualifiers t ^ strof_builtin_type bt ^ sp ^ s ^ tl
    | Tvoid -> strof_c_type_qualifiers t ^ "void" ^ sp ^ s ^ tl
    | Tpointer t1 ->
	iter t1 " " 1 ("*" ^ strof_c_type_qualifiers t ^ s) tl
    | Tfunction(at, vt, rt) when Debug.get_debug_flags ~category:12 = 10 ->
	(* alternative form *)
	"[" ^ 
	(if at = [] then
	  (if vt then "..." else "void") ^ (if argnames <> [] then "<some args... why?>" else "")
	else
	  if argnames = [] then
	    Util.string_of_list strof_c_type ", " at
	  else
	    String.concat ", " (Util.list_map2 (fun t s -> s ^ " : " ^ strof_c_type t) at argnames)) ^
	" -> " ^ strof_c_type rt ^ "]" ^ sp ^ s ^ tl
    | _ when lv >= 1 ->
	iter t " " 0 ("(" ^ s ^ tl ^ ")") ""
    | Tarray(t,None) ->
	iter t sp lv s (tl ^ "[]")
    | Tarray(t,Some sz) ->
	iter t sp lv s (tl ^ "[" ^ Big_int.string_of_big_int sz ^ "]")
    | Tfunction([],false,rt) ->
	iter rt sp lv s (tl ^ "(void)")
    | Tfunction([],true,rt) ->
	iter rt sp lv s (tl ^ "(...)")
    | Tfunction(at,vt,rt) ->
	iter rt sp lv s 
	  (tl ^ "(" 
	   ^ String.concat ", " 
	       (if List.length argnames = List.length at then
		 Util.list_map2 (fun t n -> strof_c_type ~name:n t) at argnames
	       else list_map (fun t -> strof_c_type ?name:None t) at)
	   ^ (if vt then ", ..." else "") ^ ")")
  in
  iter t (if name = "" then "" else " ") 0 name ""

let strof_union_flag = function
    Struct -> "struct"
  | Union -> "union"

let strof_binop = function
    CTTbinTimes -> "*"
  | CTTbinDiv -> "/"
  | CTTbinPlusVV -> "+"
  | CTTbinMinusVV -> "-"
  | CTTbinPostPlusVV -> "(post+)"
  | CTTbinPostMinusVV -> "(post-)"
  | CTTbinPlusPV -> ".+"
  | CTTbinMinusPP -> ".-."
  | CTTbinMinusPV -> ".-"
  | CTTbinPostPlusPV -> "(post.+)"
  | CTTbinPostMinusPV -> "(post.-)"
  | CTTbinModulo -> "%"
  | CTTbinLshift -> "<<"
  | CTTbinRshift -> ">>"
  | CTTbinLogAnd -> "&&"
  | CTTbinLogOr -> "||"
  | CTTbinIntAnd -> "&"
  | CTTbinIntOr -> "|"
  | CTTbinIntXor -> "^"
  | CTTbinLessThan -> "<"
  | CTTbinLessEqual -> "<="
  | CTTbinGtrThan -> ">"
  | CTTbinGtrEqual -> ">="
  | CTTbinEqual -> "=="
  | CTTbinNotEqual -> "!="

let strof_unaryop = function
    UnaryPlus -> "+"
  | UnaryMinus -> "-"
  | LogNot -> "!"
  | IntNot -> "~"

let strof_ctt_constant = function
    CTTconstNull -> "NULL"
  | CTTconstInteger bi -> string_of_big_int bi
  | CTTconstFloat f -> string_of_float f
  | CTTconstString s -> "\"" ^ encode_c_string s ^ "\""
  | CTTconstTypeInfo t -> "__typeinfo(" ^ strof_c_type t ^ ")"
  | CTTconstAbstract s -> "<abstract>" ^ s

let strof_local_storage_class = function
    Auto -> "auto"
  | Register -> "register"
  | LocalStatic -> "static"
  | FuncArgs -> ""

let strof_global_storage_class = function
    Extern [] -> "extern"
  | Extern _l -> "extern *extended*"
  | Global -> "global"
  | ModuleStatic -> "static"

(** CTT datatypes. *)

let strof_expr_desc, strof_expr = 
  let rec ed ~st e = 
    let str_term, st2 = match e.expr_t with
      CTTexpComma(e1,e2) ->
	s ~st e1 ^ ", " ^ s ~st e2, false
    | CTTexpAssign(e1,e2) ->
	s ~st e1 ^ " = " ^ s ~st e2, false
    | CTTexpBinAssign(bop,e1,_,e2) ->
	s ~st e1 ^ " " ^ strof_binop bop ^ "= " ^ s ~st e2, false
    | CTTexpConditional(e1,e2,e3) ->
	s ~st e1 ^ " ? " ^ s ~st e2 ^ s ~st e3, true
    | CTTexpBinExpr(bop,e1,e2) ->
	s ~st e1 ^ " " ^ strof_binop bop ^ " " ^ s ~st e2, false
    | CTTexpCoerce(ct,e1) ->
	"(" ^ strof_c_type ct ^ ")" ^ s ~st e1, false
    | CTTexpUnaryExpr(uop,e1) ->
	strof_unaryop uop ^ s ~st e1, false
    | CTTexpAddress(e1) ->
	"&" ^ s ~st e1, true
    | CTTexpPtrDeref(e1) ->
	"*" ^ s ~st e1, false
    | CTTexpInvoke(e1,es) ->
	s ~st e1 ^ "(" ^
	String.concat ", " (list_map (s ~st) es) ^ ")", true
    | CTTexpField(e1,id) ->
	s ~st e1 ^ "." ^ id, true
    | CTTexpConstant(ct) ->
	strof_ctt_constant ct, true
    | CTTexpVar (id, ct) ->
	"[" ^ id ^ " : " ^ strof_c_type ct ^ "]", true
    in
    if st && st2 then "(" ^ str_term ^ " [" ^ strof_c_type e.expr_type ^ "])"
    else"(" ^ str_term ^ ")"
  and s ~st e = 
    ed ~st (locval e)
  in
  (fun ?(show_type = true) e -> ed ~st:show_type e),
  (fun ?(show_type = true) e -> s ~st:show_type e)

let rec strof_ctt_initalizer = function
    CTTinitExp e -> strof_expr e
  | CTTinitList l -> "{" ^ String.concat ", " (list_map strof_ctt_initalizer l) ^ "}"

let strof_local_variable_declaration (sc,ct,id,initopt) = 
  strof_local_storage_class sc
  ^ " " ^ strof_c_type ~name:id ct ^ 
  match initopt with
    None -> ";"
  | Some i -> " = " ^ strof_ctt_initalizer i ^ ";"

let strof_statement st = 
  let rec sd ind = function
      CTTstmtNull -> ";"
    | CTTstmtExpr e -> strof_expr e ^ ";"
    | CTTstmtLabeled(id,st) -> id ^ ": " ^ s ind st
    | CTTstmtCase_Labeled(n,st) -> "case " ^ string_of_big_int n ^ ": " ^ s ind st
    | CTTstmtDefault_Labeled st -> "default: " ^ s ind st
    | CTTstmtCompound(decls,stmts) ->
      let new_ind = ind ^ "  " in
      "{\n" ^ new_ind ^
      String.concat ("\n" ^ new_ind)
	(list_map strof_local_variable_declaration decls)
      ^ "\n" ^ new_ind ^ 
      String.concat ("\n" ^ new_ind)
	(list_map (s new_ind) stmts)
      ^ "\n" ^ ind ^ "}"
    | CTTstmtIf(e,st,None) ->
	"if (" ^ strof_expr e ^ ") then " ^ s (ind ^ "  ") st
    | CTTstmtIf(e,st1,Some st2) ->
	"if (" ^ strof_expr e ^ ") then " ^ s (ind ^ "  ") st1
        ^ "else " ^ s (ind ^ "  ") st2
    | CTTstmtSwitch(e,st) ->
	"switch (" ^ strof_expr e ^ ") " ^ s (ind ^ "  ") st
    | CTTstmtWhile(e,st) ->
	"while (" ^ strof_expr e ^ ") " ^ s (ind ^ "  ") st
    | CTTstmtDoWhile(st,e) ->
	"do " ^ s (ind ^ "  ") st ^ "while (" ^ strof_expr e ^ ");"
    | CTTstmtFor(e1o,e2o,e3o,st) ->
	"for (" ^ 
	(match e1o with None -> "" | Some e -> strof_expr e) ^
	"; " ^
	(match e2o with None -> "" | Some e -> strof_expr e) ^
	"; " ^
	(match e3o with None -> "" | Some e -> strof_expr e) ^
	") " ^ s (ind ^ "  ") st
    | CTTstmtGoto id -> "goto " ^ id ^ ";"
    | CTTstmtContinue -> "continue;"
    | CTTstmtBreak -> "break;"
    | CTTstmtReturn(None) -> "return;"
    | CTTstmtReturn(Some e) -> "return " ^ strof_expr e ^ ";"
  and s ind st = sd ind (locval st)
  in s "" st

let strof_global_declaration = function
  CTTdeclVariable(gs,ct,id,initopt) -> begin
    strof_global_storage_class gs
    ^ " " ^ strof_c_type ~name:id ct ^ 
    match initopt with
      None -> ";\n\n"
    | Some i -> " = " ^ strof_ctt_initalizer i ^ ";\n\n"
  end
  | CTTdeclFunction(gs,ct,id,ids,st) ->
    strof_global_storage_class gs
      ^ " " ^ strof_c_type ~name:id ~argnames:ids ct ^ 
      strof_statement st ^ "\n\n"


(** CTTM datatypes. *)

let strof_mexpr_desc, strof_mexpr, strof_cttm_lv,
  strof_fields_suffix = 
  let rec ed ~st e = 
    let str_term, st2 = match e.mexpr_t with
      CTTMexpComma(e1,e2) ->
	s ~st e1 ^ ", " ^ s ~st e2, false
    | CTTMexpAddress(lv,fs) ->
	"&" ^ mo ~st lv ^ f ~st fs, true
    | CTTMexpRead(lv,fs) ->
	mo ~st lv ^ f ~st fs, true
    | CTTMexpWrite(lv,fs,None,expr) ->
	mo ~st lv ^ f ~st fs ^ " := " ^ s ~st e, false
    | CTTMexpWrite(lv,fs,Some (binop,_),expr) ->
	mo ~st lv ^ f ~st fs
	^ " :" ^ strof_binop binop ^ " = " ^ s ~st e, false
    | CTTMexpConditional(e1,e2,e3) ->
	s ~st e1 ^ " ? " ^ s ~st e2 ^ s ~st e3, true
    | CTTMexpBinExpr(bop,e1,e2) ->
	s ~st e1 ^ " " ^ strof_binop bop ^ " " ^ s ~st e2, false
    | CTTMexpCoerce(ct,e1) ->
	"(" ^ strof_c_type ct ^ ")" ^ s ~st e1, false
    | CTTMexpUnaryExpr(uop,e1) ->
	strof_unaryop uop ^ s ~st e1, false
    | CTTMexpInvoke(lv,es) ->
	mo ~st lv ^ "(" ^
	String.concat ", " (list_map (s ~st) es) ^ ")", true
    | CTTMexpConstant(ct) ->
	strof_ctt_constant ct, true
    in
    if st && st2 then "(" ^ str_term ^ " [" ^ strof_c_type e.mexpr_type ^ "])"
    else"(" ^ str_term ^ ")"
  and s ~st e = 
    ed ~st e
  and mo ~st = function
      CTTMlvPtr e ->
	"*" ^ s ~st e
    | CTTMlvVar(id,ct) ->
	id ^ if st then " [" ^ strof_c_type ct ^ "]" else ""
    | CTTMlvRvalue e ->
	"(" ^ s ~st e ^ ")"
  and f ~st l = 
    String.concat ""
      (Util.list_map
	 (fun (f,ct) -> 
	   if st then "." ^ f
	   else "." ^ f ^ " [" ^ strof_c_type ct ^ "]") l)
  in
  (fun ?(show_type = true) e -> ed ~st:show_type e),
  (fun ?(show_type = true) e -> s ~st:show_type e),
  (fun ?(show_type = true) e -> mo ~st:show_type e),
  (fun ?(show_type = true) e -> f ~st:show_type e)

let rec strof_cttm_initalizer = function
    CTTMinitExp e -> strof_mexpr e
  | CTTMinitList l -> "{" ^ String.concat ", " (list_map strof_cttm_initalizer l) ^ "}"

