(* 
   Part of Fail-Safe C Compiler. Produced by Yutaka Oiwa.
   (c) 2001-2005 Yutaka Oiwa.
   (c) 2005-2007 AIST.

   This file is written by Yutaka Oiwa in 2004. *)

(* change type of integers on memory to canonical unsigned integer *)

open Util
open Ctt_abstree
open Cttm_abstree

let canonical_types = Earray.empty ()
let canonical_signed_types = Earray.empty ()

let _ = 
  List.iter
    (fun bt ->
      Earray.set canonical_types (size_of_builtin_type bt) bt)
    [ Tulonglong; Tushort; Tulong; Tuchar; Tuint ]

let _ = 
  List.iter
    (fun bt ->
      Earray.set canonical_signed_types (size_of_builtin_type bt) bt)
    [ Tlonglong; Tshort; Tlong; Tschar; Tint ]
    
let rec translate_c_type_mem t = 
  let new_type = 
    match t.ct_ty with
      Tpointer t -> Tpointer (translate_c_type_mem t)
    | Tfunction(at,va,rt) ->
	Tfunction(Util.list_map translate_c_type_mem at, va, translate_c_type_mem rt)
    | Tbuiltin (Tchar | Tschar | Tuchar 
                | Tshort | Tushort 
		| Tint | Tuint
		| Tlong | Tulong
		| Tlonglong | Tulonglong as bt) ->
      let sz = size_of_builtin_type bt in
      Tbuiltin (Earray.get canonical_types sz)

    | Tbuiltin (Tfloat | Tdouble | Tlongdouble)
    | Tvoid
    | Tstruct _
    | Tabstract _ -> t.ct_ty
    | Tarray(ty,sz) -> Tarray(translate_c_type_mem ty, sz)
  in
  { t with ct_ty = new_type }

let rec is_signed_type t = 
  match t.ct_ty with
  | Tbuiltin (Tchar | Tschar | Tshort | Tint | Tlong | Tulonglong as bt) ->
    is_signed_builtin_type bt (* char may be either signed or unsigned *)
  | _ -> false

let translate_c_type_reg t = 
  match t.ct_ty with
  | Tbuiltin (Tchar | Tschar | Tshort | Tint | Tlong | Tulonglong as bt) 
    when is_signed_builtin_type bt (* char may be either signed or unsigned *) ->
      let sz = size_of_builtin_type bt in
      { t with ct_ty = Tbuiltin (Earray.get canonical_signed_types sz) }
  | _ -> translate_c_type_mem t

let make_mexpr e t = { mexpr_type = t; mexpr_t = e }

let coerce_to_unsigned mexp = 
  if is_signed_type mexp.mexpr_type then begin
    let t = translate_c_type_mem mexp.mexpr_type in
    make_mexpr (CTTMexpCoerce(t,mexp)) t
  end
  else mexp

let coerce_desc_to_signed mdesc t = 
  if is_signed_type t then
    CTTMexpCoerce(translate_c_type_reg t,
		  make_mexpr mdesc (translate_c_type_mem t))
  else
    mdesc

let rec translate_mexpr exp = 
  let new_exp = 
    match exp.mexpr_t with
    | CTTMexpComma(e1,e2) ->
	CTTMexpComma(translate_mexpr e1, translate_mexpr e2)
    | CTTMexpAddress(mobj,fld) ->
	CTTMexpAddress(translate_mem_object mobj, translate_fields fld)
    | CTTMexpRead(mobj,fld) ->
	let e = CTTMexpRead(translate_mem_object mobj, translate_fields fld) in
	coerce_desc_to_signed e exp.mexpr_type
    | CTTMexpWrite(mobj,fld,None,e1) ->
	let e1 = translate_mexpr e1 in
	let e1 = coerce_to_unsigned e1 in
	CTTMexpWrite(translate_mem_object mobj, translate_fields fld, None, e1)
    | CTTMexpWrite(mobj,fld,Some(bop,None),e1) ->
	let e1 = translate_mexpr e1 in
	let rty, e1 =
	  match is_signed_type exp.mexpr_type, is_signed_type e1.mexpr_type with
	    true, true ->
	      (* signed case: coerce lhs to signed, apply, and coerce back to unsigned *)
	      (* coerce back is implicit in type_annoted binop-write *)
	      Some (translate_c_type_reg exp.mexpr_type), e1
	  | false, true (* lhs is pointer, rhs is ptrdirr_t *) ->
	      None, coerce_to_unsigned e1
	  | true, false -> failwith "cttm_typechange_canonical:99"
	  | false, false -> None, e1
	in
	CTTMexpWrite(translate_mem_object mobj, translate_fields fld, 
		     Some(bop, rty), e1)
    | CTTMexpWrite(mobj,fld,Some(bop,Some imtyp),e1) ->
	let e1 = translate_mexpr e1 in
	let imtyp = translate_c_type_reg imtyp in
	CTTMexpWrite(translate_mem_object mobj, translate_fields fld, 
		     Some(bop, Some imtyp), e1)
    | CTTMexpConditional(e1,e2,e3) ->
	CTTMexpConditional(translate_mexpr e1, translate_mexpr e2, translate_mexpr e3)
    | CTTMexpBinExpr(bop,e1,e2) ->
	CTTMexpBinExpr(bop, translate_mexpr e1, translate_mexpr e2)
    | CTTMexpCoerce(cty,e1) ->
	CTTMexpCoerce(translate_c_type_reg cty, translate_mexpr e1)
    | CTTMexpUnaryExpr(uop,e1) ->
	CTTMexpUnaryExpr(uop, translate_mexpr e1)
    | CTTMexpInvoke(lv,args) ->
	let lv = translate_mem_object lv in
	let args = Util.list_map (fun x -> coerce_to_unsigned (translate_mexpr x)) args in
	coerce_desc_to_signed (CTTMexpInvoke(lv,args)) exp.mexpr_type
    | CTTMexpConstant(const) ->
	CTTMexpConstant(const)
  in
  make_mexpr new_exp (translate_c_type_reg exp.mexpr_type)

and translate_mem_object = function
    CTTMlvPtr e1 -> CTTMlvPtr (translate_mexpr e1)
  | CTTMlvVar(id,cty) -> CTTMlvVar (id, translate_c_type_mem cty)
  | CTTMlvInvoke(lv,args,cty) ->
      let lv = translate_mem_object lv in
      let args = Util.list_map (fun x -> coerce_to_unsigned (translate_mexpr x)) args in
      CTTMlvInvoke(lv,args,translate_c_type_mem cty)

and translate_fields =
  Util.list_map (fun (x,y) -> (x : identifier), translate_c_type_mem y)
    
let rec translate_cttm_initializer = function
    CTTMinitExp e1 ->
      CTTMinitExp (coerce_to_unsigned (translate_mexpr e1))
  | CTTMinitList l ->
      CTTMinitList (Util.list_map translate_cttm_initializer l)

(*let rec translate_statement stmt =
  let ndesc = 
    match stmt.mstmt_t with
      CTTMstmtNull -> CTTMstmtNull
    | CTTMstmtExpr e1 -> translate_mexpr e1
    | CTTMstmtLabeled(id,s1) -> CTTMstmtLabeled(id, translate_statement s1)
    | CTTMstmtCase_Labeled(l,s1) -> CTTMstmtCase_Labeled(l, translate_statement s1)
    | CTTMstmtDefault_Labeled s1 -> CTTMstmtDefault_Labeled(translate_statement s1)
    | CTTMstmtCompound(dcls,stmts) ->
	CTTMstmtCompound(Util.list_map translate_declaration dcls, Util.list_map translate_statement stmts)
    | CTTMstmtIf(e1, s1, s2) ->
	CTTMstmtIf(translate_mexpr e1, translate_statement s1, Option.map translate_statement s2)
    | CTTMstmtSwitch(e1, s1) ->
	CTTMstmtSwitch(translate e1, translate_statement s1)
    | CTTMstmtWhile(e1, s1) ->
	CTTMstmtWhile(translate e1, translate_statement s1)
    | CTTMstmtDoWhile(s1, e1) ->
	CTTMstmtDoWhile(translate_statement s1, translate e1)
    | CTTMstmtFor(e1,e2,e3,s1) ->
	CTTMstmtFor(Option.map translate e1,
		    Option.map translate e2,
		    Option.map translate e3,
		    translate_statement s1)
    | CTTMstmtGoto(id) -> CTTMstmtGoto(id)
    | CTTMstmtContinue -> CTTMstmtContinue
    | CTTMstmtBreak -> CTTMstmtBreak
    | CTTMstmtReturn e1 ->
	CTTMstmtReturn(Option.map translate e1)
  in
  { stmt with mstmt_t = ndesc }
*)
