(* 
   Part of Fail-Safe C Compiler. Produced by Yutaka Oiwa.
   (c) 2001-2005 Yutaka Oiwa.
   (c) 2005-     AIST.

   This file is written by Yutaka Oiwa in 2006. *)

open Printf
open Util
open Il2

let strof_tempid = Separate_side_effect.strof_tempid

let strof_fields = Separate_side_effect.strof_fields

let strof_c_type = Ctt_printer.strof_c_type

let strof_binop = Separate_side_effect.strof_binary_operator

let strof_unaryop = Ctt_printer.strof_unaryop

let strof_il2_vartype = function
    GlobalVar -> "global"
  | HeapVar -> "localH"
  | StackVar -> "localS"
  | RegVar -> "localR"

let strof_il2_lv lv fields = 
  let fields = strof_fields "." fields in
  match lv with
    IL2lvPtr(id) -> "*" ^ strof_tempid id ^ fields
  | IL2lvVar(k, id, typ) ->
      id ^ " : " ^ strof_il2_vartype k ^ " "
      ^ strof_c_type typ ^ fields
  | IL2lvTemp(id) -> strof_tempid id ^ fields

let strof_il2_expr = function
  | IL2expCoerce(ct, t) -> 
      sprintf "(%s)%s" (strof_c_type ct) (strof_tempid t)
  | IL2expConstant c -> Ctt_printer.strof_ctt_constant c
  | IL2expUndefined -> "(undefined)"
  | IL2expBinop(b, t1, t2) ->
      sprintf "%s %s %s" (strof_tempid t1) (strof_binop b) (strof_tempid t2)
  | IL2expUnaryop(u, t1) ->
      sprintf "%s %s" (Ctt_printer.strof_unaryop u) (strof_tempid t1)
  | IL2expInvoke(lv, args) ->
      sprintf "%s(%s)" (strof_il2_lv lv [])
	(Util.string_of_list strof_tempid "," args)
  | IL2expAddress(lv, flds) ->
      "&" ^ strof_il2_lv lv flds
  | IL2expArgument a ->
      sprintf "$arg%d" a
  | IL2expIdent t -> strof_tempid t

let rec strof_il2_cexp c =
  let desc = match c.il2_cexp_t with
    IL2cexpCoerce(ct, c1) -> sprintf "(%s)%s" (strof_c_type ct) (strof_il2_cexp c1)
  | IL2cexpConstant c -> Ctt_printer.strof_ctt_constant c
  | IL2cexpBinop(b, c1, c2) ->
      sprintf "%s %s %s" (strof_il2_cexp c1) (strof_binop b) (strof_il2_cexp c2)
  | IL2cexpUnaryop(u, c1) ->
      sprintf "%s %s" (strof_unaryop u) (strof_il2_cexp c1)
  | IL2cexpAddress(lv, flds) -> 
      "&" ^ strof_il2_lv lv flds
  | IL2cexpAddFieldOfs(c1, flds) ->
      sprintf "&(%s->%s)" (strof_il2_cexp c1) (strof_fields "." flds)
  in
  desc

let rec strof_il2_initializer = function
    IL2initConstantExp(t, c) -> sprintf "%s : %s" (strof_il2_cexp c) (strof_c_type t)
  | IL2initPostFill -> "(postfilled)"
  | IL2initStruct(t, l) ->
      sprintf "{ %s }: %s"
	(string_of_list (fun (n, i) -> sprintf ".%s = %s" n (strof_il2_initializer i)) ", " l)
	(strof_c_type t)
  | IL2initArray(t, l) ->
      sprintf "{ %s }: %s"
	(string_of_list strof_il2_initializer ", " l)
	(strof_c_type t)

let strof_switch_label = function
    CASE t -> sprintf "CASE %s" (Big_int.string_of_big_int t) 
  | DEFAULT -> "DEFAULT"

let rec strof_il2_stmt ind stmt = 
  let desc = match stmt.il2_t with
  | IL2stmtDeclAutoScalar(ls, ct, id, None)
  | IL2stmtDeclBulk(ls, ct, id, None) ->
      sprintf "%sDECLARE %s %s : %s;"
	ind (strof_il2_vartype ls) id (strof_c_type ct)
  | IL2stmtDeclAutoScalar(ls, ct, id, Some t) ->
      sprintf "%sDECLARE %s %s : %s := %s;"
	ind (strof_il2_vartype ls) id (strof_c_type ct) (strof_tempid t)
  | IL2stmtDeclBulk(ls, ct, id, Some i) ->
      sprintf "%sDECLARE %s %s : %s := %s;"
	ind (strof_il2_vartype ls) id (strof_c_type ct) (strof_il2_initializer i)
  | IL2stmtIf(ift, t, b) ->
      sprintf "%sIF %s%s THEN GOTO %d"
	ind (if ift = IFNOT then "!" else "") (strof_tempid t) b
  | IL2stmtSwitch(tid, l) ->
      sprintf "%sSWITCH %s WITH\n%s" ind (strof_tempid tid)
	(string_of_list 
	   (fun (l, b) -> sprintf "%s    %s => GOTO %d" ind (strof_switch_label l) b) "\n" l)
  | IL2stmtGoto b ->
      sprintf "%sGOTO %d" ind b
  | IL2stmtReturn None ->
      sprintf "%sRETURN" ind
  | IL2stmtReturn (Some t) ->
      sprintf "%sRETURN %s" ind (strof_tempid t)
  | IL2stmtAbort Il.ILabortNotReached ->
      sprintf "%sABORT NOTREACHED" ind
  | IL2stmtAbort (Il.ILabortOthers s) ->
      sprintf "%sABORT (%s)" ind s
  | IL2stmtAssign(t, ct, ex) ->
      sprintf "%s%s : %s := %s" ind (strof_tempid t) (strof_c_type ct) (strof_il2_expr ex)
  | IL2stmtRead(t, ct, lv, f) ->
      sprintf "%s%s : %s := READ %s" ind (strof_tempid t) (strof_c_type ct) (strof_il2_lv lv f)
  | IL2stmtWrite(lv, f, t) ->
      sprintf "%sWRITE %s = %s" ind (strof_il2_lv lv f) (strof_tempid t)
  | IL2stmtSequence s ->
      sprintf "%sSEQUENCE:\n%s\n%sEND"
	ind 
	(string_of_list (strof_il2_stmt (ind ^ "   ")) "\n" s)
	ind
  | IL2stmtParallel s ->
      sprintf "%sPARALLEL:\n%s\n%sEND"
	ind
	(string_of_list (strof_il2_stmt (ind ^ " | ")) "\n" s)
	ind
  in
  desc

let strof_il2_stmt = strof_il2_stmt ""

let strof_il2_funcbody a = 
  String.concat "\n\n"
    (Array.to_list
       (Array.mapi
	  (fun i b ->
	    sprintf "block %d: \n%s" i
	      (string_of_list strof_il2_stmt "\n" b.code)) a))

let strof_il2_program = 
  string_of_list
    (function
	IL2declFunction(gs, ct, id, ids, fbody, _attr) ->
	  sprintf "=== FUNCTION %s %s(%s): %s\n%s"
	    (Ctt_printer.strof_global_storage_class gs)
	    id
	    (String.concat ", " ids)
	    (strof_c_type ct)
	    (strof_il2_funcbody fbody)
      | IL2declVariable(gs, ct, id, None) ->
	  sprintf "=== VARIABLE %s %s: %s\n"
	    (Ctt_printer.strof_global_storage_class gs)
	    id
	    (strof_c_type ct)
      | IL2declVariable(gs, ct, id, Some i) ->
	  sprintf "=== VARIABLE %s %s: %s = \n    %s\n"
	    (Ctt_printer.strof_global_storage_class gs)
	    id
	    (strof_c_type ct)
	    (strof_il2_initializer i))
    "\n\n"
