(* 
   Part of Fail-Safe C Compiler. Produced by Yutaka Oiwa.
   (c) 2001-2005 Yutaka Oiwa.
   (c) 2005-2007 AIST.

   This file is written by Yutaka Oiwa in 2003-2006. *)

open Util
open Locterm
open Printf
open Big_int
open Il
open Il0
open Ctt_abstree
open Cttm_abstree

let dprintf l f = Debug.dprintf ~category:110 l f
let get_debug_flags () = Debug.get_debug_flags ~category:110

type constant_info = 
  | ConstUndef
  | Const of c_constants
  | Id of temp_id
  | Arg of int

type translate_environment = 
    {
     func_name : string;
     write_count : int earray;
     nonexp_read_count : int earray;
     (* use-count not from ILstmtDefTemp *)
     root : Int_set.t;
     typ : c_type earray;
     name : string earray;
     irreducible : bool earray;
     depends : Int_set.t earray;
     def_instr : il0 earray;
     const : constant_info earray;
     mexpr : Cttm_abstree.mexpr_desc earray;
     argnames : string list;
     argtypes : c_type list;
     additional_decls : (local_storage_class * c_type * identifier * ctt_initializer option) Glist.t
   }

let make_mexpr ty e = 
  { mexpr_type = ty; mexpr_t = e }

let type_of_id ~env id = 
  Earray.get env.typ id

let type_func_void_void =
  make_c_type (Tfunction([type_void], false, type_void))

let translated_name_of_id ~env id = 
  match Earray.get_option env.name id with
    None -> "T" ^ string_of_int id
  | Some n -> "T" ^ string_of_int id ^ "_" ^ n
					       
let lv_for_id ~env id =
  let ty = type_of_id ~env id in
  ty, CTTMlvVar(translated_name_of_id ~env id, ty)

let translate_id ~env id = 
  match Earray.get_option env.mexpr id with
    None -> 
      let ty, lv = lv_for_id ~env id in
      make_mexpr ty (CTTMexpRead(lv, []))
  | Some e ->
      let ty = type_of_id ~env id in
      dprintf 6 "%d translated to %s \n%!" id (Ctt_printer.strof_mexpr (make_mexpr ty e));
      make_mexpr ty e
	
let translate_id_ctt ~env id = 
  Cttm_to_ctt.translate_expr (translate_id ~env id)

let translate_binop = function
    ILbinTimes     -> CTTbinTimes
  | ILbinDiv       -> CTTbinDiv
  | ILbinPlusVV    -> CTTbinPlusVV
  | ILbinMinusVV   -> CTTbinMinusVV
  | ILbinPlusPV    -> CTTbinPlusPV
  | ILbinMinusPP   -> CTTbinMinusPP
  | ILbinMinusPV   -> CTTbinMinusPV
  | ILbinModulo    -> CTTbinModulo
  | ILbinLshift    -> CTTbinLshift
  | ILbinRshift    -> CTTbinRshift
  | ILbinLogAnd    -> CTTbinLogAnd
  | ILbinLogOr     -> CTTbinLogOr
  | ILbinIntAnd    -> CTTbinIntAnd
  | ILbinIntOr     -> CTTbinIntOr
  | ILbinIntXor    -> CTTbinIntXor
  | ILbinLessThan  -> CTTbinLessThan
  | ILbinLessEqual -> CTTbinLessEqual
  | ILbinGtrThan   -> CTTbinGtrThan
  | ILbinGtrEqual  -> CTTbinGtrEqual
  | ILbinEqual     -> CTTbinEqual
  | ILbinNotEqual  -> CTTbinNotEqual

let translate_lv ~env = function
    ILlvPtr(tid) ->
      CTTMlvPtr(translate_id ~env tid)
  | ILlvVar(id, typ) ->
      CTTMlvVar(id, typ)
  | ILlvTemp(tid) ->
      CTTMlvRvalue(translate_id ~env tid)

let translate_arg ~env i = 
  let name = List.nth env.argnames i
  and ty = List.nth env.argtypes i
  in
  CTTMexpRead(CTTMlvVar(name, ty),[])
  
let translate_exp_to_cttm ~env ile = 
  let id = translate_id ~env in
  match ile with
  | ILexpCoerce(t, t1) ->
      CTTMexpCoerce(t, id t1)
  | ILexpUndefined ->
      CTTMexpConstant(CTTconstInteger zero_big_int)
  | ILexpBinop(bop, t1, t2) ->
      CTTMexpBinExpr(translate_binop bop, id t1, id t2)
  | ILexpUnaryop(uop, t1) ->
      CTTMexpUnaryExpr(uop, id t1)
  | ILexpInvoke(lv, tids) ->
      CTTMexpInvoke(translate_lv ~env lv, list_map id tids)
  | ILexpAddress(lv, fields) ->
      CTTMexpAddress(translate_lv ~env lv, fields)
  | ILexpConstant(const) ->
      CTTMexpConstant(const)
  | ILexpArgument(i) ->
      translate_arg ~env i
  | ILexpIdent(t1) ->
      (id t1).mexpr_t

let translate_exp ~env ~typ e = 
  let desc = translate_exp_to_cttm ~env e in
  Cttm_to_ctt.translate_expr (make_mexpr typ desc)

let make_stmt d = 
  Locterm.locput_dummy d
let make_expr e typ = 
  Locterm.locput_dummy { expr_t = e; expr_type = typ }

let make_branch target = 
  make_stmt (CTTstmtGoto target)

let make_assign ~env id typ exp = 
  let idv = CTTexpVar(translated_name_of_id ~env id, typ) in
  if exp.locterm_v.expr_t = idv then
    [] (* redundant move *)
  else
    let assign_exp = 
      if typ.ct_ty = Tvoid then
	exp
      else begin
	make_expr
	  (CTTexpAssign(make_expr idv typ, exp))
	  typ
      end
    in
    [ CTTstmtExpr assign_exp ]

type label_kind = 
  | Lnormal of string
  | Lcase of big_int
  | Ldefault

let make_labelled l s = 
  let hd, tl =
    match s with
      [] -> make_stmt CTTstmtNull, []
    | hd::tl -> hd, tl
  in
  let new_hd = 
    match l with
    | Lnormal i -> loccopy ~orig:hd (CTTstmtLabeled(i, hd))
    | Lcase x   -> loccopy ~orig:hd (CTTstmtCase_Labeled(x, hd))
    | Ldefault  -> loccopy ~orig:hd (CTTstmtDefault_Labeled(hd))
  in
  new_hd :: tl

let make_compound_1 l = 
  (* remove redundant { } *)
  map_flatten
    (fun stmt ->
      match stmt.locterm_v with
	CTTstmtCompound([], l) -> l
      | _ -> [stmt])
    l

let make_compound_2 l =
  let rec iter = function
      [] -> []
    | hd::tl ->
	match hd.locterm_v with
	  CTTstmtLabeled(l, s) ->
	    make_labelled (Lnormal l) (iter (s :: tl))
	| CTTstmtCase_Labeled(x, s) ->
	    make_labelled (Lcase x) (iter (s :: tl))
	| CTTstmtDefault_Labeled(s) ->
	    make_labelled (Ldefault) (iter (s :: tl))
	| CTTstmtNull ->
	    iter tl
	| instr ->
	    hd :: iter tl
  in
  iter l

let make_compound body = 
  let body = make_compound_1 body in
  let body = make_compound_2 body in
  match body with
    [] -> CTTstmtNull
  | [s] -> s.locterm_v
  | body -> CTTstmtCompound([], body)

let translate_read ~env typ lv fields = 
  make_mexpr typ (CTTMexpRead (translate_lv ~env lv, fields))

let rec translate_bulkinit_partial ~env init = 
  match init with
    ILinitConstantExp _ -> failwith "i2c unused 20-2-1"
  | ILinitPostFill -> failwith "i2c panic 20-2-2" (* generated code should not have this *)
  | ILinitStruct(_, l) -> 
      CTTinitList (Util.list_map (function t, s -> translate_bulkinit_partial ~env s) l)
  | ILinitArray(_, l) ->
      CTTinitList (Util.list_map (translate_bulkinit_partial ~env) l)
  | ILinitAbstractCtt c -> CTTinitExp (Cttm_to_ctt.translate_expr c)

let rec translate ~env s = 
  let desc = 
    match s.il0_t with
      IL0stmtLabel l -> [ CTTstmtLabeled(l, make_stmt CTTstmtNull) ]
    | IL0stmtDeclAutoScalar(lsc,cty,id,None)
    | IL0stmtDeclBulk(lsc,cty,id,None) ->
	Glist.put env.additional_decls (lsc, cty, id, None); []
    | IL0stmtDeclAutoScalar(lsc,cty,id,Some _) -> failwith "unused 20-1"
    | IL0stmtDeclBulk(lsc,cty,id, Some init) -> 
	let init = translate_bulkinit_partial ~env init in
	Glist.put env.additional_decls (lsc, cty, id, Some init); []
    | IL0stmtIf(iftype,var,target) ->
	let expr = translate_id_ctt ~env var in
	let expr = 
	  match iftype with
	    IFTRUE -> expr
	  | IFNOT -> make_expr (CTTexpUnaryExpr(LogNot, expr)) type_boolean
	in
	[ CTTstmtIf(expr, make_branch target, None) ]
    | IL0stmtSwitch(var, targets) ->
	let expr = translate_id_ctt ~env var in
	let gotos = Util.list_map
	    (fun (l, t) ->
	      let stmt = make_branch t in
	      match l with
		CASE i -> make_stmt (CTTstmtCase_Labeled(i, stmt))
	      | DEFAULT -> make_stmt (CTTstmtDefault_Labeled stmt)) targets
	in
	let body = make_stmt (CTTstmtCompound ([], gotos))
	in
	[ CTTstmtSwitch (translate_id_ctt ~env var, body) ]
    | IL0stmtGoto(s) ->
	[ CTTstmtGoto s ]
    | IL0stmtReturn vopt ->
	[ CTTstmtReturn (Option.map (translate_id_ctt ~env) vopt) ]
    | IL0stmtAbort _ ->
	[ CTTstmtExpr
	    (make_expr
	       (CTTexpInvoke
		  (make_expr (CTTexpVar ("abort", type_func_void_void)) type_func_void_void,
		   []))
	       type_void) ]
    | IL0stmtDefTemp(tid, typ, exp) ->
	if Earray.get env.irreducible tid
	then begin
	  let exp = 
	    match Earray.get_option env.mexpr tid with
	      None -> 
		dprintf 6 "translating deftmp(%d), irreducible, no pc\n" tid;
		translate_exp ~env ~typ exp
	    | Some e ->
		dprintf 6 "translating deftmp(%d), irreducible, have pc\n" tid;
		Cttm_to_ctt.translate_expr (make_mexpr typ e)
	  in
	  dprintf 6 "translating deftmp(%d), reducible\n" tid;
	  make_assign ~env tid typ exp
	end else []
    | IL0stmtReadToTemp(tid, typ, lv, fields) ->
	if Earray.get env.irreducible tid
	then begin
	  let mexp = translate_read ~env typ lv fields in
	  let exp = Cttm_to_ctt.translate_expr mexp in
	  make_assign ~env tid typ exp
	end else []
    | IL0stmtWrite(lv, fields, tid) ->
	let rhs = translate_id ~env tid in
	let mexp = make_mexpr (type_of_id ~env tid) 
	    (CTTMexpWrite (translate_lv ~env lv, fields, None, rhs)) in
	let exp = Cttm_to_ctt.translate_expr mexp in
	[ CTTstmtExpr exp ]
(*    | IL0stmtInitAssign(lv, assignments) ->
	let lv = translate_lv ~env lv in
	let f (fields, rhs) =
	  failwith "unimp252initassign";
	in
        list_map f assignments*)
    | IL0stmtSequence(ils) 
    | IL0stmtParallel(ils) ->
	[ make_compound (Util.map_flatten (translate ~env) ils) ]
  in
  list_map make_stmt desc

let step1 ~name old_env = 
  let name_env =
    Earray.empty () in
  Array.iteri
    (fun i t -> match t with
      Some n -> Earray.set name_env i n
    | None -> ()) old_env;
  name_env
  
let step2 ~name args argtypes nameenv body =
  dprintf 4 "step2 for %s\n" name;
  (* gather types, set_count for temporary identifiers *)
  let root = ref (Int_set.empty) in
  let wcount = Earray.empty_with_default 0 in
  let ne_rcount = Earray.empty_with_default 0 in
  let types = Earray.empty () in
  let irreducible = Earray.empty_with_default false in
  let depends = Earray.empty_with_default Int_set.empty in
  let def_instr = Earray.empty () in
  let const = Earray.empty () in
  let set_depend tid ids = 
    Earray.set depends tid 
      (Int_set.union (Earray.get depends tid) ids)
  in
  let register_root tid = 
    dprintf 6 "  register %d as root" tid;
    root := Int_set.union !root (Int_set.singleton tid)
  in
  let set_irreducible tid = 
    dprintf 6 "  register %d as irreducible" tid;
    Earray.set irreducible tid true;
    register_root tid
  in
  let wset id typ instr = 
    Earray.incr wcount id;
    match Earray.get_option types id with
      None -> 
	Earray.set types id typ;
	Earray.set def_instr id instr;
    | Some t' ->
	if not (C_typing.equal_type typ t') then
	  failwith (sprintf "panic: type mismatch %d: %s %s" id 
		      (Ctt_printer.strof_c_type typ) (Ctt_printer.strof_c_type t'))
	else
	  Earray.unset def_instr id
  in
  let rset id = ()
  in
  let ne_rset id =
    dprintf 6 "  record %d as ne_rset" id;
    rset id;
    Earray.incr ne_rcount id;
    register_root id
  in
  let loop_lv f = function
      ILlvPtr(id) -> f id
    | ILlvVar _ -> ()
    | ILlvTemp(id) -> f id
  in
  let loop_exp target = function
      ILexpIdent(id) | ILexpCoerce(_,id) | ILexpUnaryop(_,id) ->
	rset id
    | ILexpBinop(_binop,id1,id2) -> rset id1; rset id2
	  (* Note: Gcc currently optimizing out unused divisions,
	     which might cause SIGFLT. FSC also follows this semantics.
	     if we think it is unacceptable later,
	     make target irreducible here, checking _binop.
	   *)
    | ILexpArgument _ | ILexpConstant _ | ILexpUndefined -> ()
    | ILexpInvoke(lv,ids) ->
	loop_lv rset lv;
	List.iter rset ids;
	begin
	  match lv with
	    ILlvVar(_,{ ct_const_p = true; ct_ty = Tfunction _ }) -> ()
	  | _ -> set_irreducible target
	end
    | ILexpAddress(lv,_) ->
	begin
	  match lv with
	    ILlvVar(n, t) when List.mem n args ->
	      assert false (* we assume arguments appears only as ILexpArgument *)
	  | _ -> ()
	end;
	loop_lv rset lv
  in
  let rec loop s =
    match s.il0_t with
      IL0stmtLabel l -> ()
    | IL0stmtDeclAutoScalar (lsc,cty,id,initopt)-> ()
    | IL0stmtDeclBulk (lsc,cty,id,initopt) -> ()
    | IL0stmtIf(iftype,tid,target) -> 
	ne_rset tid
    | IL0stmtSwitch(tid, targets) ->
	ne_rset tid
    | IL0stmtGoto(s) -> ()
    | IL0stmtReturn vopt -> 
	Option.iter ne_rset vopt
    | IL0stmtAbort _ -> ()
    | IL0stmtDefTemp(tid, typ, exp) ->
	set_depend tid s.il0_depends;
	wset tid typ s;
	loop_exp tid exp
    | IL0stmtReadToTemp(tid, typ, lv, fields) ->
	begin
	  match lv with
	    ILlvVar(n, t) when List.mem n args ->
	      assert false  (* we assume arguments appears only as ILexpArgument *)
          | ILlvVar _ ->
              (* TODO: reading variable is not movable, but reducible. *)
	      set_irreducible tid;
	      loop_lv ne_rset lv;
	  | _ ->
	      set_irreducible tid;
	      loop_lv ne_rset lv;
	end;
	wset tid typ s;
    | IL0stmtWrite(lv, fields, tid) ->
	begin
	  match lv with
	    ILlvVar(n, t) when List.mem n args ->
	      assert false  (* we assume arguments appears only as ILexpArgument *)
	  | _ -> ()
	end;
	loop_lv ne_rset lv;
	ne_rset tid
(*    | IL0stmtInitAssign(lv, inits) ->
	loop_lv ne_rset lv*)
    | IL0stmtSequence(ils) 
    | IL0stmtParallel(ils) -> (* TODO *)
	List.iter loop ils
  in
  loop body;
    {
     func_name = name;
     write_count = wcount;
     nonexp_read_count = ne_rcount;
     typ = types;
     name = nameenv;
     irreducible = irreducible;
     depends = depends;
     def_instr = def_instr;
     const = const;
     root = !root;
     argnames = args;
     argtypes = argtypes;
     mexpr = Earray.empty ();
     additional_decls = Glist.empty ()
   }

let optimize1 ~env body = 
  (* peep-hole optimization (1) *)
  (* replace "cons(car x, cdr x)" with "x" *)
  dprintf 3 "optimize1 for %s\n" env.func_name;
  let reduce cons car cdr = 
    Earray.iteri
      (fun i -> function
	  { il0_t = IL0stmtDefTemp(_,t,ILexpInvoke(ILlvVar(f0,_),[v1;v2]))} when f0 = cons
	  -> begin try
	    (* dprintf 6 "found %s at %d (%d,%d)\n" cons i v1 v2; *)
	    match Earray.get env.def_instr v1, Earray.get env.def_instr v2 with
	      { il0_t = IL0stmtDefTemp(_,_,ILexpInvoke(ILlvVar(f1,_),[v01]))},
	      { il0_t = IL0stmtDefTemp(_,_,ILexpInvoke(ILlvVar(f2,_),[v02]))}
	      when f1 = car && f2 = cdr
              ->
		(* dprintf 6 "found %s,%s (%d,%d)\n" car cdr v01 v02; *)
		if
		  v01 = v02
		    && Earray.get env.write_count i = 1
		    && Earray.get env.write_count v1 = 1
		    && Earray.get env.write_count v2 = 1
		then begin
		  (* dprintf 6 "replace %d\n" i; *)
		  Earray.set env.def_instr i
		    (make_il0
		       (IL0stmtDefTemp(i,t,ILexpIdent(v01))))
		end
	    | _ -> ()
	  with
	    Not_found -> ()
	  end
	| _ -> (* dprintf 6 "skip %d\n" i *) ()
      ) env.def_instr
  in
  reduce "value_of_base_vaddr" "base_of_value" "vaddr_of_value";
  reduce "dvalue_of_base_vaddr" "base_of_dvalue" "vaddr_of_dvalue"

let step3 ~env body = 
  dprintf 4 "step3 for %s\n" env.func_name;
  (* constant propagation (1) *)
  (* const and ident are propargated for all single-assignments *)
  let traced = Earray.empty_with_default false in
  let delayed = ref Int_set.empty in
  let changed = ref true in 
  let rec trace from i = 
    if Earray.get traced i then ()
    else begin
      dprintf 6 "walk: %d -> %d\n" from i;
      Earray.set traced i true;
      match Earray.get env.write_count i with
	1 -> begin
	  if get_debug_flags () >= 6 then begin
	    eprintf "      %d is wc == 1. proceed [" i;
	    Int_set.iter (fun i -> eprintf " %d" i) (Earray.get env.depends i);
	    eprintf "]\n";
	  end;
	  Int_set.iter (trace i) (Earray.get env.depends i);

	  match (Earray.get env.def_instr i).il0_t with
	    IL0stmtDefTemp(_, _, e) -> begin
	      match e with
	      | ILexpUndefined ->
		  dprintf 6 "%d <- undef-constant\n" i;
		  Earray.set env.const i (ConstUndef);
	      | ILexpConstant(c) ->
		  dprintf 6 "%d <- constant\n" i;
		  Earray.set env.const i (Const c);
	      | ILexpArgument(index) ->
		  dprintf 6 "%d <- arg %d\n" i index;
		  Earray.set env.const i (Arg index);
	      | ILexpIdent(id2) -> begin
		  dprintf 6 "%d <- ident %d\n" i id2;
		  match Earray.get_option env.const id2 with
		    Some c ->
		      dprintf 6 "      %d is already defined\n" id2;
		      Earray.set env.const i c
		  | None -> Earray.set env.const i (Id id2);
	      end
	      | _ -> ()
	    end
	  | _ -> ()
	end
      | _ -> 
	  if get_debug_flags () > 6 then begin
	    eprintf "      %d is wc != 1. delay [" i;
	    Int_set.iter (fun i -> eprintf " %d" i) (Earray.get env.depends i);
	    eprintf "]\n";
	  end;
	  delayed := Int_set.union !delayed (Earray.get env.depends i);
    end
  in
  Int_set.iter (fun i ->trace (-1) i) env.root;
  while (!delayed <> Int_set.empty) do
    let d = !delayed in
    delayed := Int_set.empty;
    Int_set.iter (fun i -> trace (-1) i) d;
  done

let optimize2 ~env ~rcount body = 
(* special case for inlining: *)
(*      If a variable is used only once,  *)
(*      it is irreducible, *)
(*      and it is used in irreducible statements immediately after definition, *)
(*     it will be treated to be reducible. *)
  let rec f prevdef insn = 
    match insn.il0_t with
      IL0stmtIf(_,id,_) |
      IL0stmtSwitch(id,_) | 
      IL0stmtReturn(Some id) |
      IL0stmtWrite(ILlvPtr(id),_,_) ->
	if id = prevdef && Earray.get rcount id <= 1 then
	  Earray.set env.irreducible id false;
	-1
    | IL0stmtDefTemp(id,_,_) -> id
    | IL0stmtReadToTemp(id,_,ILlvPtr(sid),_) ->
	if sid = prevdef && Earray.get rcount sid <= 1 then
	  Earray.set env.irreducible sid false;
	id
    | IL0stmtReadToTemp(id,_,_,_) -> id
    | IL0stmtWrite _ 
    | IL0stmtLabel _
    | IL0stmtDeclAutoScalar _ 
    | IL0stmtDeclBulk _
    | IL0stmtGoto _
    | IL0stmtReturn None
    | IL0stmtAbort _
	-> -1
    | IL0stmtParallel l (* todo *)
    | IL0stmtSequence l ->
	iter_list prevdef l
  and
      iter_list prevdef l = List.fold_left f prevdef l
  in
  ignore (f (-1) body)
    
let step4 ~env body = 
  dprintf 4 "step4 for %s\n" env.func_name;
  (* transitively count read_count *)
  let read_count = 
    Earray.map (function 0 -> 0 | x -> x - 1) env.nonexp_read_count in
  let traced = Earray.empty_with_default false in
  if get_debug_flags () >= 6 then begin
    prerr_endline "Old Read Count: ";
    Earray.iteri (fun i x -> eprintf "%d=%d " i x) read_count;
    prerr_newline ()
  end;
  let rec trace from i = 
    dprintf 6 "T%d traced from T%d\n" i from;
    Earray.incr read_count i;
    if Earray.get traced i then ()
    else
      match Earray.get_option env.const i with
	Some (ConstUndef) -> ()
      | Some (Const c) -> ()
      | Some (Arg _) -> ()
      | Some (Id id2) -> trace i id2
      | None -> begin
	  Earray.set traced i true;
	  Int_set.iter (trace i) (Earray.get env.depends i)
      end
  in
  Int_set.iter
    (fun i ->
      dprintf 6 "tracing from root %d" i;
      trace (-1) i) env.root;
  if get_debug_flags () >= 6 then begin
    prerr_endline "New Read Count: ";
    for i = 0 to Earray.length traced do
      eprintf "%d=%d " i (Earray.get read_count i)
    done;
    prerr_newline ()
  end;
  read_count

let step5 ~env read_count body = 
  dprintf 4 "step5 for %s\n" env.func_name;
  (* constant propergation (2) *)
  let traced = Earray.empty_with_default false in
  let is_unsafe =
    let unsafe_tid t = Earray.get env.write_count t >= 2 in
    let unsafe_lv = function
	ILlvPtr tid -> unsafe_tid tid
      | ILlvTemp tid -> unsafe_tid tid
      | ILlvVar _ -> false
    in
    function
      | ILexpCoerce(_,tid)
      | ILexpUnaryop(_,tid)
      | ILexpIdent tid -> unsafe_tid tid
      | ILexpBinop(_,t1,t2) -> unsafe_tid t1 || unsafe_tid t2
      | ILexpInvoke(lv,tl) -> unsafe_lv lv || List.exists unsafe_tid tl
      | ILexpAddress(lv,_) -> unsafe_lv lv
      | ILexpConstant _
      | ILexpUndefined
      | ILexpArgument _ -> false
  in
  let rec trace from i = 
    if Earray.get traced i then () else begin
      dprintf 6 "T%d traced from T%d\n" i from;
      Earray.set traced i true;
      if Earray.get env.write_count i >= 2 then begin
	dprintf 6 "T%d is wc >=2. irreducible.\n" i;
	Int_set.iter (trace i) (Earray.get env.depends i);
	Earray.set env.irreducible i true
      end
      else if
	(Earray.get read_count i >= 2 || Earray.get env.irreducible i)
      then begin
	dprintf 6 "T%d is rc >= 2 or irreducible.\n" i;
	match Earray.get_option env.const i with
	  Some (ConstUndef) -> 
	    let exp = translate_exp_to_cttm ~env (ILexpUndefined) in
	    Earray.set env.mexpr i exp
	| Some (Const c) -> 
	    let exp = translate_exp_to_cttm ~env (ILexpConstant c) in
	    Earray.set env.mexpr i exp
	| Some (Arg index) -> 
	    let exp = translate_arg ~env index in
	    Earray.set env.mexpr i exp
	| Some (Id id2) ->
	    trace i id2;
	    let exp = translate_exp_to_cttm ~env (ILexpIdent id2) in
	    Earray.set env.mexpr i exp
	| _ ->
	    dprintf 6 "T%d have no substitution.\n" i;
	    Int_set.iter (trace i) (Earray.get env.depends i);
	    Earray.set env.irreducible i true
      end
      else begin
	dprintf 6 "T%d is inlineable.\n" i;
	Int_set.iter (trace i) (Earray.get env.depends i);
	
	match (Earray.get_option env.def_instr i) with
	  Some { il0_t = IL0stmtDefTemp(_,ty,ile) } -> begin
	    dprintf 6 "T%d is defined by deftemp.\n" i;
	    if is_unsafe ile then begin
	      dprintf 5 "T%d: rhs is unsafe (wc > 2). keep it." i;
	      Earray.set env.irreducible i true
	    end else begin
	      let exp = translate_exp_to_cttm ~env ile in
	      Earray.set env.mexpr i exp
	    end
	  end
	| Some { il0_t = IL0stmtReadToTemp(_,ty,(ILlvVar _ as lv),fl) } -> begin
	    dprintf 6 "T%d is defined by readtotemp.\n" i;
	    let exp = translate_read ~env ty lv fl in
	    Earray.set env.mexpr i exp.mexpr_t
	  end
	| _ -> 
	    dprintf 6 "T%d is defined by otherones.\n" i;
	    Int_set.iter (trace i) (Earray.get env.depends i);
	    Earray.set env.irreducible i true
      end
    end
  in
  Int_set.iter (fun i ->trace (-1) i) env.root;
  ()

let finalize_step1 ~env = 
  let decls = env.additional_decls in
  (* generate declarations for temp ids *)
  Earray.iteri
    (fun id typp -> 
      match typp with
      | { ct_ty = Tvoid } -> ()
      | t ->
	  if Earray.get env.irreducible id then begin
	    let name = translated_name_of_id ~env id in
	    (* local_storage_class * c_type * identifier * ctt_initializer option *)
	    let decl = Auto, t, name, None in
	    Glist.put decls decl
	  end
    )
    env.typ;
  Glist.to_list decls

let translate_funcbody name ct args old_env body = 
  let argtypes =
    match ct.ct_ty with
      Tfunction(at,_,_) -> at
    | _ -> assert false
  in
  let env = step1 ~name old_env in
  let env = step2 ~name args argtypes env body in
  optimize1 env body;
  step3 env body;
  let read_count = step4 ~env body in
  optimize2 ~env ~rcount:read_count body;
  step5 ~env read_count body;
  dprintf 4 "main translation for %s\n" name;
  let new_body = translate ~env body in
  let new_decl = finalize_step1 ~env in
  let new_body = 
    match new_body with
      [{ locterm_v = CTTstmtCompound([],b) } as o] ->
	loccopy ~orig:o (CTTstmtCompound(new_decl, b))
    | _ ->
	make_stmt (CTTstmtCompound(new_decl, new_body))
  in
  new_body

let translate_initializer init =
  (* Translate generates CTTMinitializer instead of ILinitializer. *)
  Cttm_to_ctt.translate_initializer init

let f : _ -> Ctt_abstree.program =
  list_map
    (function
	IL0declFunction(gs, ct, id, args, f) ->
	  CTTdeclFunction(gs,ct,id,args,translate_funcbody id ct args [| |] f.il0_funcbody)
      | IL0declVariable (gs,ct,id,init) ->
	  CTTdeclVariable(gs,ct,id,
			  Option.map translate_initializer init))
