(* Part of Fail-Safe C Compiler. (c) 2002-2004 Yutaka Oiwa. *)

open Util
open Il
open Ils
open Ilc
open Ilc0

let dprintf l f = Debug.dprintf ~category:98 l f
let get_debug_flags () = Debug.get_debug_flags ~category:98

type inserted_phi_assignments = 
    (int * int) list

type block_info = 
    {
     mutable trampoline_blocks : (int * inserted_phi_assignments) list;
     mutable phi_assignments_at_bottom : inserted_phi_assignments;
     mutable phi_assignments_at_bottom_target : int option;
   }

type environment = {
     min_additional_variables : int;
     mutable max_additonal_variables : int;
     mutable additional_variables : ilr_type il_variable_attribute_base Glist.t;
     f : ilc_function
   }

let label_direct x = "B" ^ string_of_int x
let label_trampoline from target = "T" ^ string_of_int from ^ "_" ^ string_of_int target
let attr_nocheck = Insert_check.attr_nocheck

let phase1 f = 
  dprintf 5 "phase 1.";
  let n_blocks = Array.length f.Il.body in
  let env = {
	  min_additional_variables = f.Il.max_variable_number;
	  max_additonal_variables = f.Il.max_variable_number;
	  additional_variables = Glist.empty ();
	  f = f
  } in
  let info = Array.init n_blocks
      (fun _ -> 
	{ trampoline_blocks = [];
	  phi_assignments_at_bottom = [];
	  phi_assignments_at_bottom_target = None;
	};
      ) in
  for i = 0 to n_blocks - 1 do
    let b = f.Il.body.(i) in
    let bi = info.(i) in
    if b.phi_function = [] then
      dprintf 6 "block %d: no phi." i
    else begin
      let n_pred = List.length b.predecessor in
      dprintf 6 "block %d: n_pred = %d" i n_pred;
      for j = 0 to n_pred - 1 do
	let pred = List.nth b.predecessor j in
	let phi_assignments = 
	  list_map 
	    (fun (target, source_list) -> target, source_list.(j))
	    b.phi_function
	in
	if pred <> i - 1 && List.length (f.Il.body.(pred).successor) >= 2 then begin
	  (* branch edge is critical: insert trampoline block *)
	  dprintf 6 "    %d -> %d needs trampoline block. (pred.succ #%d)"
	    pred i (List.length (f.Il.body.(pred).successor));
	  bi.trampoline_blocks <- (pred, phi_assignments) :: bi.trampoline_blocks
	end 
	else begin
	  assert (info.(pred).phi_assignments_at_bottom = []);
	  dprintf 6 "    %d -> %d do not need trampoline block. put on %d's bottom. (pred.succ #%d)"
	    pred i pred (List.length (f.Il.body.(pred).successor));
	  info.(pred).phi_assignments_at_bottom_target <- Some i;
	  info.(pred).phi_assignments_at_bottom <- phi_assignments
	end
      done
    end
  done;
  env, info

let convert_target ~info ~cur ~last target = 
  if List.mem_assoc cur info.(target).trampoline_blocks
  then
    label_trampoline cur target
  else if 
    info.(cur).phi_assignments_at_bottom_target = Some target 
      && not last
  then
    label_trampoline cur target
  else
    label_direct target

let get_argument_name ~f i = fst f.Il.arguments.(i)
let typ_base typ = ILStypeBase typ
let typ_ofsval typ = 
  if Separate_fatpointer.is_pointer typ 
  then ILStypeOfs typ else ILStypeVal typ

let typ_temp ~f id = 
  f.Il.variable_environment.(id).variable_type

let make_ilc0 s = s

let enclose_parallel s = make_ilc0 (ILC0stmtParallel s) (* TODO *)
let enclose_sequence s = make_ilc0 (ILC0stmtSequence s) (* TODO *)

(* separate out last unconditional branches *)
let separate_last_branch l = 
  let rec iter = function
      [] -> [], []
    | [i] as l -> begin
	match i with
	  ILCstmtGoto t | ILCstmtSwitch(_,[DEFAULT, t]) ->
	    [], [ILCstmtGoto t]
	| ILCstmtReturn0 | ILCstmtReturn1 _ | ILCstmtReturn2 _
	  -> [], l
	| _ -> l, []
    end
    | h::t -> 
	let t1,t2 = iter t in
	h::t1, t2
  in
  iter l

let convert_block ~f cur ~info = 
  let rec convert_instruction ~last i = 
    let new_i = 
      match i with
	ILCstmtIf(pol, condition, target) ->
	  assert (not last);
	  ILC0stmtIf(pol, condition, convert_target ~info ~cur ~last target)
      | ILCstmtSwitch(temp_id, target_list) ->
	  assert (not last);
	  ILC0stmtSwitch
	    (temp_id, 
	     Util.list_map 
	       (fun (l,target) -> l, convert_target ~info ~cur ~last target)
	       target_list)
      | ILCstmtGoto target ->
	  ILC0stmtGoto(convert_target ~info ~cur ~last target)
      | ILCstmtReturn0 ->
	  ILC0stmtReturn0
      | ILCstmtReturn1(i) ->
	  ILC0stmtReturn1(i)
      | ILCstmtReturn2(i0,i1) ->
	  ILC0stmtReturn2(i0,i1)
(*      | ILCstmtAssign(id, ILCexpArgument i) ->
	  let name, ty = f.Il.arguments.(i) in
	  ILC0stmtReadToTemp1(id, ILSlvSVar(get_argument_name ~f i, typ_ofsval ty), [], attr_nocheck)
      | ILCstmtAssign(id, ILCexpArgumentB i) ->
	  let name, ty = f.Il.arguments.(i) in
	  ILC0stmtReadToTemp1(id, ILSlvSVar(get_argument_name ~f i ^ "_B", typ_base ty), [], attr_nocheck)
      | ILCstmtAssign(id, ILCexpArgumentV i) ->
	  let name, ty = f.Il.arguments.(i) in
	  ILC0stmtReadToTemp1(id, ILSlvSVar(get_argument_name ~f i ^ "_V", typ_ofsval ty), [], attr_nocheck) *)
      | ILCstmtAssign(id, ex) ->
	  ILC0stmtDefTemp(id, ex)
      | ILCstmtAssign2(i0, i1, ILCexpArgument _) ->
	  assert false (* ?? *)
      | ILCstmtAssign2(i0, i1, ex) ->
	  ILC0stmtDefTemp2(i0, i1, ex)
      | ILCstmtRead1(id, lv, fl, cc) ->
	  ILC0stmtReadToTemp1(id, lv, fl, cc)
      | ILCstmtRead2(i0, i1, lv, fl, cc) ->
	  ILC0stmtReadToTemp2(i0, i1, lv, fl, cc)
      | ILCstmtWrite1(lv, fl, id, cc) ->
	  ILC0stmtWrite1(lv, fl, id, cc)
      | ILCstmtWrite2(lv, fl, i0, i1, cc) ->
	  ILC0stmtWrite2(lv, fl, i0, i1, cc)
      | ILCstmtDeclScalar(vt, ct, id) ->
	  ILC0stmtDeclScalar(vt,ct,id)
      | ILCstmtDeclBulk(vt, ct, id) ->
	  ILC0stmtDeclBulk(vt,ct,id)
      | ILCstmtInitialize (vt, ct, id, init) ->
	  ILC0stmtInitialize (vt,ct,id,init)
      | ILCstmtSequence(is) ->
	  assert (not last);
	  ILC0stmtSequence(Util.list_map (convert_instruction ~last) is)
      | ILCstmtParallel(is) ->
	  assert (not last);
	  ILC0stmtParallel(Util.list_map (convert_instruction ~last) is)
      | ILCstmtAbort er ->
	  ILC0stmtAbort er
      | ILCstmtAbortIf(ex, er) ->
	  ILC0stmtAbortIf(ex, er)
    in
    make_ilc0 new_i
  in
  let body_instrs, last_branch = separate_last_branch f.Il.body.(cur).code in
  let body_instrs = Util.list_map (convert_instruction ~last:false) body_instrs in
  let last_branch = Util.list_map (convert_instruction ~last:true) last_branch in
  body_instrs, last_branch

let allocate_new_temp_variables ~env t = 
  let id = env.max_additonal_variables in
  env.max_additonal_variables <- env.max_additonal_variables + 1;
  Glist.put env.additional_variables t;
  id

let make_phi_assigns ~env l =
  dprintf 7 "make_phi_assigns: [%s]" (Util.string_of_list (fun (t, f) -> Printf.sprintf "%d<--%d" t f) ", " l);
  (* detect loop *)
  if l = [] then [] else 
  let s = List.length l in
  let loop_top = ref Set_list.empty in
  let walked = ref Set_list.empty in
  let bottom = ref Set_list.empty in
  let rec iter root cur = 
    if Set_list.mem cur !walked then ()
    else begin
      walked := Set_list.add !walked cur;
      let b = ref false in
      List.iter 
	(function
	    (t, f) when f = cur ->
	      b := true;
	      if t = root then begin (* loop found *)
		loop_top := Set_list.add !loop_top root;
		bottom := Set_list.add !bottom root
	      end else
		iter root t
	  | _ -> ())
	l;
      if not !b then bottom := Set_list.add !bottom cur;
    end
  in
  List.iter (function (t, f) -> iter f f) l;

  let loop_top = Set_list.to_list !loop_top in
  let bottom = Set_list.to_list !bottom in
  
  dprintf 7 "  loop tops: %s" (Util.string_of_list string_of_int ", " loop_top);
  dprintf 7 "  dependency bottoms: %s" (Util.string_of_list string_of_int ", " bottom);

  let temp_map = 
    list_map
      (fun v ->
	let tmp = allocate_new_temp_variables ~env (env.f.Il.variable_environment.(v)) in
	(v, tmp)) loop_top
  in
  let assignment_chains = 
    let dones = ref Set_list.empty in
    let result = ref [] in
    List.iter
      (fun bottom ->
	let g = Glist.empty () in
	let rec iter cur = 
	  if List.mem_assoc cur l then
	    let from = List.assoc cur l in
	    let from = 
	      if List.mem_assoc from temp_map then List.assoc from temp_map
	      else from
	    in
	    Glist.put g (cur, from);
	    dones := Set_list.add !dones cur;
	    iter from
	in
	iter bottom;
	result := (Glist.to_list g) :: !result)
      bottom;
    [map_flatten (fun (x, y) -> [y, x; y, x]) temp_map; List.flatten (!result)]
      (* TODO: a tentative DIRTY TRICK to avoid reducing this temporary variable *)
  in
  dprintf 7 "  result_chain: [%s]" 
    (Util.string_of_list
       (fun l -> (Util.string_of_list (fun (t, f) -> Printf.sprintf "%d<--%d" t f) ", " l))
       " | " assignment_chains);

  let rec iter = function
      [] -> []
    | (target, from) :: tl ->
	make_ilc0 (ILC0stmtDefTemp(target, ILCexpIdent from)) :: iter tl
  in
  [enclose_sequence (list_map (fun l -> enclose_parallel (iter l)) assignment_chains)]

let phase2 ~env ~info = 
  dprintf 5 "phase 2.";
  let f = env.f in
  let n_blocks = Array.length f.Il.body in
  let new_body = Array.make n_blocks [] in
  for i = 0 to n_blocks - 1 do
    dprintf 6 "block %d." i;
    let main, last_branch = convert_block ~f ~info i in
    (* insert unconditional GOTO to skip inserted trampoline blocks *)
    let last_branch = 
      if last_branch = [] && i < n_blocks - 1 && info.(i + 1).trampoline_blocks <> []
      then
	[make_ilc0 (ILC0stmtGoto(label_direct (i + 1)))]
      else
	last_branch
    in
    begin
      match info.(i).phi_assignments_at_bottom_target, info.(i).phi_assignments_at_bottom with
	Some t, _::_ ->
	  dprintf 6 "  generating phi %d -> %d at the bottom of %d" i t i;
      | Some t, [] ->
	  dprintf 6 "  no phi variables %d -> %d at the bottom of %d" i t i;
      | None, _::_ ->
	  assert false
      | None, [] ->
	  dprintf 6 "  no bottom phi assignment at %d" i;
    end;
    let phi = make_phi_assigns ~env info.(i).phi_assignments_at_bottom in
    let make_trampoline from assigns = 
      dprintf 6 "  generating phi trampoline %d -> %d at the top of %d" from i i;
      (make_ilc0 (ILC0stmtLabel (label_trampoline from i))) ::
      make_phi_assigns ~env assigns
    in
    let trampoline_instrs = 
      let rec iter = function
	  [] -> []
	| [from, assigns] -> make_trampoline from assigns
	| (from, assigns)::tl -> 
	    make_trampoline from assigns @ (make_ilc0 (ILC0stmtGoto (label_direct i)) :: iter tl)
      in
      iter info.(i).trampoline_blocks
    in
    let main_label = [make_ilc0 (ILC0stmtLabel (label_direct i))] in
    let next_phi_label =
      match info.(i).phi_assignments_at_bottom_target with
	None -> []
      | Some t -> [make_ilc0 (ILC0stmtLabel (label_trampoline i t))]
    in
    new_body.(i) <- trampoline_instrs @ main_label @ main @ next_phi_label @ phi @ last_branch
  done;
  let l = Array.fold_right 
      (fun l r -> l @ r) new_body []
  in
  enclose_sequence l

let extend_variables ~env venv = 
  let add_vars = Glist.to_list env.additional_variables in
  assert (List.length add_vars + env.min_additional_variables = env.max_additonal_variables);
  if env.max_additonal_variables = 0 then [| |] else
  let new_env =
    Array.create env.max_additonal_variables 
      (if Array.length venv = 0 then List.hd add_vars else venv.(0)) in
  Array.blit venv 0 new_env 0 (Array.length venv);
  list_iteri
    (fun i v -> new_env.(i + env.min_additional_variables) <- v)
    add_vars;
  new_env

let translate_function
    ({ Il.body = body;
       Il.max_variable_number = max_variable_number;
       Il.variable_environment = variable_environment;
       Il.arguments = arguments } as f) = 
  let env, info = phase1 f in
  let newbody = phase2 ~env ~info in
  { body = newbody;
    max_variable_number = env.max_additonal_variables;
    variable_environment = extend_variables ~env variable_environment;
    arguments = arguments;
    function_attributes = f.Il.function_attributes
  }

let f = 
  list_map
    (function
	ILdeclVariable (gsc,t,id,init) -> 
	  ILC0declVariable(gsc,t,id,init)
      | ILdeclFunction (gsc,t,id,arg_id,func) ->
	  dprintf 1 "ilc_decompose_ssa: %s" id;
	  ILC0declFunction (gsc,t,id,arg_id,translate_function func))
