(* 
   Part of Fail-Safe C Compiler. Produced by Yutaka Oiwa.
   (c) 2001-2005 Yutaka Oiwa.
   (c) 2005-     AIST.

   This file is written by: Yutaka Oiwa. *)

let dprintf l f = Debug.dprintf ~category:2 l f
let get_debug_flags () = Debug.get_debug_flags ~category:2

let use_c99_tokenizer = ref false

let read fname =
  C_parserhelper.reset fname;
  let chan = open_in fname in
  try
    let lexbuf = Lexing.from_channel chan in
    let t = C_parser.translation_unit C_lexer.token lexbuf in
    close_in chan;
    t
  with
    e -> close_in chan; raise e
	
let read_chan chan =
  C_parserhelper.reset "";
  if !use_c99_tokenizer then C_lexer.use_c99_constants := true;
  let lexbuf = Lexing.from_channel chan in
  C_parser.translation_unit C_lexer.token lexbuf

let print_global_declarations = 
  List.iter (fun x -> print_string (Ctt_printer.strof_global_declaration x))

let systemdir = ref ""
let includedirs = ref []
let defines = ref []

type stopat = 
    Never | Preprocessor | Parse | Type | TypePP | Type_filter 
  | Il0p | Il0 | Il2 | Il2reduce | Ils | NewSSA | Ilc | CodeGen | Translated

let stop_at_alist = 
  [
   "preprocessor", Preprocessor;
   "parsing", Parse;
   "typing", Type;
   "typing-pp", TypePP;
   "typing2", Type_filter;
   "il0p", Il0p;
   "il0", Il0;
   "il2", Il2;
   "classify-local", Il2;
   "il2-reduce", Il2reduce;
   "ils", Ils;
   "new-ssa", NewSSA;
   "ilc", Ilc;
   "translated", Translated;
   "code-gen", CodeGen;
 ]

let stop_at_symbols_list = Util.list_map fst stop_at_alist

let make_cpp_cmdline fname = 
  let sysincludedir = 
    if !systemdir = "" then failwith "panic: systemdir not specified"
    else !systemdir ^ "/include"
  in
  let definition_endian =
    if Fsc_config.is_littleendian
    then "-D__failsafeC_little_endian"
    else "-D__failsafeC_big_endian" in
  let args =
    ["cpp"; "-nostdinc"; "-x" ; "c"] @
    ["-idirafter"; sysincludedir] @
    (Util.map_flatten (fun d -> ["-I"; d]) !includedirs) @
    [if !use_c99_tokenizer then "-std=c99" else "-ansi";
     "-undef";
     "-D__failsafeC";
     definition_endian] @
    !defines @
    [fname]
  in
  (String.concat " "
     (Util.map_flatten
	(fun s -> [Filename.quote s]) args))

let () = Install_fsc_extension.install ()

let put_top_include_directive ofp = 
  if !systemdir <> "" then
    let i = open_in (!systemdir ^ "/runtime/include-header") in
    try
      while true do
	output_string ofp (input_line i ^ "\n")
      done
    with
      End_of_file -> close_in i

let force_stack_allocation = ref false

let use_polymorphic_analysis = ref true

let process ~stopat ~outfile fname = 
  let cmdline = make_cpp_cmdline fname in
  dprintf 1 "preprocessing...";
  Debug.dprintf ~category:9 1 "invoking %s ..." cmdline;
  let cmdline = 
    if Sys.os_type <> "Win32" then cmdline else 
    "sh -c " ^ Filename.quote cmdline
  in
  if stopat = Preprocessor then begin
    let cmdline = 
      if outfile = "" then cmdline
      else cmdline ^ " > " ^ Filename.quote outfile 
    in
    exit (Sys.command cmdline);
  end;
  let chan = Unix.open_process_in cmdline in
  dprintf 1 "parsing...";
  let t =
    try read_chan chan with e -> close_in chan; raise e
  in

  if Unix.close_process_in chan <> Unix.WEXITED 0 then
    (dprintf 1 "preprocessor failed."; exit 1);
  
  if stopat = Parse then
    (C_pp.pp_print_program Format.std_formatter t;
     exit 0);

  dprintf 1 "typing...";
  let genv, t = C_typing.parse_global_declarations t in
  if stopat = Type then
    (List.iter (fun x -> print_string (Ctt_printer.strof_global_declaration x)) t;
     exit 0);
  if stopat = TypePP then 
    (C_pp.pp_print_program Format.std_formatter
       (Ctt_to_ptree.convert_program ~genv t);
     exit 0);

  dprintf 1 "cleaning...";
  let t = Expunge_localstatic.reduce_program ~genv t in
  let genv, t = Ctt_pad_structures.translate_program genv t in
  let genv, t = Ctt_reduce_mdarray.reduce_program ~genv t in
  let genv, t = Ctt_reduce_unions.translate_program ~genv t in
  let genv, t = Ctt_add_tentative_initializer.reduce_program ~genv t in
  let genv, t = Ctt_remove_unuseddecls.f ~genv t in
  if stopat = Type_filter then 
    (List.iter (fun x -> print_string (Ctt_printer.strof_global_declaration x)) t;
     exit 0);

  dprintf 1 "analyzing assignments...";
  let genv, t = genv, Cttm_expunge_assignment.translate_program t in
  let genv, t = genv, Separate_side_effect.translate_program_to_il0 ~genv t in
  if stopat = Il0p then (Separate_side_effect.print_il0_program ~genv t; exit 0);
  let genv, t = Il0_translate_bitfield.translate_program ~genv t in
  let genv, t = Il0_type_canonify.translate_program ~genv t in
  if stopat = Il0 then (Separate_side_effect.print_il0_program ~genv t; exit 0);
  let orig_t = t in
  let t = Flatten_statement.translate_program ~genv t in

  dprintf 1 "classifying local variables...";

  let t = Classify_local.translate_program ~genv t in
  if stopat = Il2 then (print_endline (Il2_printer.strof_il2_program t); exit 0);

  dprintf 1 "reducing local variables...";

  let t = 
    let param = if !force_stack_allocation then 
      { Install_fsc_extension.fsc_default_il2_reduction_parameter with 
	Il2_reduce_local_variable.forced_stack_alloc = true }
    else 
      Install_fsc_extension.fsc_default_il2_reduction_parameter
    in (* TODO: use function attribute to determine setjmp functions *)
    Il2_reduce_local_variable.translate_program ~genv ~param t
  in
  if stopat = Il2reduce then (print_endline (Il2_printer.strof_il2_program t); exit 0);

  dprintf 1 "making fat pointer representation...";
  let t = Separate_fatpointer.translate_program ~genv t in
  if stopat = Ils then 
    (List.iter (Ils_formatter.pp_ils_global_declaration Format.std_formatter) t; exit 0);

  if stopat = NewSSA then (Ils_ssa_translate.test t; exit 0);

  let t = Ils_ssa_translate.translate_program t in
  dprintf 1 "analyzing pointer usage (dummy)...";
  let t = Dummy_analyzer.f ~genv t in
  
  let t = 
    if !use_polymorphic_analysis then
      (dprintf 1 "analyzing polymorphic function types...";
       Ils_basetype_analysis.f ~genv t)
    else t
  in

  dprintf 1 "inserting check condition...";
  let t = Insert_check.f ~genv t in
  if stopat = Ilc then (Insert_check.dump_program ~genv t; exit 0);
  (* dprintf 1 "translating..."; *)
  let t = Ilc_decompose_ssa.f t in
  let t = Translate.f ~genv t in
  if stopat = Translated then (Separate_side_effect.print_il0_program_cttminit ~genv t; exit 0);

  dprintf 1 "generating typed C code...";
  let t = Il0_to_ctt.f t in
  dprintf 1 "generating C syntax tree...";
  let t = Ctt_to_ptree.convert_program ~genv ~emit_structs:false t in
  if stopat = CodeGen then begin
    C_pp.pp_print_program Format.std_formatter t; exit 0
  end;

  dprintf 1 "generating support functions and declarations...";
  let t = Add_support_funcs.f ~genv ~orig_t t in
  dprintf 1 "printing...";
  if outfile <> "" then begin
    let ofp = open_out outfile in
    try
      put_top_include_directive ofp;
      C_pp.pp_print_program (Format.formatter_of_out_channel ofp) t;
      close_out ofp;
    with
      e -> close_out ofp; (try Sys.remove outfile with _ -> ()); raise e
  end;
  dprintf 1 "done."

open Arg

let usagemsg = "Usage: trans [options]"

let dummy = Unit (fun () -> assert false)

let userusage = ref (fun () -> assert false)

let f_options = 
  ["no-clearflag", Clear Translate.clear_castflag_at_memop, " do not clear redundant cast flags at memory accesses";
   "trust-intermodule-struct", Unit (fun () -> Translate.compiler_mode := Translate.TrustInterModule),
   " use unsafe module link mode";
   "multi-module", Unit (fun () -> Translate.compiler_mode := Translate.MultiModule), " use FSC linker";
   "force-setjmp-safe", Unit (fun () -> force_stack_allocation := true), " force all generated code setjmp-safe";
   "no-polymorphic-analysis", Clear use_polymorphic_analysis, " disable polymorphic function analysis";
   "no-shortcut-cast", Clear Translate.use_optimized_narrow_cast, " disable short-cut cat behavior of narrow-type pointers";
   "accmeth-expand-limit",
   Int 
     (fun i -> 
       if i >= 0 then Add_support_funcs.indexed_ofsmap_expand_limit := i 
       else
	 raise 
	   (Bad "-faccmeth-expand-limit must given a non-negative integer")),
   " (not documented)\n";
   "use-alignpad", Set Add_support_funcs.use_alignpad, " (not documented)";
   "ptr-compare-signed-offset", Set Translate.use_signed_compare_for_offset, " (not documented)";
   "gnu-autoconf-workaround", Set C_typing.enable_gnu_autoconf_workaround, " (not documented)";
   "force-emit-stack-unwinding", Set Translate.force_emit_stack_unwind, " (not documented)";
   "stdlib-implementation", Unit (fun () -> Translate.compiler_mode := Translate.StdlibImplementation),
   " (not documented)";
   "use-c99-tokens", Set use_c99_tokenizer, " (not documented)";
 ]

let parse_f str = 
  let f, argl = 
    try
      let n = String.index str '=' in
      String.sub str 0 n,
      Some (String.sub str (n + 1) (String.length str - n - 1))
    with
      Not_found-> str, None
  in
  let rec iter = function
      [] -> raise (Arg.Bad ("unknown option -f" ^ f))
    | (f', p, _)::_ when f' = f -> begin
	match p, argl with
	  Unit f, None -> f ()
	| Set v, None -> v := true
	| Clear v, None -> v := false
	| (Unit _ | Set _ | Clear _), Some _ -> 
	    raise (Arg.Bad ("-f" ^ f ^ " do not take arguments"))
	| Int _, None ->
	    raise (Arg.Bad ("-f" ^ f ^ " requires an integer argument"))
	| Int ff, Some s ->
	    let i = try int_of_string s with Failure _ ->
	      raise (Arg.Bad ("-f" ^ f ^ " requires an integer argument"))
	    in ff i
	| _, _ -> failwith "unimp237"
    end
    | _::tl -> iter tl
  in
  iter f_options
      
let _ =
  let stop_at = ref Never in
  let infile = ref "" in
  let outfile = ref "" in
  if not !Sys.interactive then begin
    let argspec = Arg.align
	([
	 "-I", String (fun s -> includedirs := !includedirs @ [s]), " specify include directories";
	 "-D", String (fun s -> defines := !defines @ ["-D" ^ s]), " declare preproessor symbols";
	 "-U", String (fun s -> defines := !defines @ ["-U" ^ s]), " undeclare preproessor symbols";
(*	 "-Xfsc-extension", Set C_lexer.use_failsafec_extension, " enable Fail-Safe C extension keywords"; *)
	 "-E", Unit (fun () -> stop_at := Preprocessor), " preprocess source and print to stdout";
	 "-f", String parse_f, " specify compiler flags"]
	 @
	   List.map (function n,f,d -> "   " ^ n, f, d) f_options
	 @[
	   "", dummy, "";
	   "--stop-at", Symbol
	   (stop_at_symbols_list ,
	    (fun s -> stop_at :=
	      try List.assoc s stop_at_alist 
	      with Not_found -> raise (Bad "bad keyword for stop_at"))),
	 "";
	   "", dummy, " stop processing at given phase";
	   "--debug", String (Debug.parse_debug_flags), " set compiler-debugging options";
	   "---systemdir", String (fun s -> systemdir := s), " specify system include directory";
	   "---input-file", String ((:=) infile), " specify name for .c input file";
	   "---output-file", String ((:=) outfile), " specify name for .safe.c output file";
	   "---help-user", Unit (fun () -> !userusage ()), " (not documented)";
	 ])
    in
    let usagemsg = "Usage: trans [options]" in
    let usage s = Arg.usage argspec (s ^ usagemsg); exit 1 in
    userusage :=
      (fun () ->
	Arg.usage
	  (List.filter (fun (s,_,_) -> String.length s < 3 or String.sub s 0 3 <> "---") argspec)
	  ""; exit 1);
    Arg.parse argspec (fun s -> raise (Bad "bad usage")) usagemsg;
    if !infile = "" then usage "no input file specified: ";
    if !outfile = "" && !stop_at = Never then usage "no output file specified: ";
    process ~stopat:!stop_at ~outfile:!outfile !infile
  end
