(* 
   Part of Fail-Safe C Compiler. Produced by Yutaka Oiwa.
   (c) 2001-2005 Yutaka Oiwa.
   (c) 2005-2006 AIST.

   This file is written by Yutaka Oiwa in 2003-2006. *)

open Ctt_abstree
open Cttm_abstree
open Ils
open Ilc
open Ilc0
open Il
open Il0
open Big_int_infix
open Util
open Transutil
open Record_globalnames

include Debug.Install (struct let category = 100 end)
let formatter f = fun pp v -> Format.pp_print_string pp (f v)

let clear_castflag_at_memop = ref true

type compiler_mode = MultiModule | MultiLinker | StandAlone | TrustInterModule | StdlibImplementation

let compiler_mode = ref StandAlone

let max_local_var_size = ref (big_int_of_int 0x2000000)

let use_signed_compare_for_offset = ref false

let force_emit_stack_unwind = ref false

let use_optimized_narrow_cast = ref true

type decl_type = Known | Unknown_monomorphic | Unknown_polymorphic

type decl_environment = (identifier, decl_type) Hashtbl.t


type size_info = 
    Size_byte | Size_hword | Size_word | Size_dword
  | Size_other of big_int option

type convfuncs = (c_type * il_lvalue option) list

type environment =
    {
     mutable max_var : int;
     f : ilc0_function;
     is_polymorphic : bool;
     global_env : Ctt_abstree.environment;
     decl_env : decl_environment;
     rettype : c_type;
     arguments_b : int earray;
     arguments_v : int earray;
     arguments_n : int earray;
     prologue0 : il0 Glist.t;
     prologue1 : il0 Glist.t;
     epilogue : il0 Glist.t;
     mutable exit_code : il0 list;
     struct_is_continuous : bool earray;
     name : identifier;
   }

(* let is_variable_argument ~env id = false

let is_global_variable ~env id = 
  not (is_variable_argument ~env id) (* TODO *) *)

let constant_zero = CTTconstInteger(zero_big_int)
let constant_one = CTTconstInteger(unit_big_int)
let make_cttm_expr body t =
  { mexpr_type = t; mexpr_t = body }

let rettype_of_functype t =
  match t.ct_ty with
    Tfunction(_,_,rt) -> rt
  | _ -> assert false

let new_tempid ~env =
  let v = env.max_var in
  env.max_var <- v + 1;
  v

let new_temp_label tid l =
  Printf.sprintf "LL_%d_%d" tid l

let type_of_tempvar ~env t =
  try
    env.f.Ilc0.variable_environment.(t).variable_type
  with
    Invalid_argument("Array.get") ->
      Printf.eprintf "panic: tries to get type of var #%d (max is %d)"
	t (Array.length env.f.Ilc0.variable_environment);
      assert false

let sizeof_type_genv ~genv t =
  match Ctt_abstree.size_of_type ~genv t with
    Some s -> s
  | None ->
      Printf.eprintf "panic: tries to get size of %s"
	(Ctt_printer.strof_c_type t);
      assert false

let sizeof_type ~env t =
  sizeof_type_genv ~genv:env.global_env t

let is_numeric t = 
  match t.ct_ty with
    Tbuiltin _ -> true
  | _ -> false

let is_pointer t = 
  match t.ct_ty with
    Tpointer _ -> true
  | _ -> false

let size_is_dword s =
  eq_big_int s
    (big_int_of_int Fsc_config.sizeof_longlong)

let type_base_t = make_c_type (Tabstract "base_t")
let type_ofs_t = make_c_type (Tabstract "ofs_t")
let type_signed_ofs_t = make_c_type (Tabstract "signed_ofs_t")
let type_vaddr_t = make_c_type (Tabstract "vaddr_t")
let type_word = make_c_type (Tabstract "word")
let type_dword = make_c_type (Tabstract "dword")

let type_byte = make_c_type (Tabstract "byte")
let type_hword = make_c_type (Tabstract "hword")
let type_value = make_c_type (Tabstract "value")
let type_dvalue = make_c_type (Tabstract "dvalue")
let type_ptrvalue = make_c_type (Tabstract "ptrvalue")
let type_ptrvalue_ptr = make_c_type (Tpointer type_ptrvalue)

let type_void_ptr = make_c_type (Tpointer type_void)

let type_ptrinit = make_c_type (Tabstract "union initU_ptrvalue")
let type_valinit = make_c_type (Tabstract "union initU_value")
let type_dvalinit = make_c_type (Tabstract "union initU_dvalue")

let type_fsc_error = make_c_type (Tabstract "enum fsc_error")
let type_typeinfo_init = make_c_type (Tabstract "struct typeinfo_init")
let type_typeinfo_struct = make_c_type (Tabstract "struct typeinfo_s")
let type_function_stub_init = make_c_type (Tabstract "struct fsc_function_stub_init")
let type_function_stub = make_c_type (Tabstract "struct fsc_function_stub")
let type_genfunc = make_c_type (Tfunction ([type_base_t; type_base_t], false, type_dvalue))

let make_func_lv name argtype rettype =
  ILlvVar(name, make_c_type (Tfunction(argtype, false, rettype)))

let make_func_lv_R name argtype rettype =
  ILlvVar(name, make_c_type ~const:true (Tfunction(argtype, false, rettype)))
    (* a constant attribute for function type means reducible functions *)

let lv_cons_val =
  make_func_lv_R "value_of_base_vaddr" [type_base_t; type_vaddr_t] type_value

let lv_value_of_ptrvalue =
  make_func_lv_R "value_of_ptrvalue" [type_ptrvalue] type_value

let lv_ptrvalue_of_value =
  make_func_lv_R "ptrvalue_of_value" [type_value] type_ptrvalue

let lv_cons_dval =
  make_func_lv_R "dvalue_of_base_vaddr"
    [type_base_t; type_unsigned_long_long] type_dvalue

let lv_cons_ptrval =
  make_func_lv_R "ptrvalue_of_base_ofs" [type_base_t; type_ofs_t] type_ptrvalue

let lv_value_of_ptr = 
  make_func_lv_R "value_of_base_ofs" [type_base_t; type_ofs_t] type_value

let lv_nullexpand_value =
  make_func_lv_R "value_of_int" [type_word] type_value

let lv_nullexpand_dvalue =
  make_func_lv_R "dvalue_of_dword" [type_dword] type_dvalue

let lv_base_of_value =
  make_func_lv_R "base_of_value" [type_value] type_base_t

let lv_base_of_dvalue =
  make_func_lv_R "base_of_dvalue" [type_dvalue] type_base_t

let lv_base_of_ptrvalue =
  make_func_lv_R "base_of_ptrvalue" [type_ptrvalue] type_base_t

let lv_vaddr_of_value =
  make_func_lv_R "vaddr_of_value" [type_value] type_vaddr_t

let lv_vaddr_of_dvalue =
  make_func_lv_R "vaddr_of_dvalue" [type_dvalue] type_dword
    
let lv_ofs_of_ptrvalue =
  make_func_lv_R "ofs_of_ptrvalue" [type_ptrvalue] type_ofs_t

let lv_ofs_of_value =
  make_func_lv_R "ofs_of_value" [type_value] type_ofs_t

let lv_vaddr_of_base_ofs =
  make_func_lv_R "vaddr_of_base_ofs" [type_base_t; type_ofs_t] type_vaddr_t

let lv_ofs_of_base_vaddr =
  make_func_lv_R "ofs_of_base_vaddr" [type_base_t; type_vaddr_t] type_ofs_t

let lv_base_remove_castflag =
  make_func_lv_R "base_remove_castflag" [type_base_t] type_base_t

let lv_base_put_castflag =
  make_func_lv_R "base_put_castflag" [type_base_t] type_base_t

let lv_check_cast =
  make_func_lv_R "is_cast" [type_base_t] type_boolean

let lv_check_ofs_ok =
  make_func_lv "is_offset_ok" [type_base_t; type_ofs_t] type_boolean

let lv_read_byte =
  make_func_lv "read_byte" [type_base_t; type_ofs_t] type_byte

let lv_read_hword =
  make_func_lv "read_hword" [type_base_t; type_ofs_t] type_hword

let lv_read_word =
  make_func_lv "read_word" [type_base_t; type_ofs_t] type_value

let lv_read_dword =
  make_func_lv "read_dword" [type_base_t; type_ofs_t] type_dvalue

let lv_write_byte =
  make_func_lv "write_byte" [type_base_t; type_ofs_t; type_byte; type_typeinfo_ptr] type_void

let lv_write_hword =
  make_func_lv "write_hword" [type_base_t; type_ofs_t; type_hword; type_typeinfo_ptr] type_void

let lv_write_word =
  make_func_lv "write_word" [type_base_t; type_ofs_t; type_value; type_typeinfo_ptr] type_void

let lv_write_dword =
  make_func_lv "write_dword" [type_base_t; type_ofs_t; type_dvalue; type_typeinfo_ptr] type_void

let lv_abort =
  make_func_lv "fsc_abort" [type_base_t; type_fsc_error] type_void

let lv_alloc_varargs =
  make_func_lv "fsc_alloc_varargs" [type_size_t] type_base_t

let lv_va_end =
  make_func_lv "fsc_finish_varargs" [type_base_t] type_void
    
let lv_dealloc_varargs_finished =
  make_func_lv "fsc_dealloc_varargs_finished" [type_base_t] type_void

let lv_put_varargs =
  make_func_lv "fsc_put_varargs" [type_base_t; type_size_t; type_value] type_void

let lv_put_varargs_2 =
  make_func_lv "fsc_put_varargs_2" [type_base_t; type_size_t; type_dvalue] type_void

let lv_alloc_valtrampoline =
  make_func_lv "fsc_alloc_valtrampoline" [type_typeinfo_ptr] type_void_ptr

let type_stack_frame = make_c_type (Tabstract "struct fsc_stack_frame")

let type_frame_variables_entry = make_c_type (Tpointer (make_c_type ~volatile:true Tvoid))

let type_frame_variables = make_c_type (Tarray (type_frame_variables_entry, None))

let type_frame_variables_n n = make_c_type (Tarray (type_frame_variables_entry, Some (big_int_of_int n)))

let id_local_frame = "fsc_local_stack_frame"
let id_local_frame_variables = "fsc_local_stack_frame_variables"

let lv_local_frame_variables = ILlvVar(id_local_frame_variables, type_frame_variables)

let lv_create_stack_frame =
  make_func_lv "enter_stack_unwind_area" 
    [make_c_type (Tpointer type_stack_frame);
     make_c_type (Tpointer type_char);
     make_c_type (Tpointer type_void_ptr);
   ]
    type_void

let lv_free_stack_frame =
  make_func_lv "exit_stack_unwind_area" [make_c_type (Tpointer type_char)] type_void

let lv_dealloc_heapvar =
  make_func_lv "fsc_dealloc_heapvar" [type_base_t] type_void

let lv_double_of_dword =
  assert (Fsc_config.sizeof_double = Fsc_config.sizeof_longlong);
  make_func_lv_R "double_of_dword" [type_dword] type_double

let lv_float_of_word =
  assert (Fsc_config.sizeof_float = Fsc_config.sizeof_long);
  make_func_lv_R "float_of_word" [type_word] type_float

let lv_dword_of_double =
  make_func_lv_R "dword_of_double" [type_double] type_dword

let lv_word_of_float =
  make_func_lv_R "word_of_float" [type_float] type_word

let lv_double_of_dvalue =
  assert (Fsc_config.sizeof_double = Fsc_config.sizeof_longlong);
  make_func_lv_R "double_of_dvalue" [type_dvalue] type_double

let lv_float_of_value =
  assert (Fsc_config.sizeof_float = Fsc_config.sizeof_long);
  make_func_lv_R "float_of_value" [type_value] type_float

let lv_dvalue_of_double =
  make_func_lv_R "dvalue_of_double" [type_double] type_dvalue

let lv_value_of_float =
  make_func_lv_R "value_of_float" [type_float] type_value

let lv_get_realoffset_ptr = 
  make_func_lv_R "get_realoffset_P" [type_base_t; type_ofs_t] type_ptrvalue_ptr

let lv_get_funcptr = 
  make_func_lv_R "get_realoffset_PF" [type_base_t] type_void_ptr

let lv_invoke_genfunc = 
  make_func_lv "fsc_invoke_func_with_varargs" [type_base_t; type_base_t; type_ofs_t; type_base_t; type_ofs_t] type_dvalue

let lv_value_of_dvalue = 
  make_func_lv_R "value_of_dvalue" [type_dvalue] type_value

let lv_dvalue_of_value = 
  make_func_lv_R "dvalue_of_value" [type_value] type_dvalue

(* predefined macros and types for constant initilizers *)

let type_fsc_header =
  make_c_type (Tabstract "struct fsc_header")

let make_cttm_func_lv name argtype rettype =
  CTTMlvVar(name, make_c_type (Tfunction(argtype, false, rettype)))

let lv_cttm_macro_emit_header =
  make_cttm_func_lv "EMIT_FSC_HEADER"
    [type_typeinfo_struct; type_size_t] type_fsc_header

let encoded_name_of_btype = function
    Tchar | Tschar | Tuchar -> "c"
  | Tshort | Tushort -> "s"
  | Tint | Tuint -> "i"
  | Tlong | Tulong -> "l"
  | Tlonglong | Tulonglong -> "q"
  | Tfloat -> "f"
  | Tdouble | Tlongdouble -> "d"

let encoded_name_of_struct ~genv id = 
  let sd = get_struct_desc ~genv id in
  try
    match List.assoc "named" sd.str_extension with
      C_abstree.Estring n ->
	"Sn" ^ string_of_int(String.length n) ^ n ^ "_"
    | _ -> assert false
  with
    Not_found -> begin
      match !compiler_mode with
	TrustInterModule -> begin
	let p = get_struct_desc ~genv id in
	let (fname, line, col, _) = p.str_loc in
	Printf.sprintf "Sl_%s_%d_%d_" (String.sub (Digest.to_hex (Digest.string fname)) 0 8) line col
      end 
      | StdlibImplementation -> failwith "all structs must be explicitly named in stdlib_impl mode"
      | StandAlone | MultiModule | MultiLinker ->
	Printf.sprintf "Sh_%s_%d_" genv.module_hash id
    end

let rec encoded_name_of_type_genv ~genv ?(array_squash = false) t =
  match t.ct_ty with
    Tvoid -> "v"
  | Tbuiltin bt -> encoded_name_of_btype bt
  | Tpointer t -> "P" ^ encoded_name_of_type_genv ~genv t
  | Tfunction(at,varp,rt) -> 
      "F" ^ String.concat "" (list_map (encoded_name_of_type_genv ~genv) at)
      ^ (if varp then "V" else "") ^ "_" ^ encoded_name_of_type_genv ~genv rt
  | Tstruct(id) -> encoded_name_of_struct ~genv id
  | Tarray(t,_) when array_squash ->
      "A" ^ encoded_name_of_type_genv ~genv t ^ "_" ^ "0"
  | Tarray(t,Some size) -> 
      "A" ^ encoded_name_of_type_genv ~genv t ^ "_" ^ 
      string_of_big_int size
  | _ -> failwith ("panic translate:345: " ^ string_of_bool array_squash ^ ": " ^ Ctt_printer.strof_c_type t)

let encoded_name_of_type ~env t = 
  encoded_name_of_type_genv ~genv:env.global_env t

let encoded_typeinfo_name ~genv t = 
  "fsc_typeinfo_" ^ encoded_name_of_type_genv ~genv t

let _ =
  Ctt_to_ptree.typeinfo_converter_hook :=
    (fun ~genv ~make t -> 
      require_name (GNtypeinfo t);
      (C_abstree.PexpAddress
	 (make
	    (C_abstree.PexpField
	       (make (C_abstree.PexpVar(encoded_typeinfo_name ~genv t)),
		"val")))))

let get_storage_type_array ~genv t s = 
  require_name (GNstorage(t, s));
  let ss = match s with
    None -> "s"
  | Some s -> string_of_big_int s
  in
  make_c_type (Tabstract ("struct fsc_storage_"
			  ^ encoded_name_of_type_genv ~genv t
			  ^ "_" ^ ss))

let size_and_elemtype_of_onmemory_type ?(incomplete_ok = false) t = 
  match t.ct_ty with
    Tarray(et, Some s) -> Some s, et
  | Tarray(et, None) ->
      if incomplete_ok then None, et else assert false
  | _ -> None, t

let storage_for_typeinfo_constant ~genv t = 
  let _, elemtype = size_and_elemtype_of_onmemory_type t in
  Record_globalnames.require_name (GNtypeinfo elemtype);
  let id = encoded_typeinfo_name ~genv elemtype in
  type_typeinfo_struct, type_typeinfo_init, ILlvVar(id, type_typeinfo_init),
  ["val", type_typeinfo_struct]

let lv_set_base_castflag ~env t =
  require_name (GNbasecast t);
  assert (is_pointer t);
  make_func_lv_R ("set_base_castflag_" ^ encoded_name_of_type ~env t)
    [type_base_t] type_base_t

let lv_add_fat_pointer ~env t =
  require_name (GNaddptr t);
  assert (is_pointer t);
  make_func_lv_R ("add_fat_pointer_" ^ encoded_name_of_type ~env t)
    [type_base_t] type_base_t

let lv_alloc_hvar =
  make_func_lv "fsc_alloc_heapvar" [type_typeinfo_ptr; type_size_t] type_void_ptr

let make_compound ~env t t1 t2 =
  let temp0 = new_tempid ~env in
  let new_t, fname = 
    match t with
      ({ct_ty = Tbuiltin bt} as t') -> begin
	match bt with
	  Tdouble | Tfloat -> assert false
	| Tlonglong | Tulonglong -> type_dvalue, lv_cons_dval
	| _ -> type_value, lv_cons_val
      end
    | ({ct_ty = Tpointer pt} as t') -> begin
	type_ptrvalue, lv_cons_ptrval
    end
    | _ -> assert false
  in
  temp0, 
  [make_il0 (IL0stmtDefTemp
	       (temp0, new_t,
		ILexpInvoke(fname,[t1; t2])))]

type type_properties = {
    size : size_info;
    is_struct : bool;
    is_pointer : bool;
    is_fat : bool;
    is_floating : bool;
    is_indirect_ret : bool;
    
    packed_t : c_type;
    cdr_t : c_type;
    unpacked_types : c_type list;
    initializer_type : c_type;
    packed_inj_cdr : convfuncs;

    generic_t : c_type; (* the return type of read_****  *)
    samesized_word_t : c_type;

    packed_to_generic : il_lvalue option;
    generic_to_packed : il_lvalue option;
    generic_reader : il_lvalue;
    generic_writer : il_lvalue;
    generic_cdr_t : c_type;
    
    gen_inj_car : convfuncs;
    gen_inj_cdr : convfuncs;
  }

let parse_type_genv ~genv t = 
  match t.ct_ty with
    Tpointer _ ->
      assert (Fsc_config.sizeof_int = Fsc_config.sizeof_pointer);
      { size = Size_word;
	is_struct = false;
	is_pointer = true;
	is_fat = true;
	is_floating = false;
	is_indirect_ret = false;
	packed_t = type_ptrvalue;
	cdr_t = type_ofs_t;
	unpacked_types = [type_base_t; type_ofs_t];
	packed_inj_cdr = [type_ofs_t, Some lv_ofs_of_ptrvalue];
	initializer_type = type_ptrinit;
	generic_t = type_value;
	packed_to_generic = Some lv_value_of_ptrvalue;
	samesized_word_t = type_size_t;
	generic_to_packed = begin
	  require_name (GNbasecast t);
	  let id = "ptrvalue_of_value_" ^ encoded_name_of_type_genv ~genv t in
	  Some (make_func_lv_R id [type_value] type_ptrvalue)
	end;
	generic_cdr_t = type_vaddr_t;
	generic_reader = lv_read_word;
	generic_writer = lv_write_word;
	gen_inj_car = [type_base_t, Some lv_base_of_value];
	gen_inj_cdr = [type_ofs_t, Some lv_ofs_of_value];
      }
  | Tbuiltin bt -> begin
      let s = sizeof_type_genv ~genv t in
      assert (is_int_big_int s);
      let s = int_of_big_int s in
      let is_real = 
	match bt with
	  Tfloat | Tdouble | Tlongdouble -> true
	| _ -> false
      in
      let can_be_fat =
	s >= Fsc_config.sizeof_pointer && not is_real
      in
      let size = 
	if s = Fsc_config.sizeof_char then Size_byte
	else if s = Fsc_config.sizeof_short then Size_hword
	else if s = Fsc_config.sizeof_long then Size_word
	else if s = Fsc_config.sizeof_longlong then Size_dword
	else assert false
      in
      let genconv, genunconv, is_floating =
	match bt with
	  Tdouble -> Some lv_double_of_dvalue, Some lv_dvalue_of_double, true
	| Tfloat -> Some lv_float_of_value, Some lv_value_of_float, true
	| Tlongdouble -> assert false
	| _ -> None, None, false (* cast *)
      in
      let t = { t with ct_const_p = false } in
      match size, can_be_fat with
	Size_byte, false ->
	  { size = Size_byte;
	    is_struct = false;
	    is_pointer = false;
	    is_fat = false;
	    is_floating = is_floating;
	    is_indirect_ret = false;
	    packed_t = t;
	    cdr_t = t;
	    packed_inj_cdr = [];
	    unpacked_types = [t];
	    initializer_type = t;
	    generic_t = type_byte;
	    samesized_word_t = type_byte;
	    packed_to_generic = genunconv;
	    generic_to_packed = genconv;
	    generic_cdr_t = type_byte;
	    generic_reader = lv_read_byte;
	    generic_writer = lv_write_byte;
	    gen_inj_car = [];
	    gen_inj_cdr = [t, genconv];
	  }
      |	Size_hword, false ->
	  { size = Size_hword;
	    is_struct = false;
	    is_pointer = false;
	    is_fat = false;
	    is_floating = is_floating;
	    is_indirect_ret = false;
	    packed_t = t;
	    cdr_t = t;
	    packed_inj_cdr = [];
	    unpacked_types = [t];
	    initializer_type = t;
	    packed_to_generic = genunconv;
	    generic_to_packed = genconv;
	    generic_t = type_hword;
	    samesized_word_t = type_hword;
	    generic_cdr_t = type_hword;
	    generic_reader = lv_read_hword;
	    generic_writer = lv_write_hword;
	    gen_inj_car = [];
	    gen_inj_cdr = [t, genconv];
	}
      |	Size_word, false ->
	  assert (is_floating);
	  { size = Size_word;
	    is_struct = false;
	    is_pointer = false;
	    is_fat = false;
	    is_floating = is_floating;
	    is_indirect_ret = false;
	    packed_t = t;
	    cdr_t = t;
	    packed_inj_cdr = [];
	    unpacked_types = [t];
	    initializer_type = t;
	    packed_to_generic = genunconv;
	    generic_to_packed = genconv;
	    generic_t = type_value;
	    samesized_word_t = type_word;
	    generic_cdr_t = type_vaddr_t;
	    generic_reader = lv_read_word;
	    generic_writer = lv_write_word;
	    gen_inj_car = [];
	    gen_inj_cdr =
	    (if genconv = None then [t, genconv; type_word, Some lv_vaddr_of_value]
	    else [t, genconv]);
	}
      |	Size_word, true ->
	  assert (genconv = None);
	  { size = Size_word;
	    is_struct = false;
	    is_pointer = false;
	    is_fat = true;
	    is_floating = is_floating;
	    is_indirect_ret = false;
	    packed_t = type_value;
	    initializer_type = type_valinit;
	    packed_to_generic = genunconv;
	    generic_to_packed = genconv;
	    cdr_t = t;
	    packed_inj_cdr = [t, genconv; type_word, Some lv_vaddr_of_value];
	    unpacked_types = [type_base_t; t];
	    generic_t = type_value;
	    samesized_word_t = type_word;
	    generic_cdr_t = type_vaddr_t;
	    generic_reader = lv_read_word;
	    generic_writer = lv_write_word;
	    gen_inj_car = [type_base_t, Some lv_base_of_value];
	    gen_inj_cdr = [t, genconv; type_word, Some lv_vaddr_of_value];
	}
      |	Size_dword, false ->
	  assert (is_floating);
	  { size = Size_dword;
	    is_pointer = false;
	    is_struct = false;
	    is_fat = false;
	    is_floating = is_floating;
	    is_indirect_ret = false;
	    packed_t = t;
	    initializer_type = t;
	    packed_to_generic = genunconv;
	    generic_to_packed = genconv;
	    cdr_t = t;
	    packed_inj_cdr = [];
	    unpacked_types = [t];
	    generic_t = type_dvalue;
	    samesized_word_t = type_dword;
	    generic_cdr_t = type_dword;
	    generic_reader = lv_read_dword;
	    generic_writer = lv_write_dword;
	    gen_inj_car = [];
	    gen_inj_cdr = 
	    if genconv = None then [t, genconv; type_dword, Some lv_vaddr_of_dvalue]
	    else [t, genconv];
	  }
      |	Size_dword, true ->
	  assert (genconv = None);
	  { size = Size_dword;
	    is_struct = false;
	    is_pointer = false;
	    is_fat = true;
	    is_floating = is_floating;
	    is_indirect_ret = false;
	    packed_t = type_dvalue;
	    initializer_type = type_dvalinit;
	    cdr_t = t;
	    packed_inj_cdr = [t, genconv; type_dword, Some lv_vaddr_of_dvalue];
	    unpacked_types = [type_base_t; t];
	    generic_t = type_dvalue;
	    samesized_word_t = type_dword;
	    packed_to_generic = genunconv;
	    generic_to_packed = genconv;
	    generic_cdr_t = type_dword;
	    generic_reader = lv_read_dword;
	    generic_writer = lv_write_dword;
	    gen_inj_car = [type_base_t, Some lv_base_of_dvalue];
	    gen_inj_cdr = [t, genconv; type_dword, Some lv_vaddr_of_dvalue];
	  }
      |	_ -> assert false

  end
  | Tstruct id ->
      let sid = encoded_name_of_struct ~genv id in
      let et = make_c_type (Tabstract ("struct struct_" ^ sid)) in
      let lv_gread = make_func_lv ("read_struct_" ^ sid) [type_base_t; type_ofs_t] et in
      let lv_gwrite = make_func_lv ("write_struct_" ^ sid) [type_base_t; type_ofs_t; et] type_void in
      { size = Size_other (t.ct_size);
	is_pointer = false;
	is_struct = true;
	is_fat = false;
	is_floating = false;
	is_indirect_ret = true;
	packed_t = et;
	initializer_type = et;
	packed_inj_cdr = [];
	cdr_t = et;
	unpacked_types = [et];
	generic_t = t;
	samesized_word_t = et;
	packed_to_generic = None;
	generic_to_packed = None;
	generic_cdr_t = et;
	generic_reader = lv_gread;
	generic_writer = lv_gwrite;
	gen_inj_car = [];
	gen_inj_cdr = [];
      }
  | Tfunction _ ->
      let t = type_function_stub in
      {
       size = Size_other None;
       is_pointer = false;
       is_struct = false;
       is_fat = false;
       is_floating = false;
       is_indirect_ret = false;
       packed_t = t;
       initializer_type = t;
       packed_inj_cdr = [];
       cdr_t = t;
       unpacked_types = [t];
       generic_t = t;
       samesized_word_t = t;
       packed_to_generic = None;
       generic_to_packed = None;
       generic_cdr_t = t;
       generic_reader = lv_abort;
       generic_writer = lv_abort;
       gen_inj_car = [];
       gen_inj_cdr = [];
     }
  | _ -> failwith ("unimp 366 " ^ Ctt_printer.strof_c_type t)

let parse_type ~env =
  parse_type_genv ~genv:env.global_env

let rec is_type_continous ~env t = 
  match t.ct_ty with
    Tvoid -> false
  | Tbuiltin _ ->
      not ((parse_type ~env t).is_fat)
  | Tpointer t -> false
  | Tarray(t,_) -> is_type_continous ~env t
  | Tabstract _ -> assert false
  | Tfunction _ -> false
  | Tstruct id -> begin
      match Earray.get_option env.struct_is_continuous id with
	Some b -> b
      | None ->
	  let st = get_struct_desc ~genv:env.global_env id in
	  let b = 
	    List.for_all
	      (function
		  (_, NormalField { sf_type = t }) -> is_type_continous ~env t
		| (_, BitField _) -> assert false)
	      st.str_fields
	  in
	  Earray.set env.struct_is_continuous id b;
	  b
  end

(*let lv_specific_reader2 ~env ty =
  let p = parse_type ~env ty in
  assert (p.is_fat);
  require_name (GNreadsp2 ty);
  make_func_lv ("read_sp2_" ^ encoded_name_of_type ~env ty)
    [type_base_t; type_ofs_t] p.packed_t

let lv_specific_writer2 ~env ty =
  let p = parse_type ~env ty in
  assert (p.is_fat);
  require_name (GNwritesp2 ty);
  make_func_lv ("write_sp2_" ^ encoded_name_of_type ~env ty)
    ([type_base_t; type_ofs_t] @ p.unpacked_types) type_void *)

let lv_get_realoffset ~env ty =
  match ty.ct_ty with
    Tpointer _ -> type_ptrvalue_ptr, lv_get_realoffset_ptr
  | Tbuiltin _ | Tstruct _ ->
      let p = parse_type ~env ty in
      (*require_name (GNgetrealofs ty);*)
      let resulttype = make_c_type (Tpointer p.packed_t) in
      resulttype,
      make_func_lv_R ("get_realoffset_" ^ encoded_name_of_type ~env ty)
	[ type_base_t; type_ofs_t ]
	resulttype
  | _ -> assert false

let get_destructors ~env ttype = (* TODO: to be depreciated, but now active *)
  let p = parse_type ~env ttype in
  if p.packed_t == type_value then
    type_value, lv_base_of_value, type_vaddr_t, lv_vaddr_of_value
  else if p.packed_t == type_dvalue then
    type_dvalue, lv_base_of_dvalue, type_dword, lv_vaddr_of_dvalue
  else if p.packed_t == type_ptrvalue then
    type_ptrvalue, lv_base_of_ptrvalue, type_ofs_t, lv_ofs_of_ptrvalue
  else
    (dprintf (-1) "panic: get_destructors: %a" (formatter Ctt_printer.strof_c_type) p.packed_t; assert false)

let get_pair_type ~env t = (parse_type ~env t).packed_t
let get_base_type ~env t = type_base_t
let get_ofs_or_value_type ~env t = 
  if is_pointer t then type_ofs_t else t

let get_generic_readers ~env ttype =
  let s = sizeof_type ~env ttype in
  assert (is_int_big_int s);
  let s = int_of_big_int s in
  if s = Fsc_config.sizeof_char then
    lv_read_byte, type_byte, None
  else if s = Fsc_config.sizeof_short then
    lv_read_hword, type_hword, None
  else if s = Fsc_config.sizeof_long then
    lv_read_word, type_value,
    Some (type_word, lv_base_of_value, lv_vaddr_of_value)
  else if s = Fsc_config.sizeof_longlong then
    lv_read_dword, type_dvalue, 
    Some (type_dword, lv_base_of_dvalue, lv_vaddr_of_dvalue)
  else assert false

let get_generic_writers ~env ttype =
  let s = sizeof_type ~env ttype in
  assert (is_int_big_int s);
  let s = int_of_big_int s in
  if s = Fsc_config.sizeof_char then
    lv_write_byte, type_byte, None
  else if s = Fsc_config.sizeof_short then
    lv_write_hword, type_hword, None
  else if s = Fsc_config.sizeof_long then
    lv_write_word, type_value,
    Some (type_word, lv_nullexpand_value)
  else if s = Fsc_config.sizeof_longlong then
    lv_write_dword, type_dvalue, 
    Some (type_dword, lv_nullexpand_dvalue)
  else assert false

let check_size_of_local_variable ~genv id cty =
  match size_of_type ~genv cty with
    None -> failwith "panic 2228: allocating size_unknown local variable"
  | Some s -> 
      if s >! !max_local_var_size then
	failwith ("2229: too big local variable allocated: " ^ id ^ ": " ^ string_of_big_int s)

let make_coerce_from_vaddr typ from =
  match typ.ct_ty with
    Tbuiltin bt -> begin
      match bt with
	Tdouble ->
	  ILexpInvoke(lv_double_of_dword, [from])
      | Tfloat ->
	  ILexpInvoke(lv_float_of_word, [from])
      | _ ->
	  ILexpCoerce(typ, from)
    end
  | _ ->
      Printf.eprintf "make_coerce_from_vaddr: from %s unimplemented\n" (Ctt_printer.strof_c_type typ);
      assert false

let make_coerce_to_vaddr typ typ_to from =
  match typ.ct_ty with
    Tbuiltin bt -> begin
      match bt with
	Tdouble ->
	  ILexpInvoke(lv_dword_of_double, [from])
      | Tfloat ->
	  ILexpInvoke(lv_word_of_float, [from])
      | _ ->
	  ILexpCoerce(typ_to, from)
    end
  | _ -> assert false

let make_coerce_to_value2 typ typ_to (f1, f2) =
  match typ.ct_ty with
    Tbuiltin bt ->
      let f = 
	match bt with
	  Tdouble | Tfloat -> assert false
	| Tlonglong | Tulonglong -> lv_cons_dval
	| _ -> lv_cons_val
      in
      ILexpInvoke(f, [f1; f2])
  | Tpointer p ->
      ILexpInvoke(lv_value_of_ptr, [f1; f2])
  | _ -> assert false

let make_specific_read1 ~env target ttype addr fields =
  let p = parse_type ~env ttype in
  let typ = p.packed_t in
  if p.is_struct then
    [ make_il0(IL0stmtReadToTemp(target, typ, ILlvPtr(addr), fields))  ]
  else
    let temp = new_tempid ~env in
    [ make_il0(IL0stmtReadToTemp(temp, typ, ILlvPtr(addr), fields)); (* TODO: ttype *)
      make_il0(IL0stmtDefTemp(target, typ, ILexpCoerce(typ, temp)))
    ]

let make_specific_write1 ~env source stype addr fields =
  [ make_il0(IL0stmtWrite(ILlvPtr(addr), fields, source)) ]

let make_specific_read2 ~env (target1,target2) ttype addr fields =
  let pairtype, injB, cdrtype, injV = get_destructors ~env ttype in
  let typeB = get_base_type ~env ttype in
  let typeV = get_ofs_or_value_type ~env ttype in
  let tid0 = new_tempid ~env in
  let tid1 = new_tempid ~env in
  [ make_il0(IL0stmtReadToTemp(tid0, pairtype,ILlvPtr(addr), fields));
    make_il0(IL0stmtDefTemp(target1, typeB, ILexpInvoke(injB, [tid0])));
    make_il0(IL0stmtDefTemp(tid1, cdrtype, ILexpInvoke(injV, [tid0])));
    make_il0(IL0stmtDefTemp(target2, typeV, ILexpCoerce(typeV, tid1)));
  ]

let make_specific_write2 ~env (source1,source2) stype addr fields =
  let id, insn = make_compound ~env stype source1 source2 in
  (insn @ [make_il0 (IL0stmtWrite (ILlvPtr(addr), fields, id))])

let make_generic_read1 ~env target ttype sttype base ofs =
  let p = parse_type ~env ttype in
  if p.is_struct then begin
    require_name (GNrwmeth ttype);
    [ make_il0 (IL0stmtDefTemp(target, p.packed_t, ILexpInvoke(p.generic_reader, [base; ofs]))) ]
  end else
  let lv, t_tid, inject = 
    get_generic_readers ~env ttype
  in
  let tid = new_tempid ~env in
  let s1 = 
    [
     make_il0 (IL0stmtDefTemp(tid, t_tid, ILexpInvoke(lv, [base; ofs])));
   ]
  in
  let tid2, s2 = match inject with
    Some (typ, _, vfunc) ->
      let tid2 = new_tempid ~env in 
      tid2, [
      make_il0 (IL0stmtDefTemp(tid2, typ, ILexpInvoke(vfunc, [tid])))
    ]
  | None -> 
      tid, []
  in
  let s3 = 
    make_il0 (IL0stmtDefTemp(target, ttype,
			     make_coerce_from_vaddr ttype tid2))
  in
  s1 @ s2 @ [s3]

let make_generic_read2 ~env (t1,t2) ttype sttype base ofs =
  let lv, t_tid, inject = 
    get_generic_readers ~env ttype
  in
  let tid = new_tempid ~env in
  let s1 = 
    [
     make_il0 (IL0stmtDefTemp(tid, t_tid, ILexpInvoke(lv, [base; ofs])));
   ]
  in
  let s2 =
    if is_pointer ttype then begin
      assert (lv == lv_read_word);
      let tid1 = new_tempid ~env in 
      [
       make_il0 (IL0stmtDefTemp(tid1, type_base_t, 
				ILexpInvoke(lv_base_of_value, [tid])));
       make_il0 (IL0stmtDefTemp(t2, type_ofs_t, ILexpInvoke(lv_ofs_of_value, [tid])));
       make_il0 (IL0stmtDefTemp
		   (t1, 
		    type_base_t,
		    ILexpInvoke
		      (lv_set_base_castflag ~env ttype, [tid1; t2])))
     ]
    end
    else if is_numeric ttype then
      let tid2 = new_tempid ~env in 
      match inject with
	Some (typ, bfunc, vfunc) ->
	  [
	   make_il0 (IL0stmtDefTemp(t1, type_base_t, 
				    ILexpInvoke(bfunc, [tid])));
	   make_il0 (IL0stmtDefTemp(tid2, typ, ILexpInvoke(vfunc, [tid])));
	   make_il0 (IL0stmtDefTemp(t2, ttype,
				    make_coerce_from_vaddr ttype tid2))
	 ]
      | None -> assert false
    else assert false
  in
  s1 @ s2

let make_generic_write1 ~env source ttype sttype base ofs =
  let tnull = new_tempid ~env in
  let p = parse_type ~env ttype in
  if p.is_struct then begin
    require_name (GNrwmeth ttype);
    [ make_il0 (IL0stmtDefTemp(tnull, type_void, ILexpInvoke(p.generic_writer, [base; ofs; source]))) ]
  end else
  let lv, t_tid, proj = 
    get_generic_writers ~env ttype
  in
  let tid1_t = match proj with None -> t_tid | Some(t,_) -> t in 
  let tid1 = new_tempid ~env in
  let s3 = 
    make_il0 (IL0stmtDefTemp(tid1, tid1_t,
			     make_coerce_to_vaddr ttype t_tid source))
  in
  let tids, s2 = match proj with
    Some (typ, vfunc) ->
      let tid2 = new_tempid ~env in 
      tid2, [
      make_il0 (IL0stmtDefTemp(tid2, t_tid, ILexpInvoke(vfunc, [tid1])))
    ]
  | None -> 
      tid1, []
  in
  let tidv = new_tempid ~env in
  let exp_strtinfo =
    match sttype with
      None -> ILexpConstant(CTTconstNull)
    | Some st ->
	let t_elem, t_st, lv, fl = storage_for_typeinfo_constant ~genv:env.global_env st in
	ILexpAddress(lv,fl)
  in
  let s1 = 
    [
     make_il0 (IL0stmtDefTemp(tnull, type_typeinfo_ptr, exp_strtinfo));
     make_il0 (IL0stmtDefTemp(tidv, type_void, ILexpInvoke(lv, [base; ofs; tids; tnull])));
   ]
  in
  s3 :: (s2 @ s1)

let make_generic_write2 ~env (src1,src2) ttype sttype base ofs =
  let tnull = new_tempid ~env in
  let lv, t_tid, proj = 
    get_generic_writers ~env ttype
  in
  let tid1 = new_tempid ~env in
  let s3 = 
    make_il0 (IL0stmtDefTemp(tid1, t_tid,
			     make_coerce_to_value2 ttype tid1 (src1, src2)))
  in
  let tidv = new_tempid ~env in
  let exp_strtinfo =
    match sttype with
      None -> ILexpConstant(CTTconstNull)
    | Some st ->
	let t_elem, t_st, lv, fl = storage_for_typeinfo_constant ~genv:env.global_env st in
	ILexpAddress(lv,fl)
  in
  let s1 = 
    [
     make_il0 (IL0stmtDefTemp(tnull, type_typeinfo_ptr, exp_strtinfo));
     make_il0 (IL0stmtDefTemp(tidv, type_void, ILexpInvoke(lv, [base; ofs; tid1; tnull])))
   ]
  in
  s3 :: s1
	   
let rettype ~env = env.rettype

let translate_type ~env t =
  match t with
    ILRtypeVal ({ ct_ty = (Tbuiltin _ | Tvoid) } as t', _) -> t'
  | ILRtypeVal ({ ct_ty = Tstruct _ } as t', _) -> 
      (parse_type env t').packed_t
  | ILRtypeVal (t', _) ->
      failwith ("panic: translate_type val " ^ Ctt_printer.strof_c_type t' );
      (* assert false *)
  | ILRtypeOfs (t,_) -> type_ofs_t
  | ILRtypeBase (t,_) -> type_base_t
  | ILRtypeBaseTemp (t,_) -> type_base_t

let translate_ilstype ~env t =
  match t with
    ILStypeVal ({ ct_ty = Tbuiltin bt } as t') -> t'
  | ILStypeVal _ -> assert false
  | ILStypeOfs (t) -> type_ofs_t
  | ILStypeBase (t) -> type_base_t
  | ILStypeBaseTemp (t) -> type_base_t

let original_type_from_val_type = function
    ILRtypeVal(t,_) | ILRtypeOfs(t,_) -> t
  | _ -> assert false

let is_type_polymorphic_func ~genv t = 
    match t.ct_ty with
      Tfunction(_,_,{ ct_ty = Tpointer { ct_ty = Tvoid }}) -> true
    | Tfunction(_,_,_) -> false
    | _ -> assert false

let rec translate_c_type_genv ~genv t = 
  match t.ct_ty with
    Tvoid -> t
  | Tbuiltin _
  | Tpointer _
  | Tstruct _ ->
      (parse_type_genv ~genv t).packed_t
  | Tfunction(at, varp, rt) ->
      let new_at =
	List.fold_right (translate_c_type_funcarg ~genv)
	  at (if varp then [type_base_t; type_ofs_t] else [])
      in
      let new_at = 
	if is_type_polymorphic_func ~genv t then type_base_t :: new_at else new_at
      in
      make_c_type (Tfunction(new_at, false, translate_c_type_genv ~genv rt))
  | Tarray _ -> assert false
  | Tabstract _ -> assert false

and translate_c_type_funcarg ~genv t acc = 
  match t.ct_ty with
    Tvoid -> assert false
  | Tbuiltin _
  | Tpointer _
  | Tstruct _ -> 
      ((parse_type_genv ~genv t).unpacked_types) @ acc
  | Tfunction _
  | Tarray _
  | Tabstract _ -> assert false

let translate_c_type ~env =
  translate_c_type_genv ~genv:env.global_env

let foldr_with_argtypes ~env t ~varargsf ~widef ~narrowf l r = 
  match t.ct_ty with
    Tfunction(at, vp, _) -> begin
      let rec loop l at = 
	match at, l with
	  [], [] ->
	    if vp then varargsf [] r else r
	| [], _::_ ->
	    if vp then varargsf l r
	    else failwith "fold_with_argtypes: too many args"
	| _::_, [] ->
	    failwith "fold_with_argtypes: not enough args"
	| t::ts, e::rest ->
	    let p = parse_type ~env t in
	    if p.is_fat then
	      widef t e (loop rest ts)
	    else
	      narrowf t e (loop rest ts)
      in
      loop l at
    end
  | _ -> assert false

let flatten_mapi_with_argtypes ~env t ~varargsf ~widef ~narrowf l = 
  let varargsf l r = (varargsf (list_map snd l)) @ r in
  let widef t (i, x) r = (widef i t x) @ r in
  let narrowf t (i, x) r = (narrowf i t x) @ r in
  let rec loop i = function
      [] -> []
    | h::t -> (i, h) :: (loop (i + 1) t)
  in
  let l = loop 0 l in
  foldr_with_argtypes ~env t ~varargsf ~widef ~narrowf l []

let rec translate_func_concrete_args ~env ft args = 
  let rest_insns = ref [] in
  let narrowf _ argt arg = 
    match arg with
      ILSFuncArgNarrow n -> [n]
    | ILSFuncArgWide _ -> assert false
  in
  let widef _ argt arg = 
    match arg with
      ILSFuncArgNarrow n -> assert false
    | ILSFuncArgWide (t1,t2) -> [t1; t2]
  in
  let varargsf l = assert (l = []); [] in
  let args = flatten_mapi_with_argtypes ~env ft ~varargsf ~widef ~narrowf args
  in
  !rest_insns, args

let rec translate_func_varargs ~env args = 
  let length = ref 0 in
  let base_id = new_tempid ~env in
  let arglen_id = new_tempid ~env in
  let ofs_id = new_tempid ~env in
  let insns = ref [] in
  let add_insns l = insns := !insns @ l in
  List.iter
    (fun t ->
      let tidpos = new_tempid ~env in
      let tidput = new_tempid ~env in
      let pos_insn = 
	make_il0 (IL0stmtDefTemp(tidpos, type_size_t, 
				 ILexpConstant(CTTconstInteger(big_int_of_int !length)))) in
      let is_struct, tid1, typ, insns = 
	match t with
	  ILSFuncArgWide (t1, t2) ->
	    let tid1 = new_tempid ~env in
	    let typ =
	      match type_of_tempvar ~env t1 with
		ILRtypeBase (t, _) -> t | _ -> assert false in
	    let p = parse_type ~env typ in
	    let arg_insn = 
	      make_il0 (IL0stmtDefTemp(tid1, p.generic_t,
				       make_coerce_to_value2 typ tid1 (t1, t2)))
	    in
	    false, tid1, typ, [arg_insn]
	| ILSFuncArgNarrow t1 ->
	    let typ = match type_of_tempvar ~env t1 with ILRtypeVal (t, _) -> t | _ -> assert false in
	    let p = parse_type ~env typ in
	    if p.is_struct then
	      true, t1, typ, []
	    else
	    let tid2 = new_tempid ~env in
	    let tid1 = new_tempid ~env in
	    let arg_insn1 = 
	      make_il0 (IL0stmtDefTemp(tid2, p.samesized_word_t,
				       make_coerce_to_vaddr typ p.samesized_word_t t1))
	    in
	    let tid0 = new_tempid ~env in
	    let nullbase_insn = 
	      make_il0 (IL0stmtDefTemp(tid0, type_base_t, 
				       ILexpConstant(constant_zero))) in
	    let arg_insn2 =
	      let f = match p.size with
		Size_byte | Size_hword | Size_word -> lv_cons_val
	      | Size_dword -> lv_cons_dval
	      | Size_other _ -> assert false
	      in
	      make_il0 (IL0stmtDefTemp(tid1, p.generic_t,
				       ILexpInvoke (f, [tid0; tid2])))
	    in
	    false, tid1, typ, [arg_insn1; nullbase_insn; arg_insn2]
      in
      let p = parse_type ~env typ in
      let arglength = int_of_big_int (someof typ.ct_size) in
      if is_struct then begin
	(* TODO: directly-written *)
	require_name (GNrwmeth typ);
	let words = (arglength + Fsc_config.sizeof_pointer - 1) / Fsc_config.sizeof_pointer in
	let pos_insn = 
	  make_il0 (IL0stmtDefTemp(tidpos, type_size_t, 
				   ILexpConstant(
				   CTTconstInteger(big_int_of_int (!length * Fsc_config.sizeof_pointer))))) in
	add_insns [ pos_insn;
		    make_il0 (IL0stmtDefTemp
				(tidput, type_void,
				 ILexpInvoke (p.generic_writer, [base_id; tidpos; tid1])))];
	length := !length + words;
      end
      else 
      let pos_insn = 
	make_il0 (IL0stmtDefTemp(tidpos, type_size_t, 
				 ILexpConstant(CTTconstInteger(big_int_of_int !length)))) in
      if arglength <= Fsc_config.sizeof_pointer then begin
	(* one-word argument *)
	(* arglength 1, 2 may appear in code invoking cast function pointers *)
	add_insns (insns @ 
		   [pos_insn;
		    make_il0 (IL0stmtDefTemp
				(tidput, type_void,
				 ILexpInvoke (lv_put_varargs, [base_id; tidpos; tid1])))]);
	length := !length + 1;
      end
      else if arglength = Fsc_config.sizeof_pointer * 2 then begin
	add_insns (insns @
		   [pos_insn;
		    make_il0 (IL0stmtDefTemp
				(tidput, type_void,
				 ILexpInvoke (lv_put_varargs_2, [base_id; tidpos; tid1])))]);
	length := !length + 2;
      end
      else failwith ("panic968: varargs size " ^ string_of_int arglength))
    args;
  if !length = 0 then
    [
     make_il0 (IL0stmtDefTemp(base_id, type_base_t, ILexpConstant(CTTconstNull)));
     make_il0 (IL0stmtDefTemp(ofs_id, type_ofs_t, ILexpConstant(constant_zero)))
   ], [base_id; ofs_id], []
  else
    let insns = 
      [
       make_il0 (IL0stmtDefTemp(arglen_id, type_size_t, 
				ILexpConstant(CTTconstInteger(big_int_of_int !length))));
       make_il0 (IL0stmtDefTemp(base_id, type_base_t, ILexpInvoke(lv_alloc_varargs, [arglen_id])));
       make_il0 (IL0stmtDefTemp(ofs_id, type_ofs_t, ILexpConstant(constant_zero)))
     ] @ !insns
    in	      
    let post_insns =
      let tid = new_tempid ~env in
      [ make_il0 (IL0stmtDefTemp(tid, type_void, 
				 ILexpInvoke(lv_dealloc_varargs_finished, [base_id]))) ]
    in
    insns, [base_id; ofs_id], post_insns

let translate_funcargs ~env ~tptr_id t args = 
  let argtypes, varargs_p, rettype = 
    match t.ct_ty with
      Tfunction(a,v,r) -> a, v, r
    | _ -> assert false
  in
  let concreteargs, varargs = 
    if varargs_p then 
      Util.split_at_nth (List.length argtypes) args
    else begin
      assert(List.length argtypes = List.length args);
      args, []
    end
  in
  let insns1, new_concrete_args =
    translate_func_concrete_args ~env t concreteargs in
  let insns2, new_var_args, post_insns =
    if varargs_p then
      translate_func_varargs ~env varargs
    else
      [], [], []
  in
  let tptr_arg = if is_type_polymorphic_func ~genv:env.global_env t then [tptr_id] else [] in
  insns1 @ insns2, tptr_arg @ new_concrete_args @ new_var_args, post_insns

let type_of_union_initializer ~env elemtype = 
  make_c_type (Tabstract ("union fsc_initU" ^ encoded_name_of_type ~env elemtype))

let get_storage_type ~genv ?(incomplete_ok = false) t = 
  let elemnum, elemtype = size_and_elemtype_of_onmemory_type ~incomplete_ok t in
  let st = get_storage_type_array ~genv elemtype elemnum in
  { st with ct_volatile_p = t.ct_volatile_p; ct_const_p = false }

let translate_global_function_name ~genv s t = 
  if is_type_polymorphic_func ~genv t then
    "FSP_" ^ encoded_name_of_type_genv ~genv t ^ "_" ^ s 
  else "FS_" ^ encoded_name_of_type_genv ~genv t ^ "_" ^ s 

let translate_global_bridge_function_name ~genv s t = 
  "FG_" ^ s 

let translate_global_variable_name ~known ~genv s t = 
  if known = Unknown_polymorphic then
    "GV_Xuf__" ^ s
  else
    "GV_" ^ encoded_name_of_type_genv ~genv ~array_squash:true t ^ "_" ^ s

let translate_ptr_local_heap_variable_name ~genv s t = 
  "LHVP_" ^ encoded_name_of_type_genv ~genv ~array_squash:false t ^ "_" ^ s

let translate_local_stack_variable_name ~genv s t = 
  "LSV_" ^ encoded_name_of_type_genv ~genv ~array_squash:false t ^ "_" ^ s

let translate_fields ~genv fl =
  match List.rev fl with
    [] -> []
  | (_,elemt)::_ ->
      let fl = list_map 
	  (fun (name,t) -> 
	    let p = parse_type_genv ~genv t in
	    (name, p.initializer_type)) fl in
      let p_elem = parse_type_genv ~genv elemt in
      if p_elem.is_fat then fl @ ["cv", p_elem.packed_t]
      else fl

let get_type_of_field vty fl = 
  let rec iter l = 
    match l with
      [] -> vty
    | [_, t] -> t
    | _::t -> iter t
  in
  iter fl

let storage_for_variable_internal ~genv f ~global vid vty fl = 
  let sz, top_t =
    (size_and_elemtype_of_onmemory_type ~incomplete_ok:true vty)
  in
  let ps = parse_type_genv ~genv top_t in
  let elemt = get_type_of_field top_t fl in
  let name = f ~genv vid vty in
  let sty = get_storage_type ~genv (if global then elemt else vty) in
  let field_add =
    ("val", ps.initializer_type) :: 
    if fl = [] && sz = None then begin
      if ps.is_fat then ["cv", ps.packed_t] else []
    end
    else
      translate_fields ~genv fl
  in
  elemt, sty, name, field_add

let get_global_string_storage_name ~genv id = 
  "GSTR_" ^ string_of_int id

let storage_for_global_string_internal ~genv s = 
  let id = Record_globalnames.get_string_constant_id s in
  let elemt = type_char in
  let ps = parse_type_genv ~genv elemt in
  let sty = get_storage_type ~genv elemt in
  let field = ["val", make_c_type(Tarray(ps.initializer_type, Some (big_int_of_int (String.length s))))] in
  elemt, sty, get_global_string_storage_name ~genv id, field

let cttm_storage_for_global_string ~genv s = 
  let elemt, sty, name, fl = storage_for_global_string_internal ~genv s in
  elemt, sty, CTTMlvVar(name, sty), fl

let storage_for_global_string ~genv s = 
  let elemt, sty, name, fl =
    storage_for_global_string_internal ~genv s in
  elemt, sty, ILlvVar(name, sty), fl

let get_field_offset ~genv =
  Separate_fatpointer.get_field_offset ~genv

let cttm_storage_for_global_variable ~genv ~known vid vty fl = 
  let elemt, sty, name, fl = 
    storage_for_variable_internal ~genv ~global:true 
      (translate_global_variable_name ~known)
      vid vty fl in
  elemt, sty, CTTMlvVar(name, sty), fl

let storage_for_variable ~env vt vid ct fl = 
  let storage_for_global_variable ~env vid vty fl = 
    let elemt, sty, name, fl =
      storage_for_variable_internal ~genv:env.global_env ~global:true
	(translate_global_variable_name ~known:(Hashtbl.find env.decl_env vid))
	vid vty fl in
    [], elemt, sty, ILlvVar(name, sty), fl
  in
  let storage_for_local_heap_variable ~env vid vty fl = 
    let elemt, sty, name, fl =
      storage_for_variable_internal 
	~genv:env.global_env ~global:false translate_ptr_local_heap_variable_name vid vty fl in
    let tid = new_tempid ~env in
    dprintf 7 "%d: elemt = %s, sty = %s vty = %s" tid
      (Ctt_printer.strof_c_type elemt) (Ctt_printer.strof_c_type sty) (Ctt_printer.strof_c_type vty);
    dprintf 7 "%d: volatile = %b %b %b" tid elemt.ct_volatile_p sty.ct_volatile_p vty.ct_volatile_p;
    [ make_il0 (IL0stmtReadToTemp(tid, make_c_type (Tpointer sty), ILlvVar(name, make_c_type (Tpointer sty)),[])) ],
    elemt, sty, ILlvPtr(tid), fl
  in
  let storage_for_local_stack_variable ~env vid vty fl = 
    let elemt, sty, name, fl =
      storage_for_variable_internal ~genv:env.global_env ~global:false translate_local_stack_variable_name vid vty fl in
    [], elemt, sty, ILlvVar(name, sty), fl
  in
  match vt with
    GlobalVar -> storage_for_global_variable ~env vid ct fl
  | HeapVar -> storage_for_local_heap_variable ~env vid ct fl
  | StackVar -> storage_for_local_stack_variable ~env vid ct fl
  | RegVar -> assert false

let cttm_macro_for_blockheader ~genv elemtype elemnum = 
  let elemsize = sizeof_type_genv ~genv elemtype in
  let storage_type = get_storage_type_array elemtype elemnum in
  let allocsize = Big_int.mult_big_int elemsize (Option.default unit_big_int elemnum) in
  let allocsize_e = 
    make_cttm_expr (CTTMexpConstant (CTTconstInteger allocsize)) type_size_t in
  let typeinfo = 
    (*make_cttm_expr (CTTMexpConstant (CTTconstTypeInfo elemtype)) type_typeinfo_ptr *)
    make_cttm_expr
      (CTTMexpRead
	 (CTTMlvVar(encoded_typeinfo_name ~genv elemtype, type_typeinfo_init),
	  ["val", type_typeinfo_struct])) type_typeinfo_struct
  in
  let header =
    make_cttm_expr (CTTMexpInvoke(lv_cttm_macro_emit_header,
				  [typeinfo; allocsize_e]))
      type_fsc_header
  in
  header

let exp_of_error_constant error = 
  let name = match error with
    ILSerrPtrMinus -> "ERR_PTRMINUS"
  | ILSerrPtrComp -> "ERR_PTRCOMP"
  | ILSerrAborted Il.ILabortNotReached -> "ERR_NOTREACHED"
  | ILSerrAborted Il.ILabortOthers s -> "ERR_UNKNOWN"
  in
  ILexpConstant(CTTconstAbstract name)

let translate_rettype_hint ~env hint = 
  let tptr_id = new_tempid ~env in
  let exp, cast = 
    match hint with
      Bunanalyzed | Bunknown | Bscalar ->
	ILexpConstant constant_zero, false
    | Bused t | Bcast t ->
	let t_elem, t_st, lv, fl = storage_for_typeinfo_constant ~genv:env.global_env t in
	ILexpAddress(lv,fl), true
    | Brettype ->
	if env.is_polymorphic then
	  ILexpArgument 0, false
	else
	  let t_elem, t_st, lv, fl = storage_for_typeinfo_constant ~genv:env.global_env env.rettype in
	  ILexpAddress(lv,fl), true
  in
  if cast then begin
    let tptr_id2 = new_tempid ~env in
    tptr_id2, [ make_il0 (IL0stmtDefTemp(tptr_id, type_typeinfo_ptr, exp));
		make_il0 (IL0stmtDefTemp(tptr_id2, type_base_t, ILexpCoerce(type_base_t, tptr_id)))
	      ]
  end 
  else
    tptr_id, [ make_il0 (IL0stmtDefTemp(tptr_id, type_base_t, exp)) ]

let resolve_reference_type ty = 
  match ty with
    ILRtypeBase({ ct_ty = (Tpointer t | Tarray(t,_)) } as tp, binfo) -> 
      make_c_type (Tpointer t), t
  | t -> 
      failwith_p "failure1468: %a@." Ils_formatter.pp_ilr_type t

let translate_expr1 is_wide ~env td e =
  match is_wide, e with
  | false, ILCexpCoerce1(td,vs) -> begin
      let ts = type_of_tempvar ~env vs in
      match ts, td.ct_ty with
	ILRtypeVal ({ ct_ty = Tbuiltin bts }, _), Tbuiltin btd ->
	  [], ILexpCoerce(td, vs), []
      | _ -> failwith "unimp1"
  end
  | false, ILCexpConstant((CTTconstNull | CTTconstInteger _ | CTTconstFloat _) as c) -> [], ILexpConstant(c), []
  | false, ILCexpConstant(CTTconstString s) ->
      let t_elem, t_st, lv, fl = storage_for_global_string ~genv:env.global_env s in
      let temp0 = new_tempid ~env in
      [ make_il0 (IL0stmtDefTemp(temp0, type_void_ptr, ILexpAddress(lv,fl))) ],
      ILexpCoerce(type_base_t, temp0), []
  | false, ILCexpConstant(CTTconstTypeInfo t) ->
      let t_elem, t_st, lv, fl = storage_for_typeinfo_constant ~genv:env.global_env t in
      let temp0 = new_tempid ~env in
      [ make_il0 (IL0stmtDefTemp(temp0, type_void_ptr, ILexpAddress(lv,fl))) ],
      ILexpCoerce(type_base_t, temp0), []
  | false, ILCexpConstant(_) -> failwith "unimp963"
  | false, ILCexpUndefined -> [], ILexpUndefined, []
  | false, ILCexpBinop(bop,t1,t2) -> begin
      match bop, (type_of_tempvar ~env t1), (type_of_tempvar ~env t2) with
	_, ILRtypeBase (_, { ilr_base_cast = bc1 }), ILRtypeBase (_, { ilr_base_cast = bc2 }) ->
	  dprintf 6 "comparing bases (%a, %a)..." 
	    Ils_formatter.pp_trivalue bc1 Ils_formatter.pp_trivalue bc2;
	  let i1, temp1 = 
	    match bc1 with
	      Never -> [], t1
	    | Maybe | Always ->
		let t = new_tempid ~env in
		[ make_il0 (IL0stmtDefTemp(t, type_base_t, 
					   ILexpInvoke(lv_base_remove_castflag, [t1]))) ], t;
	  in
	  let i2, temp2 = 
	    match bc2 with
	      Never -> [], t2
	    | Maybe | Always ->
		let t = new_tempid ~env in
		[ make_il0 (IL0stmtDefTemp(t, type_base_t, 
					   ILexpInvoke(lv_base_remove_castflag, [t2]))) ], t;
	  in
	  (i1 @ i2), ILexpBinop(bop,temp1,temp2), []
      | (ILbinLessThan | ILbinLessEqual | ILbinGtrThan | ILbinGtrEqual), ILRtypeOfs _, ILRtypeOfs _
	    when !use_signed_compare_for_offset
	-> begin
	  let t01 = new_tempid ~env in
	  let t02 = new_tempid ~env in
	  [ make_il0 (IL0stmtDefTemp(t01, type_signed_ofs_t,
				     ILexpCoerce(type_signed_ofs_t, t1)));
	    make_il0 (IL0stmtDefTemp(t02, type_signed_ofs_t,
				     ILexpCoerce(type_signed_ofs_t, t2))) ],
	  ILexpBinop(bop, t01, t02), []
	end
      | _, ILRtypeOfs _, (ILRtypeOfs _ | ILRtypeVal _)
      | _, ILRtypeVal _, ILRtypeVal _ ->
	  [], ILexpBinop(bop,t1,t2), []
      | _ -> assert false
  end
  | false, ILCexpUnaryop(uop,t1) -> [], ILexpUnaryop(uop,t1), []
  | false, ILCexpIdent(id) -> [], ILexpIdent(id), []
	
  | _, ILCexpInvoke(ILSlvVar (vt,f,t),args,cattr,rettype_hint) -> begin
      if vt <> GlobalVar then
	failwith "unimp1306: calling local function";
      if cattr <> Insert_check.attr_nocheck then
	failwith "unimp2";
      let tptr_id, tptr_insn = translate_rettype_hint ~env rettype_hint in
      let pre_insns, new_args, post_insns =
	translate_funcargs ~env ~tptr_id t args in
      let new_name = 
	translate_global_function_name ~genv:env.global_env f t in
      let new_type =
	translate_c_type ~env t in
      pre_insns @ tptr_insn, 
      ILexpInvoke(ILlvVar (new_name, new_type), new_args),
      post_insns
  end
  | is_wide, ILCexpInvoke(ILSlvPtrToFunc (b,o),args,cattr,rettype_hint) -> begin
      let td = original_type_from_val_type td in
      let is_void_ret = td.ct_ty = Tvoid in
      let p = parse_type ~env (if is_void_ret then type_char else td) in (* TODO: dirty workaround *)
      let func_type =
	match (type_of_tempvar ~env b) with
	  ILRtypeBase({ct_ty = Tpointer(t)}, _) -> t
	| _ -> assert false
      in
      let tptr_id, tptr_insn = translate_rettype_hint ~env rettype_hint in
      let concrete_pre_insns, concrete_exp, concrete_post_insns = 
	let pre_insns, new_args, post_insns =
	  translate_funcargs ~env ~tptr_id func_type args in
	let new_type =
	  translate_c_type ~env func_type
	in
	let fptrid0 = new_tempid ~env in
	let fptrid = new_tempid ~env in
	let new_type_ptr = 
	  make_c_type (Tpointer new_type) in
	let fptr_insns = 
	  [ make_il0 
	      (IL0stmtDefTemp(fptrid0, type_void_ptr, ILexpInvoke(lv_get_funcptr, [b])));
	    make_il0
	      (IL0stmtDefTemp(fptrid, new_type_ptr, ILexpCoerce(new_type_ptr, fptrid0))) ]
	in
	fptr_insns @ pre_insns, 
	ILexpInvoke(ILlvPtr (fptrid), new_args),
	post_insns
      in
      let dynamic_insns, dynamic_exp, dynamic_posinsns = 
	let args_insns, new_args, post_insns = 
	  translate_func_varargs ~env args in
	let dval_resid = new_tempid ~env in
	let tptr2_id, tptr2_vid, tptr2_insn = 
	  if p.is_indirect_ret then
	    let tid = new_tempid ~env in
	    let tid_v = new_tempid ~env in
	    let tid_t = new_tempid ~env in
	    tid, tid_v,
	    [
	     make_il0 (IL0stmtDefTemp (tid_t, type_typeinfo_ptr, ILexpConstant(CTTconstTypeInfo td)));
	     make_il0 (IL0stmtDefTemp (tid_v, make_c_type (Tpointer p.packed_t), ILexpInvoke(lv_alloc_valtrampoline, [tid_t])));
	     make_il0 (IL0stmtDefTemp (tid, type_base_t, ILexpCoerce(type_base_t, tid_v)))
	   ]
	  else
	    tptr_id, -1, []
	in
	let invoke_exp = (ILexpInvoke(lv_invoke_genfunc, [tptr2_id; b; o] @ new_args)) in
	let invoke_insn = make_il0 (IL0stmtDefTemp(dval_resid, type_dvalue, invoke_exp)) in
	if is_void_ret then begin
	  [ enclose_sequence (args_insns @ tptr2_insn @ [invoke_insn]) ], ILexpCoerce(type_void, dval_resid), post_insns
	end else if p.is_indirect_ret then begin
	  let tid = new_tempid ~env in
	  let tid_zero = new_tempid ~env in
	  let indirect_insns = 
	    [ 
	      make_il0 (IL0stmtDefTemp(tid_zero, type_void, ILexpCoerce(type_void, dval_resid)));
	      make_il0 (IL0stmtReadToTemp(tid, p.packed_t, ILlvPtr(tptr2_vid), [])) ] in
	  [ enclose_sequence (args_insns @ tptr2_insn @ [invoke_insn] @ indirect_insns) ], ILexpIdent tid, post_insns
	end else begin
	  let is_dvalue = p.generic_t == type_dvalue in
	  let generic_insns, generic_id =
	    match is_dvalue, is_wide with
	      false, false ->
		let val_id = new_tempid ~env in
		let word_id = new_tempid ~env in
		[ make_il0
		    (IL0stmtDefTemp(val_id, type_value, (ILexpInvoke(lv_value_of_dvalue, [dval_resid])))); 
		  make_il0
		  (IL0stmtDefTemp(word_id, type_word, (ILexpInvoke(lv_vaddr_of_value, [val_id])))) ],
		word_id
	    | false, true ->
		let val_id = new_tempid ~env in
		[ make_il0
		    (IL0stmtDefTemp(val_id, type_value, (ILexpInvoke(lv_value_of_dvalue, [dval_resid])))) ],
		val_id
	    | true, false ->
		let dword_id = new_tempid ~env in
		[ make_il0
		    (IL0stmtDefTemp(dword_id, type_dword, (ILexpInvoke(lv_vaddr_of_dvalue, [dval_resid])))) ],
		dword_id
	    | true, true ->
		[], dval_resid
	  in
	  let result_exp = match p.generic_to_packed with
	    Some lv -> ILexpInvoke(lv, [generic_id])
	  | None -> ILexpCoerce(p.packed_t, generic_id)
	  in
	  [enclose_sequence (args_insns @ tptr2_insn @ (invoke_insn :: generic_insns))], result_exp, post_insns
	end
      in
      let result_t = if is_void_ret then type_void else (parse_type ~env td).packed_t in
      let merge_id = new_tempid ~env in
      let temp0 = new_tempid ~env in
      let label0 = new_temp_label temp0 0 in
      let label1 = new_temp_label temp0 1 in
      tptr_insn @
      [ make_il0 (IL0stmtDefTemp (temp0, type_boolean, ILexpInvoke (lv_check_cast, [b])));
	make_il0 (IL0stmtIf (IFNOT, temp0, label0));
      ] @ dynamic_insns @ 
      [ make_il0 (IL0stmtDefTemp (merge_id, result_t, dynamic_exp))
      ] @ dynamic_posinsns @
      [ make_il0 (IL0stmtGoto label1);
	make_il0 (IL0stmtLabel label0);
      ] @ concrete_pre_insns @
      [ make_il0 (IL0stmtDefTemp (merge_id, result_t, concrete_exp)); ]
      @ concrete_post_insns @
      [	make_il0 (IL0stmtLabel label1) ],
      ILexpIdent(merge_id), []
  end
  | false, ILCexpAddress(ILSlvVar (vt, n, ct), []) -> begin
      let i0, t_elem, t_st, lv, fl = storage_for_variable ~env vt n ct [] in
      let temp0 = new_tempid ~env in
      (i0 @ [ make_il0 (IL0stmtDefTemp(temp0, type_void_ptr, ILexpAddress(lv,fl))) ]),
      ILexpCoerce(type_base_t, temp0), []
  end
  | false, ILCexpArgument(i) -> begin
      try [], ILexpArgument(Earray.get env.arguments_n i), []
      with Not_found -> assert false
  end
  | false, ILCexpArgumentB(i) -> begin
      try [], ILexpArgument(Earray.get env.arguments_b i), []
      with Not_found -> assert false
  end
  | false, ILCexpArgumentV(i) -> begin
      try [], ILexpArgument(Earray.get env.arguments_v i), []
      with Not_found -> assert false
  end
  | false, ILCexpCoerce2(td, vb, vo) -> begin
      let ts = type_of_tempvar ~env vb in
      match ts, td.ct_ty with
	ILRtypeBase ({ ct_ty = Tpointer tp }, _), Tbuiltin bt ->
	  (* e.g. char * -> short (narrowing) *)
	  let temp0 = new_tempid ~env in
	  [ make_il0 (IL0stmtDefTemp
			(temp0, type_vaddr_t,
			 ILexpInvoke(lv_vaddr_of_base_ofs, [vb; vo]))) ],
	  ILexpCoerce(td, temp0), []
      | _ -> failwith_p "unimp857 : %a" Ilc_formatter.pp_ilc_expr e
  end
  | false, e -> failwith_p "unimp4N : %a" Ilc_formatter.pp_ilc_expr e
  | true, e -> assert false
;;

let translate_expr2 ~env (ty1,ty2) e =
  match e with
  | ILCexpBinop21((ILbinPlusPV | ILbinMinusPV) as bop,(t1b,t1v),t2) -> begin
      let is_minus = (bop = ILbinMinusPV) in
      let t_t1b = match type_of_tempvar ~env t1b with
	ILRtypeBase(t, _) -> t
      | _ -> assert false
      in
      let p = parse_type ~env t_t1b in
      let tid2 = new_tempid ~env in
      let lv = lv_add_fat_pointer ~env t_t1b in
      let addofs_v, addofs_insns = 
	if is_minus then
	  let tid0 = new_tempid ~env in
	  tid0, [ make_il0 (IL0stmtDefTemp(tid0, type_ofs_t, ILexpUnaryop(UnaryMinus, t2))) ]
	else
	  t2, []
      in
      let (tt,db,cdrtype,dv) = get_destructors ~env t_t1b in
      let tid = new_tempid ~env in
      ( addofs_insns @ 
	[ make_il0 (IL0stmtDefTemp(tid, p.packed_t, ILexpInvoke(lv, [t1b; t1v; addofs_v]))) ],
	ILexpInvoke(db,[tid]),
	ILexpInvoke(dv,[tid]),
	[] )
  end
  | ILCexpCoerce2(td, vb, vo) -> begin
      (* e.g. char * -> int *)
      let ts = type_of_tempvar ~env vb in
      match ts, td.ct_ty with
	ILRtypeBase ({ ct_ty = Tpointer tp }, _), Tbuiltin bt ->
	  let temp0 = new_tempid ~env in
	  [ 
	    make_il0 (IL0stmtDefTemp
			(temp0, type_vaddr_t,
			 ILexpInvoke(lv_vaddr_of_base_ofs, [vb; vo]))) ],
	  ILexpInvoke(lv_base_remove_castflag, [vb]),
	  ILexpCoerce(td, temp0), []
      | ILRtypeBase ({ ct_ty = (Tpointer tsp | Tarray(tsp,_)) }, _), (Tpointer tdp | Tarray(tdp, _)) ->
	  if C_typing.equal_type tsp tdp then
	    [], ILexpIdent(vb), ILexpIdent(vo), []
	  else begin
	    let lv = lv_set_base_castflag ~env td in
	    [], ILexpInvoke(lv, [vb; vo]), ILexpIdent(vo), []
	  end
      | ILRtypeBase ({ ct_ty = Tbuiltin bt }, _), (Tpointer tdp | Tarray(tdp,_)) -> begin
	  let temp0 = new_tempid ~env in
	  let lv = lv_set_base_castflag ~env td in
	  [ 
	    make_il0 (IL0stmtDefTemp
			(temp0, type_ofs_t,
			 ILexpInvoke(lv_ofs_of_base_vaddr, [vb; vo])))
	  ],
	  ILexpInvoke(lv, [vb; temp0]), ILexpIdent(temp0), []
      end
      | _ -> failwith_p
	    "unimp1012: ptrcast wide: %a -> %s"
	    Ils_formatter.pp_ilr_type ts (Ctt_printer.strof_c_type td)
  end
  | ILCexpAddress(ILSlvVar (vt, n, ct), (_::_ as fl)) -> begin
      let sz, top_t =
	size_and_elemtype_of_onmemory_type ~incomplete_ok:true ct
      in
      let ofs = match get_field_offset ~genv:env.global_env ct fl with
	StrOfsNormal o -> o
      | StrOfsBitfield _ -> failwith "T1094: taking offset of bitfields"
      in
      let i0, t_elem, t_st, lv, fl = storage_for_variable ~env vt n ct [] in
      let temp0 = new_tempid ~env in
      let temp1 = new_tempid ~env in

      let targetpointertype, pointedtargettype = resolve_reference_type ty1 in
      let is_cont_opt_ok = 
	!use_optimized_narrow_cast &&
	is_type_continous ~env top_t &&
	is_type_continous ~env pointedtargettype &&
	top_t.ct_size <> None &&
	pointedtargettype.ct_size <> None &&
	mod_big_int (Option.get top_t.ct_size) (Option.get pointedtargettype.ct_size) ==! zero_big_int
      in
      i0 @ [
	   make_il0 (IL0stmtDefTemp(temp0, type_void_ptr, ILexpAddress(lv,fl)));
	   make_il0 (IL0stmtDefTemp(temp1, type_base_t, ILexpCoerce(type_base_t, temp0)))
	 ],
      (if is_cont_opt_ok then
	ILexpIdent(temp1)
      else
	ILexpInvoke(lv_base_put_castflag, [temp1])),
      ILexpConstant(CTTconstInteger ofs), []
  end
  | ILCexpAddress(ILSlvPtr(vb,vo), fields) -> begin
      let pointedsourcetype = match type_of_tempvar ~env vb with
	ILRtypeBase({ ct_ty = Tpointer t}, _) -> t
      | _ -> assert false
      in
      let targetpointertype, pointedtargettype = resolve_reference_type ty1 in
      let offset = match get_field_offset ~genv:env.global_env pointedsourcetype fields with
	StrOfsNormal o -> o
      | StrOfsBitfield _ -> failwith "SF299: taking address of bitfield"
      in
      let tempc = new_tempid ~env in
      let tempo = new_tempid ~env in
      let lv = lv_set_base_castflag ~env targetpointertype in
      [
       make_il0 (IL0stmtDefTemp (tempc, type_ofs_t, ILexpConstant(CTTconstInteger offset)));
       make_il0 (IL0stmtDefTemp (tempo, type_ofs_t, ILexpBinop(ILbinPlusVV,vo,tempc)))
     ],
      ILexpInvoke(lv, [vb;tempo]),
      ILexpIdent(tempo),
      []
  end
  | ILCexpInvoke _ -> begin
      (* bollow translate_expr1 *)
      let add_insn, new_exp, post_insns = translate_expr1 true ~env ty2 e in
      let original_type = original_type_from_val_type ty2 in
      let p = parse_type ~env original_type in
      let tid = new_tempid ~env in
      let tid2 = new_tempid ~env in
      let (tt,db,cdrtype,dv) = get_destructors ~env original_type in
      if is_numeric original_type then
	(add_insn @
	 [make_il0 (IL0stmtDefTemp(tid, tt, new_exp)) ;
	  make_il0 (IL0stmtDefTemp(tid2, cdrtype, ILexpInvoke(dv, [tid])))
	]),
	ILexpInvoke(db,[tid]),
	make_coerce_from_vaddr p.cdr_t tid2,
	post_insns
      else
	(add_insn @
	 [make_il0 (IL0stmtDefTemp(tid, tt, new_exp));
	]),
	ILexpInvoke(db,[tid]),
	ILexpInvoke(dv,[tid]),
	post_insns
  end
  | _ -> failwith_p "unimp4W : %a" Ilc_formatter.pp_ilc_expr e
      
let make_ptr_operations ~env ~idb ~ido ~genf ~specf ~tyd ~fl args =
  let needgen, needspec, needcheck, tys, typ = match type_of_tempvar ~env idb with
    ILRtypeBase ({ ct_ty = Tpointer tp } as typ, _) ->
      true, true, true, tp, typ
  | ILRtypeBase ({ ct_ty = Tarray (tp, _) } as typ, _) ->
      true, true, true, tp, typ
  | ILRtypeBaseTemp ({ ct_ty = Tstruct _ } as typ, _) ->
      false, true, false, typ, typ
  | _ -> assert false
  in
  let temp0 = new_tempid ~env in
  let temp1 = new_tempid ~env in
  let temp2 = new_tempid ~env in
  let temp3 = new_tempid ~env in
  let label0 = new_temp_label temp0 0 in
  let label1 = new_temp_label temp0 1 in
  let i_checkcast = [
    make_il0 (IL0stmtDefTemp (temp0, type_boolean, 
			      ILexpInvoke(lv_check_cast, [idb])));
    make_il0 (IL0stmtIf (IFTRUE, temp0, label0))
  ] in
  let i_spec = 
    let tsp, lv = lv_get_realoffset ~env tys in
    let fl = translate_fields ~genv:env.global_env fl in
    (if needcheck then
      [make_il0 (IL0stmtDefTemp (temp2, type_boolean,
				 ILexpInvoke(lv_check_ofs_ok, [idb; ido])));
       make_il0 (IL0stmtIf (IFNOT, temp2, label0))]
    else [])
    @ [ make_il0 (IL0stmtDefTemp (temp3, tsp, 
				ILexpInvoke(lv, [idb; ido]))) ]
    @ specf ~env args tyd temp3 fl
  in
  let i_mid = [
    make_il0 (IL0stmtGoto label1);
    make_il0 (IL0stmtLabel label0);
  ] in
  let i_gen =
    if fl = [] then
      genf ~env args tyd (Some tyd) idb ido
    else begin
      let tempc = new_tempid ~env in
      let tempo = new_tempid ~env in
      let ofs = 
	match get_field_offset ~genv:env.global_env tys fl with
	  StrOfsNormal ofs -> ofs
	| StrOfsBitfield _ -> failwith "unimp1033: bitfield access"
      in
      make_il0 (IL0stmtDefTemp (tempc, type_ofs_t, 
				ILexpConstant(CTTconstInteger ofs))) ::
      make_il0 (IL0stmtDefTemp (tempo, type_ofs_t, 
				ILexpBinop(ILbinPlusVV, ido, tempc))) ::
      genf ~env args tyd (Some tys) idb tempo
    end
  in
  let i_clearflag = 
    if !clear_castflag_at_memop && needgen then
      [make_il0 (IL0stmtDefTemp (idb, type_base_t, ILexpInvoke(lv_set_base_castflag ~env typ, [idb; ido])))]
    else []
  in
  let i_bottom = [ make_il0 (IL0stmtLabel label1) ]
  in
  let l = 
    if needgen || needcheck then
      if needspec then 
	if Fsc_config.size_check_subsumes_cast_check then
	  i_spec @ i_mid @ i_clearflag @ i_gen @ i_bottom
	else 
	  (if needgen then i_checkcast else []) @ i_spec @ i_mid 
	  @ i_clearflag @ i_gen @ i_bottom
      else
	i_gen
    else
      i_spec
  in
  IL0stmtSequence l
;;

let rec translate_global_initializer_base ~genv ~denv e = 
  let d, t = match e with
    ILSinitbaseNone -> 
      (CTTMexpConstant constant_zero), None
  | ILSinitbaseVar(vt,vid,ct) -> begin
      assert (vt = GlobalVar);
      let _, t_str, lv, fl = 
	cttm_storage_for_global_variable ~genv ~known:(Hashtbl.find denv vid) vid ct [] in
      let e = 
	make_cttm_expr (CTTMexpAddress (lv, fl))
	  (make_c_type (Tpointer t_str))
      in
      (CTTMexpCoerce(type_base_t, e)), Some ct
  end
  | ILSinitbaseString(s) ->
      let _, t_str, lv, fl = cttm_storage_for_global_string ~genv s in
      let e = 
	make_cttm_expr (CTTMexpAddress (lv, fl))
	  (make_c_type (Tpointer t_str))
      in
      (CTTMexpCoerce(type_base_t, e)), Some (make_c_type (Tpointer type_char))
  | ILSinitbaseTypeInfo t -> 
      let e = 
	make_cttm_expr (CTTMexpConstant (CTTconstTypeInfo t)) (type_void_ptr)
      in
      (CTTMexpCoerce(type_base_t, e)), None
  in
  make_cttm_expr d type_base_t, t

let constant_of_ilsinitofs = function
    ILSinitofsInt i -> CTTconstInteger i
  | ILSinitofsFloat f -> CTTconstFloat f

let is_pointer_cast ~region_target_type ~value_target_type ~unknown_p ~offset = 
  let r = 
  if C_typing.is_type_loose_function value_target_type && unknown_p
  then
    (dprintf 5 "emitting value of type %s as constant"
       (Ctt_printer.strof_c_type value_target_type);
     true)
  else if C_typing.equal_type region_target_type value_target_type then
    match region_target_type.ct_size with
      None -> not (eq_big_int offset zero_big_int)
    | Some s -> not (eq_big_int (mod_big_int offset s) zero_big_int)
  else
    true
  in
  dprintf 6 "is_ptr_cast: r=%s v=%s o=%s -> %b"
    (Ctt_printer.strof_c_type region_target_type)
    (Ctt_printer.strof_c_type value_target_type)
    (string_of_big_int offset) r;
  r

let select_macro_emit =
  let emittype_ptr_G = make_c_type (Tabstract ("union fsc_initUptr")) in
  let emitfunc_ptr_G = make_cttm_func_lv "EMIT_INITPTR"
      [type_base_t; type_ofs_t; type_boolean] emittype_ptr_G
  in
  let emitfunc_ptr_L = make_cttm_func_lv "init_ptrvalue"
      [type_base_t; type_ofs_t; type_boolean] emittype_ptr_G
  in
  fun ~genv ~is_global t ->
    let p = parse_type_genv ~genv t in
    match t.ct_ty with
      Tbuiltin bt ->
	if p.is_fat then begin
	  let tname = encoded_name_of_btype bt in
	  if is_global then begin
	    let emitname = "EMIT_INIT_" ^ tname in
	    let emittypename = "fsc_initU_" ^ tname in
	    let emittype = make_c_type (Tabstract ("union " ^ emittypename)) in
	    let emitfunc = make_cttm_func_lv emitname [type_base_t; t] emittype
	    in
	    (None, emittype, Some emitfunc, t)
	  end else begin
	    let emitname = "initialize_" ^ tname in
	    let emitfunc = make_cttm_func_lv emitname [type_base_t; t] p.packed_t
	    in
	    (None, p.packed_t, Some emitfunc, t)
	  end
	end else
	  (None, t, None, t)
    | Tpointer t ->
	if is_global then
	  (Some t,emittype_ptr_G, Some emitfunc_ptr_G, type_ofs_t)
	else
	  (Some t,type_ptrvalue, Some emitfunc_ptr_L, type_ofs_t)
    | _ -> assert false

let translate_global_initializer_value ~genv ~denv ~is_global t b o = 
  let region_targettype, emittype, emitfunc, ofs_type = 
    select_macro_emit ~genv ~is_global t in
  let exp_o = make_cttm_expr (CTTMexpConstant(constant_of_ilsinitofs o)) ofs_type in
  match emitfunc with
    Some emitfunc ->
      let exp_base, value_targettype = translate_global_initializer_base ~genv ~denv b in
      let value_targettype = Option.map
	  (fun t -> snd (size_and_elemtype_of_onmemory_type ~incomplete_ok:true t)) value_targettype in
      let cast_flag = match region_targettype,value_targettype,o with
	None, _, _ -> None
      | Some expected, Some real, ILSinitofsInt o ->
	  let unknown_p = match b with
	    ILSinitbaseVar(_vt, vid, _vty) ->
	      (Hashtbl.find denv vid <> Known)
	  | ILSinitbaseTypeInfo _ -> true
	  | ILSinitbaseString _ -> false
	  | ILSinitbaseNone -> assert false
	  in
	  Some
	    (is_pointer_cast
	       ~region_target_type:expected
	       ~value_target_type:real
	       ~unknown_p
	       ~offset:o)
      | Some _, _, _ -> Some true
      in
      let cast_flag = match cast_flag with
	None -> []
      | Some false ->
	  [make_cttm_expr (CTTMexpConstant constant_zero) type_boolean]
      | Some true ->
	  [make_cttm_expr (CTTMexpConstant constant_one) type_boolean]
      in
      make_cttm_expr
	(CTTMexpInvoke(emitfunc, (exp_base :: exp_o :: cast_flag))) emittype
  | None -> 
      assert (b = ILSinitbaseNone);
      exp_o

let translate_local_initializer ~env ~safe i =
  if (not safe) then
    failwith "unimp1155:translate_local_initializer write to unsafe region (use indirect)";
  let rec loop fields_rev = function
      ILSinitPostFill ->
	[]
    | ILSinitArray(t, b) ->
	let elemt = match t.ct_ty with
	  Tarray(t,_) -> t | _ -> assert false
	in
	let p = parse_type ~env elemt in
	Util.map_flatten_i
	  (fun i x ->
	    loop ((Iindexed (big_int_of_int i), p.initializer_type) :: fields_rev) x) b
    | ILSinitConstant(t, b, o) ->
	let p = parse_type ~env t in
	let e = 
	  translate_global_initializer_value ~genv:env.global_env ~is_global: false t b o
	in
	let add_fields = if p.is_fat then [Inamed "cv", p.packed_t] else [] in 
	[List.rev (add_fields @ fields_rev), e]
    | ILSinitStruct(t, b) ->
	let p = parse_type ~env t in
	Util.map_flatten
	  (fun (t, x) ->
	    loop ((Inamed (t), p.initializer_type) :: fields_rev) x) b
  in
  loop [] i

let is_empty_initialization i = 
  let rec loop = function
      ILSinitPostFill -> ()
    | ILSinitConstant _ -> raise Exit
    | ILSinitArray(_, l) ->
	List.iter loop l
    | ILSinitStruct(_, l) ->
	List.iter (fun (_,i) -> loop i) l
  in
  try
    loop i;
    true
  with
    Exit -> false

let translate_local_const_initialization_value ~env t addr fields b o = 
  let p = parse_type ~env t in
  let t_b = new_tempid ~env in
  let t_o = new_tempid ~env in
  let ins_b = match b with
    ILSinitbaseNone -> 
      [ make_il0 (IL0stmtDefTemp(t_b, type_base_t, ILexpConstant(CTTconstNull))) ]
  | ILSinitbaseVar(vt, vid, vty) ->
      assert (vt = GlobalVar);
      let i0, elemt, sty, lv, fl =
	storage_for_variable ~env vt vid vty []
      in
      assert (i0 = []);
      let sty_ptr = make_c_type (Tpointer sty) in
      let sty_id = new_tempid ~env in
      failwith "unimp1673"; (* TODO unimp castflag *)
      (*[ make_il0 (IL0stmtDefTemp(sty_id, sty_ptr, ILexpAddress(lv, fl)));
	make_il0 (IL0stmtDefTemp(t_b, type_base_t, ILexpCoerce(type_base_t, sty_id))) ]*)
  | ILSinitbaseString s ->
      let t_elem, t_st, lv, fl = storage_for_global_string ~genv:env.global_env s in
      let temp0 = new_tempid ~env in
      [ make_il0 (IL0stmtDefTemp(temp0, type_void_ptr, ILexpAddress(lv,fl)));
	make_il0 (IL0stmtDefTemp(t_b, type_base_t, ILexpCoerce(type_base_t, temp0))) ]
  | _ -> failwith_p "unimp1682initbase : %a" Ils_formatter.pp_ils_initializer_value (b, o)
  in
  let ins_o = match o with
    ILSinitofsInt i ->
      [ make_il0 (IL0stmtDefTemp(t_o, p.cdr_t, ILexpConstant(CTTconstInteger i))) ]
  | ILSinitofsFloat f ->
      [ make_il0 (IL0stmtDefTemp(t_o, p.cdr_t, ILexpConstant(CTTconstFloat f))) ]
  in
  if p.is_pointer then begin
    let pv = new_tempid ~env in
    let fields = 
      if fields = [] then []
      else fields @ [ "cv", type_ptrvalue ]
    in
    ins_b @ ins_o @ 
    [ make_il0 (IL0stmtDefTemp(pv, type_ptrvalue, ILexpInvoke(lv_cons_ptrval, [t_b; t_o])));
      make_il0 (IL0stmtWrite(ILlvPtr addr, fields, pv)) ]
  end
  else begin
    let ins_v, t_v = match b with
      ILSinitbaseNone -> [], t_o
    | _ ->
	let t_v = new_tempid ~env in
	[ make_il0(IL0stmtDefTemp(t_v, p.cdr_t, ILexpInvoke(lv_vaddr_of_base_ofs, [t_b; t_o]))) ],
	t_v
    in
    if p.is_fat then begin
      let cons_func, initt = match p.size with
	Size_dword -> lv_cons_dval, type_dvalue
      | Size_word -> lv_cons_val, type_value
      | _ -> assert false
      in
      let v = new_tempid ~env in
      let fields = 
	if fields = [] then []
	else fields @ [ "cv", initt ]
      in
      ins_b @ ins_o @ ins_v @ 
      [ make_il0 (IL0stmtDefTemp(v, p.packed_t, ILexpInvoke(cons_func, [t_b; t_v])));
	make_il0 (IL0stmtWrite(ILlvPtr addr, fields, v)) ]
    end else begin
      ins_b @ ins_o @ ins_v @ 
      [ make_il0 (IL0stmtWrite(ILlvPtr addr, fields, t_v)) ]
    end
  end

let rec translate_local_const_initialization ~env t inits target_base flds = 
  dprintf 6
    "translate_local_const_initialization : type = %a, init = %a"
    (formatter Ctt_printer.strof_c_type) t
    Ils_formatter.pp_ils_initializer inits;
  match t.ct_ty, inits with
    Tarray(elemtyp, _), ILSinitArray(_t, l) ->
      let p = parse_type ~env elemtyp in
      let target_ptr_t = make_c_type (Tpointer p.packed_t) in
      let tempid = new_tempid ~env in
      let i0s = 
	if flds = [] then
	  [ make_il0 (IL0stmtDefTemp(tempid, target_ptr_t, ILexpCoerce(target_ptr_t, target_base))) ]
	else
	  if p.is_fat then
	    let tid2 = new_tempid ~env in
	    [
	     make_il0 (IL0stmtDefTemp(tid2, make_c_type (Tpointer p.initializer_type),
				      ILexpAddress(ILlvPtr target_base, flds)));
	     make_il0 (IL0stmtDefTemp(tempid, target_ptr_t, ILexpCoerce(target_ptr_t, tid2)))
	    ]
	  else
	    let tid2 = new_tempid ~env in
	    [ make_il0 (IL0stmtDefTemp(tid2, type_void_ptr, ILexpAddress(ILlvPtr target_base, flds)));
	     make_il0 (IL0stmtDefTemp(tempid, target_ptr_t, ILexpCoerce(target_ptr_t, tid2)))
	    ]
      in
      let f n init = 
	let nid = new_tempid ~env in
	let pid = new_tempid ~env in
	let i0 = make_il0
	    (IL0stmtDefTemp(nid, type_size_t, ILexpConstant(CTTconstInteger (big_int_of_int n)))) in
	let i1 = make_il0
	    (IL0stmtDefTemp(pid, target_ptr_t, ILexpBinop(ILbinPlusPV, tempid, nid))) in
	i0 :: i1 :: translate_local_const_initialization ~env elemtyp init pid []
      in
      i0s @ Util.map_flatten_i f l
  | Tstruct sid, ILSinitStruct(_t, inits) ->
      let sd = get_struct_desc ~genv:env.global_env sid in
      Util.map_flatten
	(fun (id,init) ->
	  let (elemt,_ofs) = List.assoc id sd.str_fields_byname in
	  let elempackt = 
	    match elemt.ct_ty with
	      Tarray (t,sz) ->
		make_c_type (Tarray ((parse_type ~env t).initializer_type, sz))
	    | _ -> (parse_type ~env elemt).initializer_type
	  in
	  translate_local_const_initialization ~env elemt init target_base (flds @ [id, elempackt])) inits
  | _, ILSinitPostFill -> []
  | (Tbuiltin _ | Tpointer _), ILSinitConstant(_t, b, o) ->
      translate_local_const_initialization_value ~env t target_base flds b o
  | _ -> assert false

let rec translate_stmt ~env s = 
  dprintf 6 "traslate_stmt: %a" Ilc_formatter.pp_ilc0 s;
  let desc = match s with
    ILC0stmtLabel(id) -> IL0stmtLabel(id)
  | ILC0stmtIf(ift,id,tgt) -> IL0stmtIf(ift,id,tgt)
  | ILC0stmtSwitch(tid,list) -> IL0stmtSwitch(tid,list)
  | ILC0stmtGoto(tgt) -> IL0stmtGoto(tgt)
  | ILC0stmtReturn0 -> 
      IL0stmtSequence (env.exit_code @ [make_il0 (IL0stmtReturn(None))])
  | ILC0stmtReturn1(id) ->
      IL0stmtSequence (env.exit_code @ [make_il0 (IL0stmtReturn(Some id))])
  | ILC0stmtReturn2(i1,i2) ->
      let t = rettype ~env in
      let id, insn = make_compound ~env t i1 i2 in
      IL0stmtSequence
	(insn @ env.exit_code @ [make_il0 (IL0stmtReturn (Some id))])
  | ILC0stmtDefTemp(tid,exp) -> begin
      let ty = type_of_tempvar ~env tid in
      let add_insn, new_exp, post_insns = translate_expr1 false ~env ty exp in
      let insn = IL0stmtDefTemp(tid,translate_type ~env ty, new_exp) in
      match add_insn, post_insns with
	[], [] -> insn
      |	add_insn, [] ->
	  IL0stmtParallel(add_insn @ [make_il0 insn])
      | [], post_insns ->
	  IL0stmtSequence(make_il0 insn :: post_insns)
      | add_insn, post_insns ->
	  IL0stmtSequence
	    (make_il0 (IL0stmtParallel(add_insn @ [make_il0 insn])) :: post_insns)
  end
  | ILC0stmtDefTemp2(t1,t2,exp) -> begin
      let ty1 = type_of_tempvar ~env t1 in
      let ty2 = type_of_tempvar ~env t2 in
      let add_insn, new_exp1, new_exp2, post_insns = translate_expr2 ~env (ty1,ty2) exp in
      IL0stmtParallel
	(add_insn @
	 [ make_il0 (IL0stmtDefTemp(t1, translate_type ~env ty1, new_exp1));
	   make_il0 (IL0stmtDefTemp(t2,translate_type ~env ty2, new_exp2)) ]
	 @ post_insns)
  end
  | ILC0stmtSequence l ->
      IL0stmtSequence(Util.list_map (translate_stmt ~env) l)
  | ILC0stmtParallel l ->
      IL0stmtParallel(Util.list_map (translate_stmt ~env) l)
  | ILC0stmtReadToTemp1(tid,ILSlvVar (vt,vid,ct),fl,attr) -> begin
      let tyd = type_of_tempvar ~env tid in
      if attr <> Insert_check.attr_nocheck then 
	failwith "unimp6";
      let tyd = translate_type ~env tyd in
      let i0, t_elem, t_str, lv, fl = storage_for_variable ~env vt vid ct fl in
      let i1 = IL0stmtReadToTemp(tid,tyd,lv,fl) in
      if i0 = [] then i1
      else IL0stmtParallel(i0 @ [make_il0 i1])
  end
  | ILC0stmtReadToTemp1(tid,ILSlvTemp ts,fl,attr) -> begin
      let tyd = type_of_tempvar ~env tid in
      if attr <> Insert_check.attr_nocheck then 
	failwith "unimp6";
      let tyd = translate_type ~env tyd in
    let fl = translate_fields ~genv:env.global_env fl in
      IL0stmtReadToTemp(tid,tyd,ILlvTemp ts,fl)
  end
  | ILC0stmtReadToTemp2(t1, t2, ILSlvVar (vt, vid, ty), fl, attr) ->
      if attr <> Insert_check.attr_nocheck then 
	failwith "unimp6W";
      let tid0 = new_tempid ~env in
      let tid1 = new_tempid ~env in
      let i0, t_elem, t_str, lv, fl = storage_for_variable ~env vt vid ty fl in
      let p = parse_type ~env t_elem in
      let (tt,db,cdrtype,dv) = get_destructors ~env t_elem in
      IL0stmtParallel
	(i0 @ 
         [
	  make_il0 (IL0stmtReadToTemp(tid0,p.packed_t,lv,fl));
	  make_il0 (IL0stmtDefTemp(t1,get_base_type ~env t_elem, ILexpInvoke(db,[tid0])));
	  make_il0 (IL0stmtDefTemp(tid1,cdrtype,ILexpInvoke(dv,[tid0])));
	  make_il0 (IL0stmtDefTemp(t2,p.cdr_t,ILexpCoerce(p.cdr_t,tid1)));
	])
  | ILC0stmtReadToTemp2(t1, t2, ILSlvTemp ts, fl, attr) ->
      if attr <> Insert_check.attr_nocheck then 
	failwith "unimp6W";
      let tid0 = new_tempid ~env in
      let tid1 = new_tempid ~env in
      let t_elem = snd (ExtList.List.last fl) in
      let p = parse_type ~env t_elem in
      let fl = translate_fields ~genv:env.global_env fl in
      let (tt,db,cdrtype,dv) = get_destructors ~env t_elem in
      IL0stmtParallel
         [
	  make_il0 (IL0stmtReadToTemp(tid0,p.packed_t,ILlvTemp ts,fl));
	  make_il0 (IL0stmtDefTemp(t1,get_base_type ~env t_elem, ILexpInvoke(db,[tid0])));
	  make_il0 (IL0stmtDefTemp(tid1,cdrtype,ILexpInvoke(dv,[tid0])));
	  make_il0 (IL0stmtDefTemp(t2,p.cdr_t,ILexpCoerce(p.cdr_t,tid1)));
	]
  | ILC0stmtReadToTemp1(tid,ILSlvPtr (idb,ido),fl,attr) ->
      let tyd' = type_of_tempvar ~env tid in
      let tyd = original_type_from_val_type tyd' in
      make_ptr_operations ~env ~idb ~ido ~tyd ~fl
	~genf:make_generic_read1 ~specf:make_specific_read1
	tid
  | ILC0stmtReadToTemp2(t1,t2,ILSlvPtr (idb,ido),fl,attr) ->
      let tyd = original_type_from_val_type
	  (type_of_tempvar ~env t2) in
      make_ptr_operations ~env ~idb ~ido ~tyd ~fl
	~genf:make_generic_read2 ~specf:make_specific_read2
	(t1,t2)
  | ILC0stmtWrite1(ILSlvPtr (idb,ido),fl,tid,attr) ->
      let tyd' = type_of_tempvar ~env tid in
      let new_tyd = translate_type ~env tyd' in
      let tyd = original_type_from_val_type tyd' in
      make_ptr_operations ~env ~idb ~ido ~tyd ~fl
	~genf:make_generic_write1 ~specf:make_specific_write1
	tid
  | ILC0stmtWrite2(ILSlvPtr (idb,ido),fl,t1,t2,attr) ->
      let tyd = original_type_from_val_type
	  (type_of_tempvar ~env t2) in
      make_ptr_operations ~env ~idb ~ido ~tyd ~fl
	~genf:make_generic_write2 ~specf:make_specific_write2
	(t1,t2)
  | ILC0stmtWrite1(ILSlvVar (vt,vid,ty),fl,tid,attr) -> begin
      let tys = type_of_tempvar ~env tid in
      if attr <> Insert_check.attr_nocheck then 
	failwith "unimp6o";
      let i0, t_elem, t_str, lv, fl = storage_for_variable ~env vt vid ty fl in
      let i1 = IL0stmtWrite(lv,fl,tid) in
      if i0 = [] then i1
      else IL0stmtParallel (i0 @ [ make_il0 i1 ])
  end
  | ILC0stmtWrite2(ILSlvVar (vt,vid,ty),fl,t1,t2,attr) -> begin
      let tys = type_of_tempvar ~env t2 in
      if attr <> Insert_check.attr_nocheck then
	failwith "unimp6oW";
      let i00, t_elem, t_str, lv, fl = storage_for_variable ~env vt vid ty fl in
      let p = parse_type ~env t_elem in
      assert p.is_fat;
      let tid, i0 = make_compound ~env t_elem t1 t2 in
      IL0stmtParallel (i0 @ i00 @ [ make_il0 (IL0stmtWrite(lv,fl,tid)) ])
  end
  | ILC0stmtAbort error -> begin
      let tid1 = new_tempid ~env in
      let tid_error = new_tempid ~env in
      let tid2 = new_tempid ~env in
      IL0stmtSequence
	[
	  make_il0 (IL0stmtDefTemp (tid1, type_base_t,
				    ILexpConstant(constant_zero)));
	  make_il0 (IL0stmtDefTemp (tid_error, type_fsc_error,
				    exp_of_error_constant error));
	  make_il0 (IL0stmtDefTemp (tid2, type_void,
				    ILexpInvoke(lv_abort, [tid1; tid_error])));
	]
  end
  | ILC0stmtAbortIf(expr, error) -> begin
      let tid = new_tempid ~env in
      let tid1 = new_tempid ~env in
      let tid_error = new_tempid ~env in
      let tid2 = new_tempid ~env in
      let label0 = new_temp_label tid 0 in
      let add_insn, new_exp, post_insns = 
	translate_expr1 false ~env (ILRtypeVal (type_boolean, Value_any)) expr
      in
      let insn = IL0stmtDefTemp(tid, type_boolean, new_exp) in
      let icond = 
	match add_insn, post_insns with
	  [], [] -> insn
	| add_insn, post_insns ->
	    IL0stmtSequence(add_insn @ [make_il0 insn] @ post_insns)
      in
      IL0stmtSequence
	[ make_il0 icond;
	  make_il0 (IL0stmtIf (IFNOT, tid, label0));
	  make_il0 (IL0stmtDefTemp (tid1, type_base_t,
				    ILexpConstant(constant_zero)));
	  make_il0 (IL0stmtDefTemp (tid_error, type_fsc_error,
				    exp_of_error_constant error));
	  make_il0 (IL0stmtDefTemp (tid2, type_void,
				    ILexpInvoke(lv_abort, [tid1; tid_error])));
	  make_il0 (IL0stmtLabel label0)
	]
  end
  | ILC0stmtDeclScalar (HeapVar, cty, id)
  | ILC0stmtDeclBulk (HeapVar, cty, id)
  | ILC0stmtDeclScalar (StackVar, cty, id)
  | ILC0stmtDeclBulk (StackVar, cty, id) ->
      IL0stmtParallel [] (* now in prescan *)
  | ILC0stmtInitialize((HeapVar | StackVar (* TODO *)) as vt, cty, id, inits) -> begin
      let wholetyp, elemtyp, size = match cty.ct_ty with
	Tarray(t, Some s) -> cty, t, s
      | Tstruct _ -> cty, cty, unit_big_int
      | _ -> assert false
      in
      dprintf 6 "BULKINIT: %s" (Ctt_printer.strof_c_type cty);
      let inits = 
	begin
	  dprintf 6 ">>";
	  let t1 = new_tempid ~env in 
	  let p = parse_type ~env elemtyp in
	  let pt = make_c_type (Tpointer p.packed_t) in
	  dprintf 6 "  cty=%s" (Ctt_printer.strof_c_type cty);
	  let i0, _elemt, sty, lv, fl = storage_for_variable ~env vt id cty [] in
	  let dt = get_type_of_field sty fl in
	  dprintf 6 "  storage type=%s" (Ctt_printer.strof_c_type dt);
	  let ptarget = new_tempid ~env in 
	  let i1 = make_il0 (IL0stmtDefTemp (ptarget, make_c_type (Tpointer dt), ILexpAddress(lv, fl))) in
	  (* let ctarget = new_tempid ~env in 
	  let i2 = make_il0 (IL0stmtDefTemp (ctarget, pt, ILexpCoerce(pt, ptarget))) in *)
	  let ctarget = ptarget in
	  let r = i0 @ (i1 (* :: i2 *) :: translate_local_const_initialization ~env wholetyp inits ctarget []) in
	  dprintf 6 "<<"; r
	end
      in
      IL0stmtParallel inits
  end
  | ILC0stmtDeclScalar ((GlobalVar | RegVar), _, _)
  | ILC0stmtDeclBulk ((GlobalVar | RegVar), _, _)
  | ILC0stmtInitialize ((GlobalVar | RegVar), _, _, _) -> assert false

  | ILC0stmtReadToTemp1
      (_, (ILSlvSVar (_, _, _)|ILSlvPtrToFunc (_, _)), _, _)
  | ILC0stmtReadToTemp2
      (_, _, (ILSlvSVar (_, _, _)|ILSlvPtrToFunc (_, _)), _, _)
  | ILC0stmtWrite1
      ((ILSlvSVar (_, _, _)|ILSlvPtrToFunc (_, _)|ILSlvTemp _), _, _, _)
  | ILC0stmtWrite2
      ((ILSlvSVar (_, _, _)|ILSlvPtrToFunc (_, _)|ILSlvTemp _), _, _, _, _)
    -> failwith_p "unimp5 %a" Ilc_formatter.pp_ilc0 s
  in
  make_il0 desc

(* prescan phase: pickup all local variable declarations *)

type prescan_env = 
    { heapvars : (c_type * identifier) Glist.t;
      stackvars : (c_type * identifier) Glist.t; }

let rec prescan_stmt ~penv s = 
  dprintf 6 "presccan: %a" Ilc_formatter.pp_ilc0 s;
  match s with
    ILC0stmtLabel(id) -> ()
  | ILC0stmtIf(ift,id,tgt) -> ()
  | ILC0stmtSwitch(tid,list) -> ()
  | ILC0stmtGoto(tgt) -> ()
  | ILC0stmtReturn0 -> ()
  | ILC0stmtReturn1(id) -> ()
  | ILC0stmtReturn2(i1,i2) -> ()
  | ILC0stmtDefTemp(tid,exp) -> ()
  | ILC0stmtDefTemp2(t1,t2,exp) -> ()
  | ILC0stmtSequence l ->
      List.iter (prescan_stmt ~penv) l
  | ILC0stmtParallel l ->
      List.iter (prescan_stmt ~penv) l
  | ILC0stmtReadToTemp1(tid,ILSlvVar (vt,vid,ct),fl,attr) -> ()
  | ILC0stmtReadToTemp1(tid,ILSlvTemp ts,fl,attr) -> ()
  | ILC0stmtReadToTemp2(t1, t2, ILSlvVar (vt, vid, ty), fl, attr) -> ()
  | ILC0stmtReadToTemp2(t1, t2, ILSlvTemp ts, fl, attr) -> ()
  | ILC0stmtReadToTemp1(tid,ILSlvPtr (idb,ido),fl,attr) -> ()
  | ILC0stmtReadToTemp2(t1,t2,ILSlvPtr (idb,ido),fl,attr) -> ()
  | ILC0stmtWrite1(ILSlvPtr (idb,ido),fl,tid,attr) -> ()
  | ILC0stmtWrite2(ILSlvPtr (idb,ido),fl,t1,t2,attr) -> ()
  | ILC0stmtWrite1(ILSlvVar (vt,vid,ty),fl,tid,attr) -> ()
  | ILC0stmtWrite2(ILSlvVar (vt,vid,ty),fl,t1,t2,attr) -> ()
  | ILC0stmtAbort error -> ()
  | ILC0stmtAbortIf(expr, error) -> ()

  | ILC0stmtDeclScalar (HeapVar, cty, id)
  | ILC0stmtDeclBulk (HeapVar, cty, id) ->
      Glist.put penv.heapvars (cty, id);
      ()

  | ILC0stmtDeclScalar (StackVar, cty, id)
  | ILC0stmtDeclBulk (StackVar, cty, id) ->
      Glist.put penv.stackvars (cty, id);
      ()

  | ILC0stmtInitialize((HeapVar | StackVar (* TODO *)) as vt, cty, id, inits) -> ()
  | ILC0stmtDeclScalar ((GlobalVar | RegVar), _, _)
  | ILC0stmtDeclBulk ((GlobalVar | RegVar), _, _)
  | ILC0stmtInitialize ((GlobalVar | RegVar), _, _, _) -> assert false

  | ILC0stmtReadToTemp1
      (_, (ILSlvSVar (_, _, _)|ILSlvPtrToFunc (_, _)), _, _)
  | ILC0stmtReadToTemp2
      (_, _, (ILSlvSVar (_, _, _)|ILSlvPtrToFunc (_, _)), _, _)
  | ILC0stmtWrite1
      ((ILSlvSVar (_, _, _)|ILSlvPtrToFunc (_, _)|ILSlvTemp _), _, _, _)
  | ILC0stmtWrite2
      ((ILSlvSVar (_, _, _)|ILSlvPtrToFunc (_, _)|ILSlvTemp _), _, _, _, _)
    -> failwith_p "unimp5p %a" Ilc_formatter.pp_ilc0 s

let prescan ~make_frame ~env body = 
  let penv =
    { heapvars = Glist.empty ();
      stackvars = Glist.empty (); } in
  
  prescan_stmt ~penv body;
  
  let heapvars, stackvars = Glist.to_list penv.heapvars, Glist.to_list penv.stackvars in

  List.iter
    (fun (cty, id) ->
      let elemtyp, size = match cty.ct_ty with
	Tarray(t, Some s) -> t, s
      | Tarray(t, None) -> failwith "panic 2063: allocating unsized local array"
      | _ -> cty, unit_big_int
      in
      check_size_of_local_variable ~genv:env.global_env id cty;
      let stype = get_storage_type ~genv:env.global_env cty in
      let tid_p = new_tempid ~env in
      let tid_t = new_tempid ~env in
      let tid_s = new_tempid ~env in
      let tid_v = new_tempid ~env in
      let ptrval_name = translate_ptr_local_heap_variable_name ~genv:env.global_env id cty in
      let ptrval_type = 
	make_c_type (Tpointer { stype with ct_volatile_p = cty.ct_volatile_p }) in
      Glist.put env.prologue0
	(make_il0 
	   (IL0stmtDeclAutoScalar (Auto, ptrval_type, ptrval_name, None)));
      Glist.put env.prologue1
	(make_il0 
	   (IL0stmtParallel
	      [make_il0 (IL0stmtDefTemp (tid_s, type_size_t, ILexpConstant(CTTconstInteger size)));
	       make_il0 (IL0stmtDefTemp (tid_t, type_typeinfo_ptr, ILexpConstant(CTTconstTypeInfo elemtyp)));
	       make_il0 (IL0stmtDefTemp (tid_p, type_void_ptr, ILexpInvoke(lv_alloc_hvar, [tid_t; tid_s])));
	       make_il0 (IL0stmtWrite (ILlvVar(ptrval_name, ptrval_type), [], tid_p))
	     ]));
      if not make_frame then
	(* in make_frame mode, variables are released by stack_frame management. *)
	env.exit_code <-
	  env.exit_code @ 
	  [make_il0
	     (IL0stmtSequence
		[make_il0 (IL0stmtDefTemp (tid_v, type_void, ILexpInvoke(lv_dealloc_heapvar, [tid_p])))])];
      ())
    heapvars;

  if make_frame then begin
    let n_hvar = List.length heapvars in
    let inst_fvar = 
      if n_hvar > 0 then begin
	Glist.put env.prologue0
	  (make_il0 
	     (IL0stmtDeclBulk
		(Auto, type_frame_variables_n (n_hvar + 1), id_local_frame_variables, 
		 Some
		   (ILinitArray
		    (type_frame_variables_entry, 
		     [ILinitAbstractCtt
			(make_cttm_expr
			   (CTTMexpConstant constant_zero)
			   type_frame_variables_entry)])))));
	Util.list_iteri
	  (fun i (cty, id) ->
	    let tid_p = new_tempid ~env in
	    let tid_pp = new_tempid ~env in
	    let tid_i = new_tempid ~env in
	    let tid_v = new_tempid ~env in
	    
	    Glist.put env.prologue1
	      (make_il0
		 (IL0stmtParallel
		    [
		     make_il0
		       (IL0stmtDefTemp(tid_pp, type_frame_variables,
				       ILexpAddress(lv_local_frame_variables, [])));
		     make_il0
		       (IL0stmtDefTemp(tid_i, type_int,
				       ILexpConstant(CTTconstInteger (big_int_of_int i))));
		     make_il0
		       (IL0stmtDefTemp(tid_p, make_c_type (Tpointer type_frame_variables_entry),
				       ILexpBinop(ILbinPlusPV, tid_pp, tid_i)));
		     make_il0
		       (IL0stmtReadToTemp(tid_v, type_frame_variables_entry,
					  ILlvVar(translate_ptr_local_heap_variable_name ~genv:env.global_env id cty,
						  type_frame_variables_entry), []));
		     make_il0
		       (IL0stmtWrite(ILlvPtr tid_p, [], tid_v))])))
	  heapvars;
	(fun tid -> IL0stmtReadToTemp(tid, make_c_type (Tpointer type_frame_variables_entry), 
				      lv_local_frame_variables, [])) (* type ok? *)
      end
      else
	(fun tid -> IL0stmtDefTemp(tid, make_c_type (Tpointer type_frame_variables_entry), ILexpConstant(constant_zero)))
    in

    Glist.put env.prologue0
      (make_il0
	 (IL0stmtDeclAutoScalar(Auto, type_stack_frame, id_local_frame, None)));

    let tid_f = new_tempid ~env in
    let tid_fv = new_tempid ~env in
    let tid_s = new_tempid ~env in
    let tid_v = new_tempid ~env in

    Glist.put env.prologue1
      (make_il0 
	 (IL0stmtParallel
	    [make_il0
	       (IL0stmtDefTemp(tid_f, make_c_type (Tpointer type_stack_frame),
			       ILexpAddress(ILlvVar(id_local_frame, type_stack_frame), [])));
	     make_il0
	       (inst_fvar tid_fv);
	     make_il0 (IL0stmtDefTemp (tid_s, make_c_type (Tpointer type_char),
				       ILexpConstant(CTTconstString env.name)));
	     make_il0 (IL0stmtDefTemp (tid_v, type_void, ILexpInvoke(lv_create_stack_frame,
								     [tid_f; tid_s; tid_fv])));
	   ]));
    let tid_v = new_tempid ~env in
    env.exit_code <-
      env.exit_code @ [make_il0 (IL0stmtDefTemp (tid_v, type_void, ILexpInvoke(lv_free_stack_frame, [])))]
  end else ();

  List.iter
    (fun (cty, id) ->
      let elemtyp, size = match cty.ct_ty with
	Tarray(t, Some s) -> t, s
      | Tarray(t, None) -> failwith "panic 2063: allocating unsized local array"
      | _ -> cty, unit_big_int
      in
      check_size_of_local_variable ~genv:env.global_env id cty;
      let stype = get_storage_type ~genv:env.global_env cty in
      
      let val_name = translate_local_stack_variable_name ~genv:env.global_env id cty in
      
      let header =
	cttm_macro_for_blockheader ~genv:env.global_env elemtyp (Some size)
      in
      
      let init = 
	ILinitStruct 
	  (stype,
	   ["header", ILinitAbstractCtt header]) in
      
      Glist.put env.prologue0
	(make_il0 
	   (IL0stmtDeclBulk (Auto, stype, val_name, Some init)));
      ())
    stackvars;
  ()

let translate_function_body ~genv ~denv ~name ct args
    ({ Ilc0.body = body;
       Ilc0.max_variable_number = max_variable_number;
       Ilc0.variable_environment = variable_environment;
       Ilc0.arguments = arguments } as f) =
  let arguments_b = Earray.empty () in
  let arguments_v = Earray.empty () in
  let arguments_n = Earray.empty () in
  let is_polymorphic = is_type_polymorphic_func ~genv ct in
  let env =
    { max_var = max_variable_number;
      f = f;
      is_polymorphic = is_polymorphic;
      global_env = genv;
      decl_env = denv;
      rettype = rettype_of_functype ct;
      arguments_b = arguments_b;
      arguments_v = arguments_v;
      arguments_n = arguments_n;
      prologue0 = Glist.empty ();
      prologue1 = Glist.empty ();
      epilogue = Glist.empty ();
      exit_code = [];
      struct_is_continuous = Earray.empty ();
      name = name;
    } in
  let arglen = List.length args in
  let arg_is_wide = Array.create arglen false in
  let widef i t name =
    arg_is_wide.(i) <- true;
    ["FAB_" ^ name; "FAV_" ^ name]
  in
  let narrowf i t name = 
    arg_is_wide.(i) <- false;
    ["FA_" ^ name]
  in
  let varargsf l = ["FAva_b"; "FAva_v"] in
  let args = 
    let args = flatten_mapi_with_argtypes ~env ~widef ~narrowf ~varargsf ct args
    in
    if is_polymorphic then "FAtptr_b" :: args else args
  in
  let _ = 
    let n = ref (if is_polymorphic then 1 else 0) in
    for i = 0 to arglen - 1 do
      if arg_is_wide.(i) then begin
	Earray.set arguments_b i !n;
	incr n;
	Earray.set arguments_v i !n;
	incr n;
      end else begin
	Earray.set arguments_n i !n;
	incr n
      end
    done;
  in
  
  prescan
    ~make_frame:(f.Ilc0.function_attributes.setjmp_called || !force_emit_stack_unwind)
    ~env body;

  let body = translate_stmt ~env body in
  translate_c_type ~env ct, args, 
  make_il0 (IL0stmtSequence (Glist.to_list env.prologue0 
			     @ Glist.to_list env.prologue1
			     @ [body] @ 
			     Glist.to_list env.epilogue))

let translate_function ~genv ~denv gs ct id args f = 
  let new_id = 
    translate_global_function_name ~genv id ct in
  let ct, args, f = translate_function_body ~genv ~denv ~name:id ct args f in
  let f = { il0_var_types = Earray.empty (); il0_funcbody = f } in
  IL0declFunction(gs, ct, new_id, args, f)

let rec translate_global_initializer_list ~genv ~denv = function
    ILSinitArray(t,l) ->
      CTTMinitList 
	(Util.list_map (translate_global_initializer_list ~genv ~denv) l)
  | ILSinitConstant(t,b,o) -> 
      CTTMinitExp (translate_global_initializer_value ~genv ~denv ~is_global:true t b o)
  | ILSinitPostFill -> assert false
  | ILSinitStruct(t,l) -> 
      CTTMinitList
	(Util.list_map (fun (name, i) -> translate_global_initializer_list ~genv ~denv i) l)

let translate_global_initializer ~genv ~denv t i =
  let elemnum, elemtype = size_and_elemtype_of_onmemory_type t in
(*  let typeinfo = 
    (*make_cttm_expr (CTTMexpConstant (CTTconstTypeInfo elemtype)) type_typeinfo_ptr *)
    make_cttm_expr
      (CTTMexpRead
	 (CTTMlvVar(encoded_typeinfo_name ~genv elemtype, type_typeinfo_init),
	  ["val", type_typeinfo_struct])) type_typeinfo_struct
  in
  let elemsize = sizeof_type_genv ~genv elemtype in
  let storage_type = get_storage_type_array elemtype elemnum in
  let elemnum' = Option.default unit_big_int elemnum in
  let allocsize = Big_int.mult_big_int elemsize elemnum' in
  let allocsize_e = 
    make_cttm_expr (CTTMexpConstant (CTTconstInteger allocsize)) type_size_t in
  let header =
    make_cttm_expr (CTTMexpInvoke(lv_cttm_macro_emit_header,
				  [typeinfo; allocsize_e]))
      type_fsc_header*)
  let header = cttm_macro_for_blockheader ~genv elemtype elemnum
  in
  let inits = match i with
    None -> []
  | Some l -> [translate_global_initializer_list ~genv ~denv l]
  in
  CTTMinitList ((CTTMinitExp header) :: inits)

exception FG_not_supported (* TODO: unimplemented *)

let generate_bridge_function ~genv gs id ct args = 
  let at, varf, rt = match ct.ct_ty with Tfunction(a,v,r) -> a,v,r | _ -> assert false in
  let env = { max_var = 0; f = Obj.magic 0; is_polymorphic = false; global_env = genv; decl_env = Obj.magic 0;
	      rettype = rettype_of_functype ct; 
	      arguments_b = Earray.empty (); arguments_n = Earray.empty (); arguments_v = Earray.empty ();
	      prologue0 = Glist.empty (); prologue1 = Glist.empty (); epilogue = Glist.empty ();
	      exit_code = [];
	      struct_is_continuous = Earray.empty ();
	      name = ""
	    } in
  let new_id = translate_global_bridge_function_name ~genv id ct in
  let spec_id = translate_global_function_name ~genv id ct in
  let argofs = ref 0 in
  let args_rev = ref [] in
  let base_tid = new_tempid ~env in
  let instr_rev = ref [ make_il0 (IL0stmtDefTemp(base_tid, type_base_t, ILexpArgument 1)) ] in
  let put v l = v := List.rev_append l !v in
  List.iter2
    (fun ty name ->
      let p = parse_type_genv ~genv ty in
      let incr, ofs = match p.size with
	Size_byte | Size_hword | Size_word -> Fsc_config.sizeof_pointer, !argofs
      | Size_dword -> Fsc_config.sizeof_pointer * 2, !argofs
      | Size_other (Some s) -> (int_of_big_int s + Fsc_config.sizeof_pointer - 1) / Fsc_config.sizeof_pointer * Fsc_config.sizeof_pointer, !argofs
      | Size_other None ->
	  let s = sizeof_type_genv ~genv ty in
	  dprintf 0 "oop: size of %s unknown (%s)?" (Ctt_printer.strof_c_type ty) (string_of_big_int s);
	  raise FG_not_supported (* TODO? *)
      in
      argofs := !argofs + incr;
      let tido = new_tempid ~env in
      put instr_rev [ make_il0 (IL0stmtDefTemp(tido, type_size_t,
					       ILexpConstant (CTTconstInteger (big_int_of_int ofs)))) ];
      match p.is_fat with
	true ->
	  let tid1, tid2 = new_tempid ~env, new_tempid ~env in
	  put instr_rev (make_generic_read2 ~env (tid1, tid2) ty None base_tid tido);
	  put args_rev [tid1; tid2];
      | false -> begin
	  let b_name = "FAB_" ^ name in
	  match p.size with
	    Size_byte | Size_hword -> begin
	      assert (not p.is_fat);
	      if p.is_floating then
		  failwith "unimp2015";
	      (* read by word size *)
	      let tidv = new_tempid ~env in
	      let tidw = new_tempid ~env in
	      let tid1 = new_tempid ~env in
	      put instr_rev [make_il0 
			       (IL0stmtDefTemp
				  (tidv, type_value,
				   ILexpInvoke (lv_read_word, [base_tid; tido])));
			     make_il0 
			       (IL0stmtDefTemp
				  (tidw, type_word,
				   ILexpInvoke (lv_vaddr_of_value, [tidv])));
			     make_il0 
			       (IL0stmtDefTemp
				  (tid1, type_word,
				   ILexpCoerce (p.packed_t, tidw)))];
	      put args_rev [tid1]
	    end
	  | _ -> begin
	      let tid1 = new_tempid ~env in
	      put instr_rev (make_generic_read1 ~env tid1 ty None base_tid tido);
	      put args_rev [tid1]
	  end
      end)
    at args;
  if varf then begin
    let tido = new_tempid ~env in
    put instr_rev [ make_il0 (IL0stmtDefTemp(tido, type_size_t,
					     ILexpConstant (CTTconstInteger (big_int_of_int !argofs)))) ];
    put args_rev [base_tid; tido];
  end;
  if is_type_polymorphic_func ~genv ct then begin
    let tid_tptr = new_tempid ~env in
    put instr_rev [ make_il0 (IL0stmtDefTemp(tid_tptr, type_base_t,
					     ILexpArgument 0)) ];
    args_rev := !args_rev @ [ tid_tptr ] (* put to the top *)
  end;
  let va_end_insn = 
    if not varf then begin
      let tid = new_tempid ~env in
      [ make_il0 (IL0stmtDefTemp
		    (tid, type_void,
		     ILexpInvoke(lv_va_end, [base_tid]))) ]
    end else
      []
  in
  if rt.ct_ty = Tvoid then begin
    let retp_id = new_tempid ~env in
    let retg_id = new_tempid ~env in
    let tid0 = new_tempid ~env in
    put instr_rev
      [ make_il0 (IL0stmtDefTemp(retp_id, type_void, ILexpInvoke(ILlvVar(spec_id, type_void), List.rev !args_rev))) ];
    put instr_rev va_end_insn;
    put instr_rev
      [ make_il0 (IL0stmtDefTemp(tid0, type_dword, ILexpConstant(constant_zero)));
	make_il0 (IL0stmtDefTemp(retg_id, type_dvalue, ILexpInvoke(lv_nullexpand_dvalue, [tid0])));
	make_il0 (IL0stmtReturn(Some (retg_id)))
      ]
  end else begin
    let rt_p = parse_type ~env rt in
    if rt_p.is_indirect_ret then begin
      let retp_id = new_tempid ~env in
      let retg_id = new_tempid ~env in
      let tid0 = new_tempid ~env in
      put instr_rev
	[ make_il0 (IL0stmtDefTemp(retp_id, rt_p.packed_t, ILexpInvoke(ILlvVar(spec_id, type_void), List.rev !args_rev))) ];
      put instr_rev va_end_insn;
      let dummy1_id = new_tempid ~env in
      let tid_tptr = new_tempid ~env in
      put instr_rev [ make_il0 (IL0stmtDefTemp(tid_tptr, type_base_t, ILexpArgument 0)) ];
      require_name (GNrwmeth rt); (* for generic_writer *)
      put instr_rev
	[ 
	  make_il0 (IL0stmtDefTemp(tid0, type_dword, ILexpConstant(constant_zero)));
	  make_il0 (IL0stmtDefTemp(dummy1_id, type_void, ILexpInvoke(rt_p.generic_writer, [tid_tptr; tid0; retp_id])));
	  make_il0 (IL0stmtDefTemp(retg_id, type_dvalue, ILexpInvoke(lv_nullexpand_dvalue, [tid0])));
	  make_il0 (IL0stmtReturn(Some (retg_id)))
	]
    end else begin
      let retp_id = new_tempid ~env in
      put instr_rev
	[make_il0 (IL0stmtDefTemp
		     (retp_id, rt_p.packed_t, 
		      ILexpInvoke(ILlvVar(spec_id, type_void), List.rev !args_rev)))];
      put instr_rev va_end_insn;
      let retg_id = match rt_p.packed_to_generic with
	None -> retp_id
      | Some lv ->
	  let retg_id = new_tempid ~env in
	  put instr_rev [make_il0 (IL0stmtDefTemp(retg_id, rt_p.generic_t, 
						  ILexpInvoke(lv, [retp_id])))];
	  retg_id
      in
      if rt_p.size < Size_word then begin
	let dv_id = new_tempid ~env in
	put instr_rev 
	  [make_il0 (IL0stmtDefTemp
		       (dv_id, type_dvalue,
			ILexpInvoke(lv_nullexpand_dvalue, [retg_id])));
	   make_il0 (IL0stmtReturn (Some dv_id))]
      end else if rt_p.generic_t == type_dvalue then
	put instr_rev [make_il0 (IL0stmtReturn (Some retg_id))]
      else if rt_p.generic_t == type_value then begin
	let dv_id = new_tempid ~env in
	put instr_rev 
	  [make_il0 (IL0stmtDefTemp
		       (dv_id, type_dvalue,
			ILexpInvoke(lv_dvalue_of_value, [retg_id])));
	   make_il0 (IL0stmtReturn (Some dv_id))]
      end else assert false
    end
  end;
  IL0declFunction(gs, type_genfunc, new_id, ["FAtptr_b"; "FAva_b"],
		  { il0_var_types = Earray.empty (); il0_funcbody = enclose_sequence (List.rev !instr_rev) })

let generate_dummy_bridge_function ~genv gs id ct args = 
  let new_id = translate_global_bridge_function_name ~genv id ct in
  IL0declFunction
    (gs, type_genfunc, new_id, ["FAtptr_b"; "FAva_b"],
     {
      il0_var_types = Earray.empty (); 
      il0_funcbody =
      enclose_sequence
	[
	 make_il0 (IL0stmtDefTemp (1, type_base_t,
				   ILexpConstant(constant_zero)));
	 make_il0 (IL0stmtDefTemp (2, type_fsc_error,
				   ILexpConstant(CTTconstAbstract "ERR_UNIMPLEMENTED")));
	 make_il0 (IL0stmtDefTemp (3, type_void, ILexpInvoke(lv_abort, [1; 2])));
       ]
    })
      
let generate_function_stub ~genv gs id ct args = 
  require_name (GNtypeinfo ct);
  let typeinfo = 
    make_cttm_expr
      (CTTMexpRead
	 (CTTMlvVar(encoded_typeinfo_name ~genv ct, type_typeinfo_init),
	  ["val", type_typeinfo_struct])) type_typeinfo_struct in
  let allocsize_e = 
    make_cttm_expr (CTTMexpConstant (CTTconstInteger unit_big_int)) type_size_t in
  let header =
    make_cttm_expr (CTTMexpInvoke(lv_cttm_macro_emit_header,
				  [typeinfo; allocsize_e]))
      type_fsc_header
  in
  IL0declVariable
    (gs, type_function_stub_init, 
     translate_global_variable_name ~genv ~known:Known id ct,
     Some 
       (CTTMinitList
	  [CTTMinitExp header;
	   CTTMinitList
	     [CTTMinitExp 
		(make_cttm_expr 
		   (CTTMexpCoerce
		      (type_void_ptr,
		       make_cttm_expr 
			 (CTTMexpRead
			    (CTTMlvVar (translate_global_function_name ~genv id ct, type_void_ptr),
			     []))
			 type_void_ptr))
		   type_void_ptr);
	      CTTMinitExp
		(make_cttm_expr
		   (CTTMexpRead 
		      (CTTMlvVar (translate_global_bridge_function_name ~genv id ct, type_genfunc),
		       []))
		   type_genfunc)]]))

let translate_funcdecl_to_prototype f = 
  match f with
    IL0declFunction(gs, ct, id, args, f) ->
      IL0declVariable(gs, ct, id, None)
  | _ -> assert false

let create_denv t = 
  let denv = Hashtbl.create 16 in
  dprintf 6 "Creating denv...";
  List.iter
    (function
	ILC0declFunction(_gs, _ct, id, _args, _t) ->
	  dprintf 6 "known: %s" id;
	  Hashtbl.add denv id Known;
      | ILC0declVariable(_gs, ct, id, None) ->
	  if C_typing.is_type_loose_function ct then begin
	    dprintf 6 "poly.: %s" id;
	    Hashtbl.add denv id Unknown_polymorphic
	  end else begin
	    dprintf 6 "mono.: %s" id;
	    Hashtbl.add denv id Unknown_monomorphic
	  end
      | ILC0declVariable(_gs, _ct, id, Some _init) ->
	  dprintf 6 "known: %s" id;
	  Hashtbl.add denv id Known)
    t;
  denv

let f ~genv t =
  let denv = create_denv t in
  let function_predeclarations = Glist.empty () in
  dprintf_start "translating...";
  let new_decls = 
    Util.map_flatten
      (function
	  ILC0declFunction(gs, ct, id, args, f) -> begin
	    dprintf_progress "%s" id;
	    let f = translate_function ~genv ~denv gs ct id args f in
	    if !compiler_mode <> StdlibImplementation then begin
	      let stubf = 
		try
		  generate_bridge_function ~genv gs id ct args 
		with
		  FG_not_supported ->
		    generate_dummy_bridge_function ~genv gs id ct args
	      in
	      let stubrec = generate_function_stub ~genv gs id ct args in
	      Glist.append function_predeclarations 
		(Util.list_map translate_funcdecl_to_prototype [f; stubf]);
	      Glist.put function_predeclarations
	      (IL0declVariable(gs,
			       type_function_stub_init,
			       translate_global_variable_name ~genv ~known:Known id ct, None));
	      [f; stubf; stubrec]
	    end
	    else begin
	      Glist.append function_predeclarations 
		(Util.list_map translate_funcdecl_to_prototype [f]);
	      [f]
	    end
	  end
	| ILC0declVariable (gs,ct,id,init) ->
	    match ct.ct_ty with
	      Tfunction _ -> begin
		match gs with
		  Extern l when not (List.mem_assoc "external" l )->
		    let new_id = 
		      translate_global_function_name ~genv id ct in
		    let new_ct = translate_c_type_genv ~genv ct in
		    [IL0declVariable(Extern [], new_ct, new_id, None);
		     IL0declVariable(Extern [],
				     type_function_stub_init,
				     translate_global_variable_name ~genv
				       ~known:(Hashtbl.find denv id) id ct, None)]
		      (* TODO: definition and/or declaration of cast stub function *)
		| _ -> []
	      end
	    | _ ->
		let is_extern, emit, init =
		  match gs with
		    Extern s when List.mem_assoc "emit_typeinfo" s -> 
		      true, false, None
		  | Extern _ -> true, true, None
		  | Global | ModuleStatic -> 
		      false, true, Some (translate_global_initializer ~genv ~denv ct init) 
		in
		let emit_t = get_storage_type ~genv ~incomplete_ok:true ct in
		let emit_t = { emit_t with ct_volatile_p = ct.ct_volatile_p } in
		if emit then begin
		  Glist.append function_predeclarations
		    [IL0declVariable(gs, emit_t, 
				     translate_global_variable_name ~genv ~known:Known id ct, None) ];
		  [IL0declVariable(gs, emit_t,
				   translate_global_variable_name ~genv ~known:Known id ct, init) ]
		end else begin
		  (* get_storage_type is called above, side-effect (require_name) is already performed *)
		  []
		end
      )
      t
  in
  dprintf_end "done.";
  Glist.to_list function_predeclarations @ new_decls
